/*!***************************************************************************
*! FILE NAME  : buffer_monitor.c
*! DESCRIPTION: A useless program to monitor frame size + circ buffer free space and adjust quality. 
*!              Based on the code in autoexposure.c & camogm.c.
*! AUTHOR     : Oleg K Dzhimiev <oleg@elphel.com>
*! Copyright (C) 2010 Elphel, Inc.
*! -----------------------------------------------------------------------------**
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log: buffer_monitor.c,v $
*!  Revision 1.1  2010/09/12 02:02:28  dzhimiev
*!  initial version
*!
*!
*/ 
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <time.h>
#include <string.h>

#include <netinet/in.h> /*little <-> big endian?*/
#include <sys/mman.h>   /* mmap */
#include <sys/ioctl.h>

#include <asm/elphel/c313a.h>
//#include <asm/elphel/exifa.h>
#include <asm/byteorder.h>

#include <ogg/ogg.h> // has to be before ogmstreams.h
//#include "ogmstreams.h" // move it to <>?

#include "buffer_monitor.h"

/// set max frames size to 5.2MB
#define FRAME_SIZE_MAX  5200000
#define FRAME_SIZE_MIN  2500000
#define QUALITY_IDEAL 98

const char ctlFileName[]="/dev/frameparsall";
unsigned long * ccam_dma_buf;           /* mmapped array */

struct framepars_all_t   *frameParsAll;
struct framepars_t       *framePars;
unsigned long            *globalPars; /// parameters that are not frame-related, their changes do not initiate any actions

camogm_state sstate;
camogm_state * state;

FILE* debug_file;

extern int debug_level=0;

int main(int argc, char *argv[]){

  int circbufFreeSpace, circbufSize;
  int quality;
  int frameSize;
  
  int ix=0;
  int jx=0;
    
  const char usage[]= "This program allows something. \
Takes 6 parameters:\n\
  RECORDS_LIMIT \n\
  CONTROL_QUALITY_EN - 0 or 1\n\
  COMPRESSION_QUALITY_GOAL,\n\
  FRAME_SIZE_MAX(in bytes),\n\
  FRAME_SIZE_MIN(in bytes),\n\
  debug level - 0 or 1,\n\
  <full_path>/<log file name>\n\
  EXAMPLE:  eyesis_quality_control 5000 1 98 5200000 2500000 1 /usr/html/var/buffer_monitor.log\n";

  if ((argc < 6) || (argv[1][1]=='-')) {
    printf (usage);
    return 0;
  } 

  debug_file=stderr;
  
  int records_number = atoi(argv[1]);
  int qualityControl = atoi(argv[2]);
  int qualityIdeal = atoi(argv[3]);
  int frameSizeMax = atoi(argv[4]);
  int frameSizeMin = atoi(argv[5]);

  const char *log_file_name = argv[7];
  
  state= &sstate; //extern

  /// Set read/write access (mmap) to Elphel parameters
  initAccessToParameters();

  /// Log file + debug level
  camogm_debug_level(atoi(argv[6]));
  camogm_debug(log_file_name);
  
  /// CIRCBUFSIZE parameter is a constant in fact
  circbufSize=GLOBALPARS(G_CIRCBUFSIZE);

  while(1) {
     ix++;
    
    //get parameters
    quality = GLOBALPARS(P_QUALITY);
    frameSize=GLOBALPARS(G_FRAME_SIZE);
    circbufFreeSpace=GLOBALPARS(G_FREECIRCBUF);
 
    if (qualityControl==1) {
      if (frameSize>frameSizeMax) {
	if (circbufFreeSpace<circbufSize/2){
	  setParameter(P_QUALITY,quality-5);
	  D1(fprintf(debug_file, "1 - fs: %d, q: %d, cbfs: %d, cbs: %d\n", frameSize, quality,circbufFreeSpace,circbufSize));
	}else{
	  setParameter(P_QUALITY,quality-1);
	  D1(fprintf(debug_file, "2 - fs: %d, q: %d, cbfs: %d, cbs: %d\n", frameSize, quality,circbufFreeSpace,circbufSize));
	}
      }else{
	if ((quality<qualityIdeal)&&(frameSize<frameSizeMin)&&(circbufFreeSpace>circbufSize/2)){
	  setParameter(P_QUALITY,quality+1);
	  D1(fprintf(debug_file, "3 - fs: %d, q: %d, cbfs: %d, cbs: %d\n", frameSize, quality,circbufFreeSpace,circbufSize));
	}else{
	  //wait peacefully for the next frame
	  lseek(state->fd_fparmsall, GLOBALPARS(G_THIS_FRAME)+1+LSEEK_FRAME_WAIT_ABS, SEEK_END);
	  D1(fprintf(debug_file, "4 - fs: %d, q: %d, cbfs: %d, cbs: %d\n", frameSize, quality,circbufFreeSpace,circbufSize));
	}
      }
    }else{
      lseek(state->fd_fparmsall, GLOBALPARS(G_THIS_FRAME)+1+LSEEK_FRAME_WAIT_ABS, SEEK_END);
      D1(fprintf(debug_file, "5 - fs: %d, q: %d, cbfs: %d, cbs: %d\n", frameSize, quality,circbufFreeSpace,circbufSize));      
    }  
    
    if (ix==records_number) {
      ix=0;
      fclose(debug_file);
      log_file_name=argv[7];
      //rename(argv[6],strcat(log_file_name,jx));
      jx++;
      
      fopen(argv[7],"w+");
    }
  }
  return 0;
}

int initAccessToParameters(){
    //! Get access to the camera parameters

    //! Open fparmsall and mmap it
    state->fd_fparmsall = open(ctlFileName, O_RDWR);
    if (state->fd_fparmsall<0) { // check control OK
       D0(fprintf (debug_file,"%s:%d:%s: Error opening %s\n",__FILE__,__LINE__,__FUNCTION__, ctlFileName));
       return -2;
    }

    //! Now try to mmap
    /// frameParsAll = (struct framepars_all_t *) mmap(0, sizeof (struct framepars_all_t) , PROT_READ, MAP_SHARED, state->fd_fparmsall, 0);
    /// PROT_WRITE - only to write acknowledge
    frameParsAll = (struct framepars_all_t *) mmap(0, sizeof (struct framepars_all_t) , PROT_READ | PROT_WRITE, MAP_SHARED, state->fd_fparmsall, 0);

    if((int)frameParsAll == -1) {
      D0(fprintf(debug_file,"%s:%d:%s: Error in mmap in %s\n",__FILE__,__LINE__,__FUNCTION__,ctlFileName));
      close(state->fd_fparmsall);
      close(state->fd_circ);
      return -3;
    }
    framePars=frameParsAll->framePars;
    globalPars=frameParsAll->globalPars;

    return 0;
// end of Get access to the camera parameters
}

int setParameter(int par,int value){
  unsigned long this_frame=GLOBALPARS(G_THIS_FRAME);
  unsigned long next_frame=this_frame + 1;
  unsigned long write_data[4]; 
  //MDF2(fprintf(stderr,"next_frame=0x%08lx, this_frame=0x%08lx\n",next_frame,this_frame)); ///======= 0 here
  
  //if (next_frame <= this_frame) return 0; /// too_late
  //else if (next_frame > (this_frame+5) ) { /// wait heer as it is too early to schedule histograms
  //  lseek(state->fd_fparmsall, next_frame-5+LSEEK_FRAME_WAIT_ABS, SEEK_END);
  //  this_frame=GLOBALPARS(G_THIS_FRAME);
  //}
  
  write_data[0]= FRAMEPARS_SETFRAME;
  write_data[1]= next_frame;
  write_data[2]= par;
  write_data[3]= value;
  
  int rslt=write(state->fd_fparmsall, write_data, sizeof(write_data));
  
  if (rslt < sizeof(write_data)) return -errno;
  /// now wait for that next frame (if needed)
  //if (next_frame > this_frame ) {
  lseek(state->fd_fparmsall, next_frame+1+LSEEK_FRAME_WAIT_ABS, SEEK_END);
  //  this_frame=GLOBALPARS(G_THIS_FRAME);
  //  ELP_FERR(fprintf(stderr, "Hey!!!\n"));
    //return 1;
  //}  
  return 0;
}

/**
 * @brief Read either G_* parameter (these are 'regular' values defined by number) or P_* parameter
 *         (it can be read for up to 6 frames ahead, but current interface only allows to read last/current value)
 * @param GPNumber parameter number (as defined in c313a.h), G_* parameters have numbers above FRAMEPAR_GLOBALS, P_* - below)
 * @return parameter value
 */
unsigned long getGPValue(unsigned long GPNumber) {
   return (GPNumber>=FRAMEPAR_GLOBALS)?
              GLOBALPARS(GPNumber):
              framePars[GLOBALPARS(G_THIS_FRAME) & PARS_FRAMES_MASK].pars[GPNumber];
}

/**
 * @brief Set value of the specified global (G_*) parameter
 * @param GNumber - parameter number (as defined in c313a.h)
 * @param value  - value to set
 */
void setGValue(unsigned long  GNumber,   unsigned long value) {
    GLOBALPARS(GNumber)=value;
}

int camogm_debug(const char * fname) {
  int none=1;
  if (fname && strlen(fname) && strcmp(fname, "none") && strcmp(fname, "null")  && strcmp(fname, "/dev/null")) none=0;
  if (debug_file){
    if (strcmp(state->debug_name, "stdout") && strcmp(state->debug_name, "stderr")) fclose (debug_file);
    debug_file=NULL;
    state->debug_name[0]='\0';
  } 
  if (!none) {
    if      (strcmp(fname, "stdout") ==0) debug_file=stdout;
    else if (strcmp(fname, "stderr") ==0) debug_file=stderr;
    else                                  debug_file=fopen(fname,"w+");
  }
  if (debug_file) {
      strncpy(state->debug_name,fname,sizeof(state->debug_name)-1);
      state->debug_name[sizeof(state->debug_name)-1]='\0';
  }
  return 0;
}

int camogm_debug_level(int d) {
  debug_level=d;
  return 0;
}