 
/*!***************************************************************************
*! FILE NAME  : camogm.c
*! DESCRIPTION: Program to write captured video (and audio) to camera file system
*! using Ogg container.
*! Original implementation will copy package data to a buffer to use library calls?
*! Copyright (C) 2007 Elphel, Inc.
*! -----------------------------------------------------------------------------**
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log: camogm.c,v $
*!  Revision 1.5  2009/03/13 13:26:53  spectr_rain
*!  fixed wrong record about videoframes for the first movie with start command, when frames before the audio data was skipped
*!
*!  Revision 1.2  2009/02/25 17:50:02  spectr_rain
*!  removed deprecated dependency
*!
*!  Revision 1.1.1.1  2008/11/27 20:04:01  elphel
*!
*!
*!  Revision 1.6  2008/11/21 01:52:53  elphel
*!  updated for 8.0
*!
*!  Revision 1.5  2008/11/20 23:21:32  elphel
*!  Put FIXME notes and removed parameters that are not used anymore
*!
*!  Revision 1.4  2008/10/29 04:18:28  elphel
*!  v.8.0.alpha10 made a separate structure for global parameters (not related to particular frames in a frame queue)
*!
*!  Revision 1.3  2008/10/13 16:55:53  elphel
*!  removed (some) obsolete P_* parameters, renamed CIRCLSEEK to LSEEK_CIRC constants (same as other similar)
*!
*!  Revision 1.2  2008/09/07 19:48:08  elphel
*!  snapshot
*!
*!  Revision 1.9  2008/04/13 21:05:19  elphel
*!  Fixing KML generation
*!
*!  Revision 1.8  2008/04/11 23:09:33  elphel
*!  modified to handle kml generation
*!
*!  Revision 1.7  2008/04/07 09:13:34  elphel
*!  Changes related to new Exif generation/processing
*!
*!  Revision 1.6  2008/01/14 22:59:00  elphel
*!  7.1.7.4 - added timelapse mode to camogm
*!
*!  Revision 1.5  2007/12/03 08:28:45  elphel
*!  Multiple changes, mostly cleanup
*!
*!  Revision 1.4  2007/11/29 00:38:57  elphel
*!  fixed timescale bug
*!
*!  Revision 1.3  2007/11/19 05:07:19  elphel
*!  fixed 2 typos
*!
*!  Revision 1.2  2007/11/19 03:23:21  elphel
*!  7.1.5.5 Added support for *.mov files in camogm.
*!
*!  Revision 1.1  2007/11/16 08:49:56  elphel
*!  Initial release of camogm - program to record video/image to the camera hard drive (or other storage)
*!
*/ 
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
//#include <ctype.h>
//#include <getopt.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>

#include <netinet/in.h> /*little <-> big endian ?*/
#include <sys/mman.h>		/* mmap */
#include <sys/ioctl.h>
#include <sched.h>

#include <asm/elphel/c313a.h>
#include <asm/byteorder.h>

#include <ogg/ogg.h> // has to be before ogmstreams.h
#include "ogmstreams.h" // move it to <>?

//#include "camogm_exif.h"
#include "camogm_ogm.h"
#include "camogm_jpeg.h"
#include "camogm_mov.h"
#include "camogm_kml.h"
#include "camogm.h"
#include "camogm_wav.h"

#define TRAILER_SIZE   0x02
#define MAP_OPTIONS MAP_FILE|MAP_PRIVATE

char trailer[TRAILER_SIZE] = {0xff,0xd9};

const char ExifFileName[]="/dev/exif_exif";

const char HeadFileName[]="/dev/jpeghead";
const char ctlFileName[]="/dev/frameparsall";
unsigned long * ccam_dma_buf;           /* mmapped array */

int lastDaemonBit= DAEMON_BIT_CAMOGM;
struct framepars_all_t   *frameParsAll;
struct framepars_t       *framePars;
unsigned long            *globalPars; /// parameters that are not frame-related, their changes do not initiate any actions


int buff_size;

#define DEFAULT_DURATION 60 /*!default segment duration (seconds) */
#define DEFAULT_LENGTH 100000000 /*!default segment length (B) */
//!Next 2 for Quicktime (mostly)
#define DEFAULT_FRAMES           16384 /* Maximal number of frames in file segment (each need 4* (1 + 1/frames_per_chunk) bytes for the frame index */
//#define DEFAULT_FRAMES_PER_CHUNK    10 /*second sparse index - for Quicktime fast forward */
#define DEFAULT_FRAMES_PER_CHUNK    1 /*second sparse index - for Quicktime fast forward */

#define DEFAULT_LENGTH 100000000 /*!default segment length (B) */
#define DEFAULT_EXIF 1 /* use Exif */

#define DEFAULT_AUDIO_RATE  44100
#define DEFAULT_AUDIO_CHANNELS  2

#define AUDIO_RATE_MAX  44100
#define AUDIO_RATE_MIN  11025
#define AUDIO_CHANNELS_MAX  2
#define AUDIO_CHANNELS_MIN  1
#define AUDIO_DEF_VOLUME	"58981"

static char cmdbuf[1024];
static int  cmdbufp=0; // current input pointer in the command buffer (read from pipe)
static int  cmdstrt=0; // start of the next partial command

camogm_state sstate;
camogm_state * state;

int   debug_level;
FILE* debug_file;

int camogm_init(void);

int camogm_start(int by_command);	// by_command == 1 if started outside,
int camogm_stop(int by_command);	// == 0 - if restart for next movie file

int camogm_reset(void); //! reset circbuf read pointer
int camogm_debug(const char * fname);
int camogm_debug_level(int d);
void  camogm_set_segment_duration(int sd);
void  camogm_set_segment_length(int sl);
void  camogm_set_save_gp(int d);
void  camogm_set_prefix (const char * p);
void  camogm_set_exif(int d);
void  camogm_set_timescale(double d); //! set timescale, default=1.0
void  camogm_set_frames_skip(int d); //! set number of frames to skip, if negative - seconds between frames
void  camogm_set_format(int d);

void  camogm_kml_set_enable(int d);
void  camogm_kml_set_horHalfFov (double dd);
void  camogm_kml_set_vertHalfFov(double dd);
void  camogm_kml_set_height_mode(int d);
void  camogm_kml_set_height(double dd);
void  camogm_kml_set_period(int d);
void  camogm_kml_set_near(double dd); // distance to PhotoOverlay


int   parse_cmd(FILE* npipe);
char * getLineFromPipe(FILE* npipe);

int  sendImageFrame (void);

void  camogm_set_max_frames(int d);
void  camogm_set_frames_per_chunk(int d);

void camogm_set_audio_state(char *args);
void camogm_set_audio_format(char *args);
void camogm_set_audio_volume(char *args);
void camogm_set_sync_state(char *args);
void audio_process(void);
void audio_init(int by_command);
void audio_start(int by_command);
void audio_end(int by_command);
void audio_free(void);
void camogm_audio_initialize(void);
void audio_finish(int by_command);
void audio_get_sys_fpga_times(unsigned long *rtc);
void audio_set_volume(int nvolume);
//!======================================================================================================
int time_comp(struct timeval *t1, struct timeval *t2) {
	if(t1->tv_sec > t2->tv_sec)
		return 1;
	if(t1->tv_sec == t2->tv_sec) {
		if(t1->tv_usec > t2->tv_usec)
			return 1;
		if(t1->tv_usec == t2->tv_usec)
			return 0;
	}
	return -1;
}

void time_normalize(struct timeval *tv) {
	tv->tv_sec += tv->tv_usec / 1000000;
	tv->tv_usec = tv->tv_usec % 1000000;
/*
	while(tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec += 1;
	}
*/
}

struct timeval time_delta(struct timeval *t1, struct timeval *t2) {
	struct timeval tv;
	tv.tv_sec = t1->tv_sec - t2->tv_sec;
	if(t1->tv_usec >= t2->tv_usec) {
		tv.tv_usec = t1->tv_usec - t2->tv_usec;
	} else {
		tv.tv_usec = 1000000 + t1->tv_usec - t2->tv_usec;
		tv.tv_sec -= 1;
	}
	return tv;
}

void put_uint16(void *buf, u_int16_t val)
{
	unsigned char  *tmp;

	tmp = (unsigned char *) buf;

	tmp[0] = val & 0xff;
	tmp[1] = (val >>= 8) & 0xff;
}

void put_uint32(void *buf, u_int32_t val)
{
	unsigned char  *tmp;

	tmp = (unsigned char *) buf;

	tmp[0] = val & 0xff;
	tmp[1] = (val >>= 8) & 0xff;
	tmp[2] = (val >>= 8) & 0xff;
	tmp[3] = (val >>= 8) & 0xff;
}

void put_uint64(void *buf, u_int64_t val)
{
	unsigned char  *tmp;

	tmp = (unsigned char *) buf;

	tmp[0] = val & 0xff;
	tmp[1] = (val >>= 8) & 0xff;
	tmp[2] = (val >>= 8) & 0xff;
	tmp[3] = (val >>= 8) & 0xff;
	tmp[4] = (val >>= 8) & 0xff;
	tmp[5] = (val >>= 8) & 0xff;
	tmp[6] = (val >>= 8) & 0xff;
	tmp[7] = (val >>= 8) & 0xff;
}

int camogm_init(void) {
   const char sserial[]="elp0";
   int * ipser= (int*) sserial;
   state->running=0; // mo
   state->starting=0; // mo
//   state->vf=NULL;
   state->vf=0;
   camogm_set_segment_duration(DEFAULT_DURATION);
   camogm_set_segment_length(DEFAULT_LENGTH);
   camogm_set_max_frames(DEFAULT_FRAMES);
   camogm_set_frames_per_chunk(DEFAULT_FRAMES_PER_CHUNK);
   camogm_set_prefix ("\0");
   camogm_set_save_gp(0);
   camogm_reset(); //! sets    state->buf_overruns=-1; //!first does not count
   state->serialno= ipser[0];
   state->last= 0;
   debug_file= stderr;
   camogm_debug_level(1);
   strcpy(state->debug_name,"stderr");
   camogm_set_timescale(1.0);
   camogm_set_frames_skip(0); //! don't skip
   camogm_set_format(CAMOGM_FORMAT_OGM);
   camogm_set_audio_state("on");
   camogm_set_audio_format("44100/2");
   camogm_set_audio_volume(AUDIO_DEF_VOLUME);
   camogm_set_audio_state("off");
// camogm_set_sync_state("on");
   camogm_set_sync_state("off");
   state->exifSize=0;
   state->exif= DEFAULT_EXIF;
   state->frame_lengths=NULL;
   state->frameno=0;
   state->formats=0;

   camogm_audio_initialize();

///kml stuff
   camogm_kml_set_enable(0);
   state->kml_file=NULL;
   camogm_kml_set_horHalfFov (20.0);
   camogm_kml_set_vertHalfFov(15.0);
   camogm_kml_set_height_mode(0);
   camogm_kml_set_height(10.0);
   camogm_kml_set_period(2); // 2 sec
   camogm_kml_set_near(40.0); // 40 m (distance to PhotoOverlay)
   state->kml_path[0]='\0';


/*
void  camogm_kml_set_enable(int d);
void  camogm_kml_set_horHalfFov (double dd);
void  camogm_kml_set_vertHalfFov(double dd);
void  camogm_kml_set_height_mode(int d);
void  camogm_kml_set_height(double dd);
void  camogm_kml_set_decimate(int d);

  int                   kml_enable;       //! enable KML file generation
  FILE*                 kml_file;         //! stream to write kml file
  double                kml_horHalfFov;   //! half horizontal Fov (degrees)
  double                kml_vertHalfFov;  //! half vertical Fov (degrees)
  int                   kml_height_mode;  //! 1 - actual, 0 - ground
  double                kml_height;       //! extra height to add
  int                   kml_decimate;     //! generate PhotoOverlay for each kml_decimate frame;
  int                   kml_decimate_counter; //! counter to skip kml frames
  double                kml_near;         //! Use in KML "near" parameter (<=0 - don't use it)

*/



   return 0;
}


int camogm_debug(const char * fname) {
  int none=1;
  if (fname && strlen(fname) && strcmp(fname, "none") && strcmp(fname, "null")  && strcmp(fname, "/dev/null")) none=0;
  if (debug_file){
    if (strcmp(state->debug_name, "stdout") && strcmp(state->debug_name, "stderr")) fclose (debug_file);
    debug_file=NULL;
    state->debug_name[0]='\0';
  }
  if (!none) {
    if      (strcmp(fname, "stdout") ==0) debug_file=stdout;
    else if (strcmp(fname, "stderr") ==0) debug_file=stderr;
    else                                  debug_file=fopen(fname,"w+");
  }
  if (debug_file) {
      strncpy(state->debug_name,fname,sizeof(state->debug_name)-1);
      state->debug_name[sizeof(state->debug_name)-1]='\0';
  }
  return 0;
}

int camogm_debug_level(int d) {
  debug_level=d;
  return 0;
}




int camogm_start(int by_command) {
  int timestamp_start;
  int rslt;
  int next_metadata_start, next_jpeg_len, fp;
  D1(fprintf (debug_file,"Starting recording(%d)\n", by_command));
  state->frameno=0;
  state->timescale=        state->set_timescale;//! current timescale, default 1
///debug
  int * ifp =      (int *) &(state->frame_params) ;
  int * ifp_this = (int *) &(state->this_frame_params) ;
  if (state->kml_enable) camogm_init_kml() ; // do nothing

  audio_init(by_command);
  if (state->format != state->set_format) {
     state->format=  state->set_format;
     switch (state->format) {
      case CAMOGM_FORMAT_NONE: rslt= 0; break;
      case CAMOGM_FORMAT_OGM:  rslt= camogm_init_ogm(); break;
      case CAMOGM_FORMAT_JPEG: rslt= camogm_init_jpeg();break;
      case CAMOGM_FORMAT_MOV:  rslt= camogm_init_mov(); break;
     }
     state->formats |= 1 << (state->format);
    //! exit on unknown formats?
  }
  state->max_frames=       state->set_max_frames;
  state->frames_per_chunk= state->frames_per_chunk;
  state->starting=1; //!may be already set
//! Check/set circbuf read pointer
D3(fprintf (debug_file,"1: state->cirbuf_rp=0x%x\n",state->cirbuf_rp));
  if ((state->cirbuf_rp <0) || (lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET) < 0) || (lseek(state->fd_circ,LSEEK_CIRC_VALID,SEEK_END) < 0 )) {
D3(fprintf (debug_file,"2: state->cirbuf_rp=0x%x\n",state->cirbuf_rp));
    state->cirbuf_rp=lseek(state->fd_circ,LSEEK_CIRC_LAST,SEEK_END);
    if (((fp=lseek(state->fd_circ,LSEEK_CIRC_PREV,SEEK_END)))>=0) state->cirbuf_rp=fp; //!try to have 2 frames available for fps
    state->buf_overruns++;
//! file pointer here should match state->rp; so no need to do    lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET);
//    state->buf_min=lseek(state->fd_circ,LSEEK_CIRC_FREE,SEEK_END);
//    lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET); //! restore after moving (side effect)
    state->buf_min=getGPValue(G_FREECIRCBUF);

  } else {
//      b_free=lseek(state->fd_circ,LSEEK_CIRC_FREE,SEEK_END);
//      lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET); //! restore after moving (side effect)
       if (state->buf_min > getGPValue(G_FREECIRCBUF)) state->buf_min=getGPValue(G_FREECIRCBUF);

  }
D3(fprintf (debug_file,"3: state->cirbuf_rp=0x%x\n",state->cirbuf_rp));
D3(fprintf (debug_file,"4:lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END)=%d\n",(int) lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END)));

//! is this frame ready?
  if (lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END) <0) return -CAMOGM_FRAME_NOT_READY; //! frame pointer valid, but no frames yet
D3(fprintf (debug_file,"5: state->cirbuf_rp=0x%x\n",state->cirbuf_rp));
  state->metadata_start=(state->cirbuf_rp)-32;
  if (state->metadata_start<0) state->metadata_start+=state->circ_buff_size;

///==================================

  memcpy (&(state->frame_params), (unsigned long * ) &ccam_dma_buf[state->metadata_start>>2],32);
  state->jpeg_len=state->frame_params.frame_length; //! frame_params.frame_length are now the length of bitstream


  if (state->frame_params.signffff !=0xffff) {
    D0(fprintf(debug_file, "%s:%d: wrong signature - %d\r\n",__FILE__,__LINE__,(int) state->frame_params.signffff));
     state->cirbuf_rp=-1;
     D1(fprintf(debug_file, "state->cirbuf_rp=0x%x\r\n",(int) state->cirbuf_rp));
     D1(fprintf(debug_file, "%08x %08x %08x %08x %08x %08x %08x %08x\r\n",ifp[0],ifp[1],ifp[2],ifp[3],ifp[4],ifp[5],ifp[6],ifp[7]));
     return -CAMOGM_FRAME_BROKEN;
  }
//!   find location of the timestamp and copy it to the frame_params structure
///==================================
  timestamp_start=(state->cirbuf_rp)+((state->jpeg_len+CCAM_MMAP_META+3) & (~0x1f)) + 32 - CCAM_MMAP_META_SEC; //! magic shift - should index first byte of the time stamp
  if (timestamp_start >= state->circ_buff_size) timestamp_start-=state->circ_buff_size;
  memcpy (&(state->frame_params.timestamp_sec), (unsigned long * ) &ccam_dma_buf[timestamp_start>>2],8);

//! see if next frame is available   
  if ((lseek(state->fd_circ,LSEEK_CIRC_NEXT,SEEK_END) < 0 ) ||
//! is that next frame ready?
      (((fp=lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END))) < 0)) {
        lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET); //!just in case - restore pointer
        return -CAMOGM_FRAME_NOT_READY; //! frame pointer valid, but no frames yet
  }
  next_metadata_start=fp-32;
  if (next_metadata_start<0) next_metadata_start+= state->circ_buff_size;
  memcpy (&(state->this_frame_params), (unsigned long * ) &ccam_dma_buf[next_metadata_start>>2],32);
  next_jpeg_len=state->this_frame_params.frame_length; //! frame_params.frame_length are now the length of bitstream
  if (state->this_frame_params.signffff !=0xffff) { //! should not happen ever
     D0(fprintf(debug_file, "%s:%d: wrong signature - %d\r\n",__FILE__,__LINE__,(int) state->this_frame_params.signffff));
     D1(fprintf(debug_file, "fp=0x%x\r\n",(int) fp));
     D1(fprintf(debug_file, "%08x %08x %08x %08x %08x %08x %08x %08x\r\n",ifp_this[0],ifp_this[1],ifp_this[2],ifp_this[3],ifp_this[4],ifp_this[5],ifp_this[6],ifp_this[7]));
//  int * ifp =      (int *) &(state->this_frame_params) ;
//  int * ifp_this = (int *) &(state->this_frame_params) ;

     state->cirbuf_rp=-1;
     return -CAMOGM_FRAME_BROKEN;
  }
//! find location of the timestamp and copy it to the frame_params structure
  timestamp_start=fp+((next_jpeg_len+CCAM_MMAP_META+3) & (~0x1f)) + 32 - CCAM_MMAP_META_SEC; //! magic shift - should index first byte of the time stamp
  if (timestamp_start >= state->circ_buff_size) timestamp_start-=state->circ_buff_size;
  memcpy (&(state->this_frame_params.timestamp_sec), (unsigned long * ) &ccam_dma_buf[timestamp_start>>2],8);
//! verify that the essential current frame params did not change, if they did - return an error (need new file header) 
  if ((state->frame_params.width  != state->this_frame_params.width) ||
      (state->frame_params.height != state->this_frame_params.height)) {
//! Advance frame pointer to the next (caller should try again)
     state->cirbuf_rp=fp;
     return -CAMOGM_FRAME_CHANGED; // no yet checking for the FPS
  }
//! calcualte the frame period - time difference (in microseconds)
   state->frame_period=(state->this_frame_params.timestamp_usec - state->frame_params.timestamp_usec)+
               1000000*(state->this_frame_params.timestamp_sec  - state->frame_params.timestamp_sec);

//! correct for timelapse modes:
  state->frames_skip=      state->set_frames_skip;
  if (state->frames_skip > 0) {
    state->frames_skip_left=0;
    state->frame_period *= (state->frames_skip+1);
//    state->frames_skip_left= state->set_frames_skip;
  } else if (state->frames_skip < 0) {
    state->frame_period=-(state->frames_skip) ; //! actual frame period will fluctuate to the nearest frame acquired (free running)
    state->frames_skip_left=state->frame_params.timestamp_sec;
  }
   state->time_unit= (ogg_int64_t) (((double) state-> frame_period) * ((double) 10) / ((double) state-> timescale));
   state->width=state->frame_params.width;
   state->height=state->frame_params.height;

//!read JPEG header - it should stay the same for the whole file (restart new file if any parameters changed)
//!rebuild JPEG header:
   lseek(state->fd_head,state->cirbuf_rp+1,SEEK_END); //!+1 to avoid condition when jpeg_start==0. overloaded lseek will ignore 5 LSBs when SEEK_END
   state->head_size=lseek(state->fd_head,0,SEEK_END); /// In 8.0 the header size might change for some jp4 modes 
   if (state->head_size>JPEG_HEADER_MAXSIZE) {
     D0(fprintf (debug_file,"%s:%d: Too big JPEG header (%d > %d)",__FILE__,__LINE__,state->head_size, JPEG_HEADER_MAXSIZE ));
     return -2;
   }
//! and read it
   lseek(state->fd_head,0,0);
   read (state->fd_head,state->jpegHeader,state->head_size);
//! Restore read pointer to the original (now there may be no frame ready there yet)
   lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET);

//!here we are ready to initialize Ogm file
   audio_start(by_command);
   switch (state->format) {
    case CAMOGM_FORMAT_NONE: rslt= 0;  break;
    case CAMOGM_FORMAT_OGM:  rslt= camogm_start_ogm();  break;
    case CAMOGM_FORMAT_JPEG: rslt= camogm_start_jpeg(); break;
    case CAMOGM_FORMAT_MOV:  rslt= camogm_start_mov();  break;
    default: rslt=0; // do nothing
   }
   if (rslt) return rslt;
   if (state->kml_enable) rslt=camogm_start_kml() ; // will turn on state->kml_used if it can
   if (rslt) return rslt;
   state->running=1;
   state->starting=0;
   D1(fprintf (debug_file,"Started OK\n"));
   return 0;
}

int  sendImageFrame (void) {
   int rslt;
//   struct interframe_params_t this_frame_params;
//   elph_packet_chunk packetchunks[7];
   unsigned char frame_packet_type = PACKET_IS_SYNCPOINT;
//   int chunk_index=0;
//   int i;
//   unsigned long * tts= (unsigned long *) &(state->granulepos); //to word-swap timestamt to granule
   int timestamp_start;
///debugging:
//   int * ifp =      (int *) &(state->frame_params) ;
   int * ifp_this = (int *) &(state->this_frame_params) ;

//! This is probably needed only for Quicktime (not to exceed already allocated frame index)
   if (state->frameno >= (state->max_frames)) return -CAMOGM_FRAME_CHANGED;
//! Format changed?
//   D3(fprintf (debug_file,"sendImageFrame: format=%d, set_format=%d\n", state->format, state->set_format));

   if (state->format != state->set_format) return -CAMOGM_FRAME_CHANGED;
//!   check if file size is exceeded (assuming fopen),-CAMOGM_FRAME_CHANGED will trigger a new segment
//   if ((state->vf) && (state->segment_length >=0) && (ftell(state->vf) > state->segment_length)) return -CAMOGM_FRAME_CHANGED;
   if((state->vf) && (state->segment_length >= 0) && (lseek(state->vf, 0, SEEK_CUR) > state->segment_length))
      return -CAMOGM_FRAME_CHANGED;
//!same for open
   if (((state->ivf)>=0) && (state->segment_length >=0) && (lseek(state->ivf, 0, SEEK_CUR) > state->segment_length)) return -CAMOGM_FRAME_CHANGED;
//! check the frame pointer is valid
   if (lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET) <0) return -CAMOGM_FRAME_INVALID; //!it will probably be that allready
//! is the frame ready?
   if (lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END) <0) return -CAMOGM_FRAME_NOT_READY; //! frame pointer valid, but no frames yet
/*
//! correct for timelapse modes:
  state->frames_skip=      state->set_frames_skip;
  if (state->frames_skip > 0) {
    state->frames_skip_left=0;
    state->frame_period *= (state->frames_skip+1);
//    state->frames_skip_left= state->set_frames_skip;
  } else if (state->frames_skip < 0) {
    state->frame_period=-(state->frames_skip) ; //! actual frame period will fluctuate to the nearest frame acquired (free running)
    state->frames_skip_left=state->frame_params.timestamp_sec;
  }
*/

//! process skipping frames. TODO: add - skipping time between frames (or better -  actual time period - use the nearest frame) instead of the frame number
   if ( (state->frames_skip > 0) && (state->frames_skip_left > 0 )) { //!skipping frames, not seconds.
     state->cirbuf_rp=lseek(state->fd_circ,LSEEK_CIRC_NEXT,SEEK_END);
//!optionally save it to global read pointer (i.e. for debugging with imgsrv "/pointers")
     if (state->save_gp) lseek(state->fd_circ,LSEEK_CIRC_SETP,SEEK_END);
     state->frames_skip_left--;
     return -CAMOGM_FRAME_NOT_READY; //! the required frame is not ready
   }

//! Get metadata
//D3(fprintf (debug_file,"_1_"));
   state->metadata_start=state->cirbuf_rp-32;
   if (state->metadata_start<0) state->metadata_start+=state->circ_buff_size;
   memcpy (&(state->this_frame_params), (unsigned long * ) &ccam_dma_buf[state->metadata_start>>2],32);
   state->jpeg_len=state->this_frame_params.frame_length; //! frame_params.frame_length are now the length of bitstream
   if (state->this_frame_params.signffff !=0xffff) {
     D0(fprintf(debug_file, "%s:%d: wrong signature - %d\r\n",__FILE__,__LINE__,(int) state->this_frame_params.signffff));
     D1(fprintf(debug_file, "state->cirbuf_rp=0x%x\r\n",(int) state->cirbuf_rp));
     D1(fprintf(debug_file, "%08x %08x %08x %08x %08x %08x %08x %08x\r\n",ifp_this[0],ifp_this[1],ifp_this[2],ifp_this[3],ifp_this[4],ifp_this[5],ifp_this[6],ifp_this[7]));


     return -CAMOGM_FRAME_BROKEN;
   }
//D3(fprintf (debug_file,"_2_"));
//!   find location of the timestamp and copy it to the frame_params structure
   timestamp_start=state->cirbuf_rp+((state->jpeg_len+CCAM_MMAP_META+3) & (~0x1f)) + 32 - CCAM_MMAP_META_SEC; //! magic shift - should index first byte of the time stamp
   if (timestamp_start >= state->circ_buff_size) timestamp_start-=state->circ_buff_size;
//D3(fprintf (debug_file,"_3_"));
   memcpy (&(state->this_frame_params.timestamp_sec), (unsigned long * ) &ccam_dma_buf[timestamp_start>>2],8);
//! verify that the essential current frame params did not change, if they did - return an error (need new file header) 
   if ((state->frame_params.width  != state->this_frame_params.width) ||
       (state->frame_params.height != state->this_frame_params.height)) return -CAMOGM_FRAME_CHANGED; //! not yet checking for the FPS
//!   check if file duration (in seconds) exceeded ,-CAMOGM_FRAME_CHANGED will trigger a new segment
   if ((state->segment_duration > 0) && 
       ((state->this_frame_params.timestamp_sec-state->frame_params.timestamp_sec) > state->segment_duration))
           return -CAMOGM_FRAME_CHANGED;

	struct timeval ts;
	ts.tv_sec = state->this_frame_params.timestamp_sec;
	ts.tv_usec = state->this_frame_params.timestamp_usec;

//! check if (in timelapse mode)  it is too early for the frame to be stored
   if ((state->frames_skip < 0) && (state->frames_skip_left > state->this_frame_params.timestamp_sec) ) {
     state->cirbuf_rp=lseek(state->fd_circ,LSEEK_CIRC_NEXT,SEEK_END);
//!optionally save it to global read pointer (i.e. for debugging with imgsrv "/pointers")
     if (state->save_gp) lseek(state->fd_circ,LSEEK_CIRC_SETP,SEEK_END);
     return -CAMOGM_FRAME_NOT_READY; //! the required frame is not ready
   }


//D3(fprintf (debug_file,"_4_"));
   if (state->exif) {
D3(fprintf (debug_file,"_5_"));
//! update the Exif header with the current frame metadata
//     updateExif(ep, state->ed, &(state->frame_params));
     state->exifSize=lseek(state->fd_exif,1,SEEK_END); // at the beginning of page 1 - position == page length
//     if (state->exifSize < 0) state->exifSize=0; // error from lseek;
     if (state->exifSize > 0) {
//state->this_frame_params.meta_index
       lseek(state->fd_exif,state->this_frame_params.meta_index,SEEK_END); //! select meta page to use (matching frame)
       rslt=read (state->fd_exif, state->ed, state->exifSize);
       if (rslt<0) rslt=0;
       state->exifSize=rslt;
     } else state->exifSize=0;
   } else state->exifSize=0;

//D3(fprintf (debug_file,"_6_"));

//! prepare a packet to be sent (a lst of memory chunks)
   state->chunk_index=0;
   state->packetchunks[state->chunk_index  ].bytes=1;
   state->packetchunks[state->chunk_index++].chunk=&frame_packet_type;
   if (state->exif>0) {//! insert Exif
//D3(fprintf (debug_file,"_7_"));
     state->packetchunks[state->chunk_index  ].bytes=2;
     state->packetchunks[state->chunk_index++].chunk=state->jpegHeader;
     state->packetchunks[state->chunk_index  ].bytes=state->exifSize;
     state->packetchunks[state->chunk_index++].chunk= state->ed;
     state->packetchunks[state->chunk_index  ].bytes=state->head_size-2;
     state->packetchunks[state->chunk_index++].chunk= &(state->jpegHeader[2]);
   } else {
//D3(fprintf (debug_file,"_8_"));
     state->packetchunks[state->chunk_index  ].bytes=state->head_size;
     state->packetchunks[state->chunk_index++].chunk= state->jpegHeader;
   }
//D3(fprintf (debug_file,"_9_"));

/*! JPEG image data may be split in two segments (rolled over buffer end) - process both variants */
//     if ((jpeg_start+jpeg_len) > state->circ_buff_size) { // two segments
   if ((state->cirbuf_rp + state->jpeg_len) > state->circ_buff_size) { //! two segments
/*! copy from the beginning of the frame to the end of the buffer */
//D3(fprintf (debug_file,"_10_"));

//        sendBuffer((void *) &ccam_dma_buf[jpeg_start>>2], state->circ_buff_size-state->cirbuf_rp);
     state->packetchunks[state->chunk_index  ].bytes=state->circ_buff_size-state->cirbuf_rp;
     state->packetchunks[state->chunk_index++].chunk= (unsigned char*) &ccam_dma_buf[state->cirbuf_rp>>2];
/*! copy from the beginning of the buffer to the end of the frame */
//        sendBuffer((void *) &ccam_dma_buf[0], jpeg_len-(state->circ_buff_size-state->cirbuf_rp));
     state->packetchunks[state->chunk_index  ].bytes=state->jpeg_len - (state->circ_buff_size-state->cirbuf_rp);
     state->packetchunks[state->chunk_index++].chunk= (unsigned char*) &ccam_dma_buf[0];
   } else { // single segment
//D3(fprintf (debug_file,"_11_"));

/*! copy from the beginning of the frame to the end of the frame (no buffer rollovers) */
//        sendBuffer((void *) &ccam_dma_buf[state->cirbuf_rp>>2], state->jpeg_len);
     state->packetchunks[state->chunk_index  ].bytes=state->jpeg_len;
     state->packetchunks[state->chunk_index++].chunk= (unsigned char*)  &ccam_dma_buf[state->cirbuf_rp>>2];
   }
//D3(fprintf (debug_file,"_12_"));
   state->packetchunks[state->chunk_index  ].bytes=2;
   state->packetchunks[state->chunk_index++].chunk= (unsigned char*) trailer;

	// synchronize audio and video here - skip video if needed...
	// analyze frame timestamp...
/*
	long delta = ts.tv_sec - v_ts.tv_sec;
	delta *= 1000000;
	delta += ts.tv_usec;
	delta -= v_ts.tv_usec;
//	fprintf(stderr, "timestamp: %d:%06d; delta == %06d\n", ts.tv_sec, ts.tv_usec, delta);
	v_ts = ts;
*/
	state->ts_video = ts;

	int flag = 1;
	if(ctx_a.begin_of_stream_with_audio) {
		if(ctx_a.begin_of_stream_with_audio_trigger) {
			state->ts_video_start = state->ts_audio;
			state->ts_video_start.tv_usec += state->frame_period / 2;
			time_normalize(&state->ts_video_start);
			ctx_a.begin_of_stream_with_audio_trigger = 0;
			// compute how many audio samples we need to skip here
			long audio_to_skip = 0;
			struct timeval tv = state->ts_video;
			while(time_comp(&tv, &state->ts_video_start) < 0) {
				tv.tv_usec += state->frame_period;
				time_normalize(&tv);
			}
			tv.tv_sec -= 1;
			tv.tv_usec += 1000000;
			tv.tv_usec -= state->frame_period / 2;
			time_normalize(&tv);
			if(tv.tv_sec != state->ts_audio.tv_sec) {
				audio_to_skip = tv.tv_sec - state->ts_audio.tv_sec;
				audio_to_skip *= 1000000;
			}
			audio_to_skip += tv.tv_usec;
			audio_to_skip -= state->ts_audio.tv_usec;
			double s = state->audio_rate;
			s /= 1000.0;
			s *= audio_to_skip;
			s /= 1000.0;
			ctx_a.audio_skip_samples = (long)s;
			ctx_a.movie_start = tv;
DA(			fprintf(stderr, "____ AUDIO started from: %d:%06d; we need to record it from: %d:%06d; audio_to_skip_us == %d; audio samples to skip == %d\n", state->ts_audio.tv_sec, state->ts_audio.tv_usec, tv.tv_sec, tv.tv_usec, audio_to_skip, ctx_a.audio_skip_samples););
		}
DA(		fprintf(stderr, "AUDIO (start): %d:%06d; VIDEO (current): %d:%06d; frame period is: %d\n", state->ts_audio.tv_sec, state->ts_audio.tv_usec, state->ts_video.tv_sec, state->ts_video.tv_usec, state->frame_period););
		if(time_comp(&state->ts_video_start, &state->ts_video) > 0) {
DA(			fprintf(stderr, "skip the video frame\n"););
			flag = 0;
		} else {
DA(			fprintf(stderr, "save the VIDEO frame with time: %d:%06d\n", state->ts_video.tv_sec, state->ts_video.tv_usec););
			ctx_a.begin_of_stream_with_audio = 0;
		}
	}
	if(flag) {
		switch (state->format) {
			case CAMOGM_FORMAT_NONE: rslt=0; break;
			case CAMOGM_FORMAT_OGM:  rslt = camogm_frame_ogm(); break;
			case CAMOGM_FORMAT_JPEG: rslt = camogm_frame_jpeg(); break;
			case CAMOGM_FORMAT_MOV:  rslt = camogm_frame_mov(); break;
			default: rslt = 0; // do nothing
		}
	} else {
		//! skip the only first videoframes that are before of the audio data by timestamp;
		//! this will happen only at the first movie after "start" command
		rslt = 0;
	}

   if (rslt) return rslt;
   if (state->kml_used) rslt=camogm_frame_kml() ; // will turn on state->kml_used if it can
   if (rslt) return rslt;

//D3(fprintf (debug_file,"_14_"));
//!advance frame pointer
	if(flag)
		state->frameno++;
   state->cirbuf_rp=lseek(state->fd_circ,LSEEK_CIRC_NEXT,SEEK_END);
//!optionally save it to global read pointer (i.e. for debugging with imgsrv "/pointers")
   if (state->save_gp) lseek(state->fd_circ,LSEEK_CIRC_SETP,SEEK_END);
//D3(fprintf (debug_file,"_15_\n"));
//D3(fprintf (debug_file,"state->cirbuf_rp=0x%08x granulepos=0x%x 0x%x (%d.%06d)\n", state->cirbuf_rp, (int) ((state->granulepos)>>32), (int) (state->granulepos),
//                                                                (int) state->this_frame_params.timestamp_sec, (int)state->this_frame_params.timestamp_usec));
  if (state->frames_skip > 0) {
     state->frames_skip_left= state->frames_skip;
  } else if (state->frames_skip < 0) {
     state->frames_skip_left+= -(state->frames_skip);
  }
   return 0;
}

int camogm_stop(int by_command) {
  int rslt=0;
  if (!state->running) {
    if (!state->starting) {
      D2(fprintf (debug_file,"Recording was not running, nothing to stop\n"));
    } else {
      state->starting=0;
      D1(fprintf (debug_file,"Dropping attempt to start\n"));
    }
    return 0;
  }
  D1(fprintf (debug_file,"Ending recording\n"));
   if (state->kml_used) camogm_end_kml() ;
	if(ctx_a.audio_present)
		audio_finish(by_command);

   switch (state->format) {
    case CAMOGM_FORMAT_NONE: rslt= 0; break;
    case CAMOGM_FORMAT_OGM:  rslt= camogm_end_ogm(); break;
    case CAMOGM_FORMAT_JPEG: rslt= camogm_end_jpeg();break;
    case CAMOGM_FORMAT_MOV:  rslt= camogm_end_mov(); break;
///    default: return 0; // do nothing
   }
//! now close video file (if it is open)
	if(state->vf)
		close(state->vf);
//		fclose(state->vf);
//	state->vf = NULL;
	state->vf = 0;
	if(rslt)
		return rslt;
	state->last = 1;
//!state->running=0 should be output after file is finished and closed
	state->running = 0;
	state->starting = 0;
	return 0;
}

void camogm_free() {
   int f;
//! free all file format handlers that were used
//add kml when needed
   for (f=0;f<31;f++) {
     if (state->formats & ( 1 << (state->format))) {
       switch (f) {
        case CAMOGM_FORMAT_NONE: break;
        case CAMOGM_FORMAT_OGM:  camogm_free_ogm(); break;
        case CAMOGM_FORMAT_JPEG: camogm_free_jpeg();break;
        case CAMOGM_FORMAT_MOV:  camogm_free_mov(); break;
       }
     }
   }
   state->formats=0;
}

int camogm_reset(void) { //! reset circbuf read pointer
   state->cirbuf_rp=-1;
   state->buf_overruns=-1; //!first will not count
   return 0;
}

///kml stuff
void  camogm_kml_set_enable(int d) {
   state->kml_enable=d;
}
void  camogm_kml_set_horHalfFov (double dd) {
   state->kml_horHalfFov=dd;
}
void  camogm_kml_set_vertHalfFov(double dd) {
   state->kml_vertHalfFov=dd;
}
void  camogm_kml_set_height_mode(int d) {
   state->kml_height_mode=d;
}
void  camogm_kml_set_height(double dd) {
   state->kml_height=dd;
}
void  camogm_kml_set_period(int d) {
   state->kml_period=d;
   state->kml_last_ts=0;
   state->kml_last_uts=0;
}
void  camogm_kml_set_near(double dd) { // distance to PhotoOverlay
   state->kml_near=dd;
}


void  camogm_set_segment_duration(int sd) {
   state->segment_duration=sd;
}
void  camogm_set_segment_length(int sl) {
   state->segment_length=   sl;
}
void  camogm_set_save_gp(int d) {
   state->save_gp=   d;
}
void  camogm_set_exif(int d) {
   state->exif=   d;
}

void  camogm_set_prefix (const char * p) {
  strncpy(state->path_prefix, p, sizeof(state->path_prefix)-1);
  state->path_prefix[sizeof(state->path_prefix)-1]='\0';
}

void  camogm_set_timescale(double d) { //! set timescale, default=1,000,000
   state->set_timescale=  d;
   if ((state->running==0) && (state->starting==0)) {
     state->timescale=state->set_timescale;
   }
}

void  camogm_set_frames_skip(int d) { //! set frames to skip (for time lapse)
   state->set_frames_skip=  d;
   if ((state->running==0) && (state->starting==0)) {
     state->frames_skip=      state->set_frames_skip;
//     state->frames_skip_left= state->set_frames_skip;
     state->frames_skip_left= 0;
   }
}


void  camogm_set_format(int d) {
//   int rslt=0;
   state->set_format=  d;
/* set format with the "start" command
   if ((state->running==0) && (state->starting==0)) {
     state->format=  state->set_format;
     switch (state->format) {
      case CAMOGM_FORMAT_NONE: rslt= 0; break;
      case CAMOGM_FORMAT_OGM:  rslt= camogm_init_ogm(); break;
      case CAMOGM_FORMAT_JPEG: rslt= camogm_init_jpeg();break;
      case CAMOGM_FORMAT_MOV:  rslt= camogm_init_mov(); break;
     }
     if (rslt) {
       D0(fprintf (debug_file,"%s:%d: Error setting format to=%d\n",__FILE__,__LINE__, state->format));
     }
     state->formats |= 1 << (state->format);
   }
*/
}

//! needed for Quicktime - maybe something else?
void  camogm_set_max_frames(int d) {
   state->set_max_frames=  d;
   if ((state->running==0) && (state->starting==0)) state->max_frames=  d;
}

void  camogm_set_frames_per_chunk(int d) {
   state->set_frames_per_chunk=  d;
   if ((state->running==0) && (state->starting==0)) state->frames_per_chunk=  d;
}

void  camogm_status(char * fn, int xml) {
  int _len=0;
  int _dur,_udur;
//TODO:make it XML file
  FILE* f;
  char *_state, *_output_format, *_using_exif, *_using_global_pointer, *_compressor_state;
  int _b_free, _b_used; // , save_p;
  int _frames_remain=0;
  int _sec_remain=0;
  int _frames_skip=0;
  int _sec_skip=0;
  char *_kml_enable, *_kml_used, *_kml_height_mode;

  _kml_enable=      state->kml_enable?"yes":"no";
  _kml_used=        state->kml_used?"yes":"no";
  _kml_height_mode= state->kml_height_mode?"GPS altitude":"map ground level";  //! 1 - actual, 0 - ground
  char *_audio_enable, *_allow_sync;
  _audio_enable=      state->set_audio_enable?"yes":"no";
  _allow_sync=      state->set_allow_sync?"yes":"no";


  _b_free=getGPValue(G_FREECIRCBUF);
  _b_used=getGPValue(G_CIRCBUFSIZE)-getGPValue(G_FREECIRCBUF);

  if (!fn) f=stdout;
  else if (strcmp(fn, "stdout")==0) f=stdout;
  else if (strcmp(fn, "stderr")==0) f=stderr;
  else {
    if (!((f=fopen (fn,"w")))) {
       D0(fprintf (debug_file,"Error opening %s\n", fn));
      return;
    }
  }
//  if      (state->vf)       _len=ftell(state->vf); //! for ogm
  if      (state->vf)       _len=lseek(state->vf, 0, SEEK_CUR); //! for ogm
  else if ((state->ivf)>=0) _len=lseek(state->ivf, 0, SEEK_CUR); //!for mov
  _dur= state->this_frame_params.timestamp_sec-state->frame_params.timestamp_sec;
  _udur=state->this_frame_params.timestamp_usec-state->frame_params.timestamp_usec;
  if (_udur<0) {
    _dur-=1;
    _udur+=1000000;
  } else if (_udur>=1000000) {
    _dur+=1;
    _udur-=1000000;
  }
  _state=         state->running? "running":(state->starting?"starting":"stopped");
  _output_format= state->format?((state->format==CAMOGM_FORMAT_OGM)?"ogm":
                                             ((state->format==CAMOGM_FORMAT_JPEG)?"jpeg":
                                             ((state->format==CAMOGM_FORMAT_MOV)?"mov":
                                              "other"))):"none";
  _using_exif=    state->exif?"yes":"no";
  _using_global_pointer=state->save_gp?"yes":"no";
  _compressor_state=(getGPValue(P_COMPRESSOR_RUN)==2)?"running":"stoppped";
    if ( state->frames_skip >0 ) {
      _frames_remain= state->frames_skip_left;
      _frames_skip=state->frames_skip;
    } else if ( state->frames_skip <0 ) {
      _sec_remain= (state->frames_skip_left - state->this_frame_params.timestamp_sec);
      _sec_skip=-(state->frames_skip);
    }


  if (xml) {
    fprintf (f,"<?xml version=\"1.0\"?>\n" \
             "<camogm_state>\n" \
             "  <state>\"%s\"</state>\n" \
             "  <compressor_state>\"%s\"</compressor_state>\n" \
             "  <file_name>\"%s\"</file_name>\n" \
             "  <frame_number>%d</frame_number>\n" \
             "  <file_duration>%d.%06d</file_duration>\n" \
             "  <file_length>%d</file_length>\n" \
             "  <frame_period>%d</frame_period>\n" \
             "  <frames_skip>%d</frames_skip>\n" \
             "  <seconds_skip>%d</seconds_skip>\n" \
             "  <frames_skip_left>%d</frames_skip_left>\n" \
             "  <seconds_skip_left>%d</seconds_skip_left>\n" \
             "  <frame_width>%d</frame_width>\n" \
             "  <frame_height>%d</frame_height>\n" \
             "  <format>\"%s\"</format>\n" \
             "  <exif>\"%s\"</exif>\n" \
             "  <prefix>\"%s\"</prefix>\n" \
             "  <max_duration>%d</max_duration>\n" \
             "  <max_length>%d</max_length>\n" \
             "  <max_frames>%d</max_frames>\n" \
             "  <timescale>%f</timescale>\n" \
             "  <frames_per_chunk>%d</frames_per_chunk>\n" \
             "  <buffer_overruns>%d</buffer_overruns>\n" \
             "  <buffer_minimal>%d</buffer_minimal>\n" \
             "  <buffer_free>%d</buffer_free>\n" \
             "  <buffer_used>%d</buffer_used>\n" \
             "  <circbuf_rp>%d</circbuf_rp>\n" \
             "  <debug_output>\"%s\"</debug_output>\n" \
             "  <debug_level>%d</debug_level>\n" \
             "  <use_global_rp>\"%s\"</use_global_rp>\n" \
             "  <kml_enable>\"%s\"</kml_enable>\n" \
             "  <kml_used>\"%s\"</kml_used>\n" \
             "  <kml_path>\"%s\"</kml_path>\n" \
             "  <kml_horHalfFov>\"%f\"</kml_horHalfFov>\n" \
             "  <kml_vertHalfFov>\"%f\"</kml_vertHalfFov>\n" \
             "  <kml_near>\"%f\"</kml_near>\n" \
             "  <kml_height_mode>\"%s\"</kml_height_mode>\n" \
             "  <kml_height>\"%f\"</kml_height>\n" \
             "  <kml_period>%d</kml_period>\n" \
             "  <kml_last_ts>%d.%06d</kml_last_ts>\n" \
             "  <audio_enable>\"%s\"</audio_enable>\n" \
             "  <audio_channels>\"%d\"</audio_channels>\n" \
             "  <audio_rate>\"%d\"</audio_rate>\n" \
             "  <audio_volume>\"%d\"</audio_volume>\n" \
             "  <allow_sync>\"%s\"</allow_sync>\n" \
             "</camogm_state>\n",
             _state,_compressor_state,state->path,state->frameno,_dur,_udur,_len,state->frame_period, \
             _frames_skip,_sec_skip,_frames_remain, _sec_remain, \
             state->width,state->height,_output_format,_using_exif, \
             state->path_prefix, state->segment_duration, state->segment_length, state->max_frames, state->timescale, \
             state->frames_per_chunk, state->buf_overruns, state->buf_min, _b_free, _b_used, state->cirbuf_rp, \
             state->debug_name, debug_level, _using_global_pointer, \
             _kml_enable,_kml_used,state->kml_path,state->kml_horHalfFov,state->kml_vertHalfFov,state->kml_near,\
             _kml_height_mode,state->kml_height,state->kml_period,state->kml_last_ts,state->kml_last_uts, \
             _audio_enable,state->set_audio_channels,state->set_audio_rate,state->audio_volume,_allow_sync);
  } else {
    fprintf (f,"state              %s\n",        _state);
    fprintf (f,"compressor state   %s\n",        _compressor_state);
    fprintf (f,"file               %s\n",        state->path);
    fprintf (f,"frame              %d\n",        state->frameno);
    fprintf (f,"file duration      %d.%06d sec\n",_dur,_udur);
    fprintf (f,"file length        %d B\n",      _len);
    fprintf (f,"frame period       %d (0x%x)\n", state->frame_period,state->frame_period);
    if ( _frames_skip >0 ) fprintf (f,"frames to skip   %d (left %d)\n",_frames_skip, _frames_remain);
    if ( _sec_skip    <0 ) fprintf (f,"timelapse period  %d sec (remaining %d sec)\n", _sec_skip, _sec_remain);
    fprintf (f,"width              %d (0x%x)\n", state->width,state->width);
    fprintf (f,"height             %d (0x%x)\n", state->height,state->height);
    fprintf (f,"\n");
    fprintf (f,"output format      %s\n",        _output_format);
    fprintf (f,"using exif         %s\n",        _using_exif);
    fprintf (f,"path prefix:       %s\n",        state->path_prefix);
    fprintf (f,"max file duration: %d sec\n",    state->segment_duration);
    fprintf (f,"max file length:   %d B\n",      state->segment_length);
    fprintf (f,"max frames         %d\n",        state->max_frames);
    fprintf (f,"timescale          %f\n",        state->timescale);
    fprintf (f,"frames per chunk   %d\n",        state->frames_per_chunk);
    fprintf (f,"\n");
    fprintf (f,"buffer overruns    %d\n",        state->buf_overruns);
    fprintf (f,"buffer minimal     %d\n",        state->buf_min);
    fprintf (f,"buffer free        %d\n",        _b_free);
    fprintf (f,"buffer used        %d\n",        _b_used);
    fprintf (f,"circbuf_rp         %d (0x%x)\n", state->cirbuf_rp,state->cirbuf_rp);
    fprintf (f,"\n");
    fprintf (f,"debug output to    %s\n",        state->debug_name);
    fprintf (f,"debug level        %d\n",        debug_level);
    fprintf (f,"use global pointer %s\n",        _using_global_pointer);
    fprintf (f,"\n\n");
    fprintf (f,"kml_enable         %s\n",        _kml_enable);
    fprintf (f,"kml_used           %s\n",        _kml_used);
    fprintf (f,"kml_path           %s\n",        state->kml_path);
    fprintf (f,"kml_horHalfFov     %f degrees\n",state->kml_horHalfFov);
    fprintf (f,"kml_vertHalfFov    %f degrees\n",state->kml_vertHalfFov);
    fprintf (f,"kml_near           %f m\n",      state->kml_near);
    fprintf (f,"kml height mode    %s\n",        _kml_height_mode);
    fprintf (f,"kml_height (extra) %f m\n",      state->kml_height);
    fprintf (f,"kml_period         %d\n",        state->kml_period);
    fprintf (f,"kml_last_ts        %d.%06d\n",   state->kml_last_ts,state->kml_last_uts);
    fprintf (f,"audio_enable       %s\n",        _audio_enable);
    fprintf (f,"audio_channels     %d\n",        state->set_audio_channels);
    fprintf (f,"audio_rate         %d\n",        state->set_audio_rate);
    fprintf (f,"allow_sync         %s\n",        _allow_sync);
    fprintf (f,"\n\n");

  }
  if ((f!=stdout) && (f!=stderr)) fclose (f);
  if (state->buf_overruns>=0) state->buf_overruns=0; //! resets overruns after reading status , so "overruns" means since last reading status
  state->buf_min=_b_free;
}

//! will read from pipe, return pointer to null terminated string if available, NULL otherwise
char * getLineFromPipe(FILE* npipe) {
      int fl;
      char * nlp;
//!remove used string if any
      if (cmdstrt > 0) {
//!moving overlapping strings
        memmove(cmdbuf, &cmdbuf[cmdstrt], sizeof(cmdbuf)- cmdstrt);
        cmdbufp-=cmdstrt;
        cmdstrt=0;
      }
//! is there any complete string in a buffer?
      if (!cmdbufp) cmdbuf[cmdbufp]=0; //!null-terminate first access (probably not needed for the static buffer
//      nlp= strchr(cmdbuf,'\n');
      nlp= strpbrk(cmdbuf,";\n");
      if (!nlp) { //!no complete string, try to read more
        fl=fread(&cmdbuf[cmdbufp], 1,sizeof(cmdbuf)-cmdbufp-1,npipe);
        cmdbuf[cmdbufp+fl]=0;
//! is there any complete string in a buffer after reading?
//        nlp= strchr(&cmdbuf[cmdbufp],'\n'); //! there were no new lines before cmdbufp
        nlp= strpbrk(&cmdbuf[cmdbufp],";\n"); //! there were no new lines before cmdbufp
        cmdbufp+=fl; //!advance pointer after pipe read
      }
      if (nlp) {
//printf ("++nlp=%d\n", (int) (nlp-cmdbuf));
        nlp[0]=0;
        cmdstrt=nlp-cmdbuf+1;
//printf ("++cmdstrt=%d\n", cmdstrt);
//printf ("cmdbuf[0]=%d, cmdbuf[1]=%d, cmdbuf[2]=%d, cmdbuf[3]=%d, \n",cmdbuf[0],cmdbuf[1],cmdbuf[2],cmdbuf[3]);
        for (fl=0; cmdbuf[fl] && strchr(" \t",cmdbuf[fl]); fl++);
//printf ("++fl=%d\n", fl);
        return &cmdbuf[fl];
      } else {
//printf ("notready: cmdbufp=%d, cmdstrt=%d\n",cmdbufp, cmdstrt);
          return NULL;
      }
}
// command[= \t]*args[ \t]*
int parse_cmd(FILE* npipe) {
  char * cmd;
  char * args;
  char * argse;
  int d;
  double dd;
//  if (!((cmd=getLineFromPipe(npipe)))) return 0; //! nothing in the pipe
//!skip empty commands
  while(((cmd=getLineFromPipe(npipe))) && !cmd[0]) ;
  if (!cmd) return 0; //! nothing in the pipe
//  printf ("cmd[0]=%d:%s\n",(int) cmd[0],cmd);
  args=strpbrk(cmd,"= \t");
//! is it just a single word command or does it have parameters?
  if (args) {
    args[0]=0;
    args++;
    while (strchr("= \t",args[0])) args++;
    if (args[0]) {
//! ltrim (args)
      for (argse=strchr(args,'\0')-1; strchr("= \t",argse[0]);argse--) argse[0]='\0';
    }
    if (!args[0]) args=NULL;
  }
//! now cmd is trimmed, arg is NULL or a pointer to trimmed command arguments
  if      (strcmp(cmd, "start")==0) {
     camogm_start(1);
     return 1;
  } else if (strcmp(cmd, "reset")==0) { //! will reset pointer to the last acquired frame (if any)
     camogm_reset();
     return 2;
  } else if (strcmp(cmd, "stop")==0) {
     camogm_stop(1);
     return 3;
  } else if (strcmp(cmd, "exit")==0) { 
     camogm_stop(1);
     camogm_free();
     exit (0);
  } else if (strcmp(cmd, "duration")==0) {
     if (!(args) || (((d= strtol(args, NULL, 10)))<=0)) d=DEFAULT_DURATION;
     camogm_set_segment_duration(d);
     return 4;
  } else if (strcmp(cmd, "length")==0) {
     if (!(args) || (((d= strtol(args, NULL, 10)))<=0)) d=DEFAULT_LENGTH;
     camogm_set_segment_length(d);
     return 5;
  } else if (strcmp(cmd, "prefix")==0) {
     if (args) camogm_set_prefix (args);
     return 6;
  } else if (strcmp(cmd, "status")==0) { 
     camogm_status(args, 0);
     return 7;
  } else if (strcmp(cmd, "xstatus")==0) { 
     camogm_status(args, 1);
     return 7;
  } else if (strcmp(cmd, "save_gp")==0) {
     if ((args) && (((d= strtol(args, NULL, 10)))>=0)) camogm_set_save_gp(d);
     return 8;
  } else if (strcmp(cmd, "exif")==0) {
     if ((args) && (((d= strtol(args, NULL, 10)))>=0)) camogm_set_exif(d);
     return 8;
  } else if (strcmp(cmd, "debug")==0) {
     camogm_debug(args);
     return 9;
  } else if (strcmp(cmd, "timescale")==0) {
     dd= strtod(args,NULL);
     camogm_set_timescale(dd?dd:1.0);
     return 10;
//!TODO: fix period calculation/check for frame skipping (just disable in frame skip mode?)
//!TODO: add time period (system clock), not just frame skipping


  } else if (strcmp(cmd, "frameskip")==0) {
     d= strtol(args, NULL, 10);
     camogm_set_frames_skip(d);
     return 11;
  } else if (strcmp(cmd, "timelapse")==0) { //! period (in seconds) between stored frames
     d= strtol(args, NULL, 10);
     camogm_set_frames_skip(-d);
     return 11;
  } else if (strcmp(cmd, "format")==0) {
     if (args) {
      if      (strcmp(args,  "none")==0) camogm_set_format(0);
      else if ((strcmp(args, "ogm" )==0) || (strcmp(args, "ogg")==0)) camogm_set_format(CAMOGM_FORMAT_OGM);
      else if ((strcmp(args, "jpeg")==0) || (strcmp(args, "jpg")==0)) camogm_set_format(CAMOGM_FORMAT_JPEG);
      else if (strcmp(args,  "mov" )==0)                              camogm_set_format(CAMOGM_FORMAT_MOV);
     }
     return 12;
  } else if (strcmp(cmd, "debuglev")==0) {
     d= strtol(args, NULL, 10);
     camogm_debug_level(d?d:0);
     return 13;
  } else if (strcmp(cmd, "kml")==0) {
     if ((args) && (((d= strtol(args, NULL, 10)))>=0)) camogm_kml_set_enable(d);
     return 14;
  } else if (strcmp(cmd, "kml_hhf")==0) {
     dd= strtod(args,NULL);
     camogm_kml_set_horHalfFov(dd);
     return 15;
  } else if (strcmp(cmd, "kml_vhf")==0) {
     dd= strtod(args,NULL);
     camogm_kml_set_vertHalfFov(dd);
     return 16;
  } else if (strcmp(cmd, "kml_near")==0) {
     dd= strtod(args,NULL);
     camogm_kml_set_near(dd);
     return 17;
  } else if (strcmp(cmd, "kml_alt")==0) {
     if (args) {
      if      (strcmp(args, "gps"   )==0) camogm_kml_set_height_mode(1);
      else if (strcmp(args, "ground")==0) camogm_kml_set_height_mode(0);
     }
     return 18;
  } else if (strcmp(cmd, "kml_height")==0) {
     dd= strtod(args,NULL);
     camogm_kml_set_height(dd);
     return 19;
  } else if (strcmp(cmd, "kml_period")==0) {
     d= strtol(args, NULL, 10);
     camogm_kml_set_period(d?d:1);
     return 20;
  } else if (strcmp(cmd, "audio")==0) {
     if (args) {
        camogm_set_audio_state(args);
     }
     return 21;
  } else if (strcmp(cmd, "audio_format")==0) {
     if (args) {
        camogm_set_audio_format(args);
     }
     return 22;
  } else if (strcmp(cmd, "audio_volume")==0) {
     if (args) {
        camogm_set_audio_volume(args);
     }
     return 23;
  } else if (strcmp(cmd, "allow_sync")==0) {
     if (args) {
        camogm_set_sync_state(args);
     }
     return 24;
  }

  return -1;
}

int main(int argc, char *argv[])
{
   const char circbufFileName[]="/dev/circbuf";
//     int fd_circ;
   FILE * cmd_file;
   const char usage[]=   "This program allows recording of the video/images acquired by Elphel camera to the storage media.\n" \
                         "It is designed to run in the background and accept commands through a named pipe.\n\n" \
                         "Usage:\n\n" \
                         "%s <named_pipe_name>\n\n" \
                         "i.e.:\n\n" \
                         "%s /var/state/camogm_cmd\n\n" \
                         "When the program is runninig you may send commands by writing strings to the command file\n" \
                         "(/var/state/camogm_cmd in the example above). The complete list of available commands is available\n" \
                         "on Elphel Wiki (http://wiki.elphel.com/index.php?title=Camogm), here is the example of usage\n" \
                         "from the shell prompt in the camera:\n\n" \
                         "echo \"status; exif=1; format=jpeg;status=/var/tmp/camogm.status\" > /var/state/camogm_cmd\n\n" \
                         "That will print status information on the standard output (may not be visible if the program was not\n" \
                         "started from the same session), set exif mode on (each frame will have the full Exif header including\n" \
                         "a precise time stamp), set output format to a series of individual JPEG files, and then send status\n" \
                         "information to a file /var/tmp/camogm.status in the camera file system.\n\n" \
                         "This program does not control the process of acquisition of the video/images to the camera internal\n" \
                         "buffer, it only retrieves that data from the buffer (waiting when needed), packages it to selected\n" \
                         "format and stores the result files.\n\n";
   int go=1;
   int cmd;
   int i,rslt;
   state= &sstate; //extern
//! no command line options processing yet
   if ((argc < 2) || (argv[1][1]=='-'))  { 
     printf (usage,argv[0],argv[0]);
     return 0;
   }
   camogm_init();

//! open Exif header file
   state->fd_exif = open(ExifFileName, O_RDONLY);
   if (state->fd_exif<0) { // check control OK
     D0(fprintf (debug_file,"Error opening %s\n", ExifFileName));
     return -1;
   }

//! open JPEG header file
   state->fd_head = open(HeadFileName, O_RDWR);
   if (state->fd_head<0) { // check control OK
     D0(fprintf (debug_file,"Error opening %s\n", HeadFileName));
     return -1;
   }
   state->head_size=lseek(state->fd_head,0,SEEK_END);
   if (state->head_size>JPEG_HEADER_MAXSIZE) {
     D0(fprintf (debug_file,"%s:%d: Too big JPEG header (%d > %d)",__FILE__,__LINE__,state->head_size, JPEG_HEADER_MAXSIZE ));
     return -2;
   }

//! open circbuf and mmap it (once at startup)
   state->fd_circ = open(circbufFileName, O_RDWR);
   if (state->fd_circ<0) { // check control OK
      D0(fprintf (debug_file,"Error opening %s\n", circbufFileName));
      return -2;
   }
/*! find total buffer length (it is in defines, actually in c313a.h */
   state->circ_buff_size=lseek(state->fd_circ,0,SEEK_END);
   ccam_dma_buf = (unsigned long *) mmap(0, state->circ_buff_size, PROT_READ, MAP_SHARED, state->fd_circ, 0);
   if((int)ccam_dma_buf == -1) {
     D0(fprintf (debug_file,"Error in mmap of %s\n",circbufFileName));
//     close (fd_head);
     close(state->fd_circ);
     return -3;
   }

//! Now open/mmap file to read sensor/compressor parameters (currently - just free memory in circbuf and compressor state)

//! open circbuf and mmap it (once at startup)
    state->fd_fparmsall = open(ctlFileName, O_RDWR);
    if (state->fd_fparmsall<0) { // check control OK
       D0(fprintf (debug_file,"%s:%d:%s: Error opening %s\n",__FILE__,__LINE__,__FUNCTION__, ctlFileName));
       return -2;
    }

//! now try to mmap
    frameParsAll = (struct framepars_all_t *) mmap(0, sizeof (struct framepars_all_t) , PROT_READ, MAP_SHARED, state->fd_fparmsall, 0);
    if((int)frameParsAll == -1) {
      D0(fprintf(debug_file,"%s:%d:%s: Error in mmap in %s\n",__FILE__,__LINE__,__FUNCTION__,ctlFileName));
      close(state->fd_fparmsall);
      close(state->fd_circ);
      return -3;
    }
    framePars=frameParsAll->framePars;
    globalPars=frameParsAll->globalPars;

//!create a named pipe
//!always delete the pipe if it existed, start a fresh one
   i=unlink (argv[1]);
   if (i) {
       D1(fprintf (debug_file,"Unlink %s returned %d, errno=%d \n", argv[1], i, errno));
   }
   i=mkfifo(argv[1], 0777); //EEXIST
//! now should not exist
   if (i) {
    if (errno==EEXIST) {
       D1(fprintf (debug_file,"Named pipe %s already exists, will use it.\n", argv[1]));
    } else {
       D0(fprintf (debug_file,"Can not create a named pipe %s, errno=%d \n", argv[1], errno));
       return -4;
    }
   }

//!now open the pipe - will block until something will be written (or just open for writing
//!Reads themselves will not block
   if (!((cmd_file=fopen(argv[1],"r")))) {
     D0(fprintf (debug_file,"Can not open command file %s\n",argv[1]));
     return -5;
   }
   D1(fprintf (debug_file,"Pipe %s open for reading\n",argv[1]));

//! Here is a main loop. If recording is on, it will check for commands after each frame, if it is off - poll with fixed usleep
#define COMMAND_LOOP_DELAY 500000 //0.5sec
 
	while (go) {
//   D3(fprintf (debug_file,"%s:%d: format=%d, set_format=%d\n",__FILE__,__LINE__, state->format, state->set_format));

//! look at command queue first
		cmd=parse_cmd(cmd_file);
		if (cmd) {
			if (cmd<0) D0(fprintf (debug_file,"Unrecognized command\n"));
		} else {
			if(state->running) { //!no commands in queue, started 
//   D3(fprintf (debug_file,"%s:%d: format=%d, set_format=%d\n",__FILE__,__LINE__, state->format, state->set_format));
				switch (rslt=-sendImageFrame()) {
				case 0:
					audio_process();
/*
          D3(fprintf (debug_file,"%s:line %d - sendImageFrame() returned %d\n" \
                                  "state->cirbuf_rp= 0x%x\n",__FILE__,__LINE__,rslt,state->cirbuf_rp));
*/
					break; //! frame sent OK, nothing to do (TODO: check file length/duration)
				case CAMOGM_FRAME_NOT_READY:        //!  just wait for the frame to appear at the current pointer
//! we'll wait for a frame, not to waste resources. But if the compressor is stopped this program will not respond to any commands
//! TODO - add another wait with (short) timeout?
//				D3(fprintf (debug_file,"%s:line %d - sendImageFrame() returned -%d\n",__FILE__,__LINE__,rslt));
/*
     D3(fprintf (debug_file,"%s:line %d - sendImageFrame() returned -%d\n" \
                                  "state->cirbuf_rp= 0x%x\n",__FILE__,__LINE__,rslt,state->cirbuf_rp));
*/
					lseek(state->fd_circ,LSEEK_CIRC_WAIT,SEEK_END);
					break;
				case CAMOGM_FRAME_CHANGED:   //! frame parameters have changed
				case CAMOGM_FRAME_NEXTFILE:  //! next file needed (need to switch to a new file (time/size exceeded limit)
				case CAMOGM_FRAME_INVALID:   //! invalid frame pointer
				case CAMOGM_FRAME_BROKEN:    //! frame broken (buffer overrun)
//! restart the file
//     D3(fprintf (debug_file,"%s:line %d - sendImageFrame() returned -%d\n",__FILE__,__LINE__,rslt));
					camogm_stop(0);
					camogm_start(0);
					break;
				case CAMOGM_FRAME_FILE_ERR:  //! error with file I/O
				case CAMOGM_FRAME_OTHER:     //! other errors
					D0(fprintf (debug_file,"%s:line %d - error=%d\n",__FILE__,__LINE__,rslt));
					break;
				default:
					D0(fprintf (debug_file,"%s:line %d - should not get here (rslt=%d)\n",__FILE__,__LINE__,rslt));
					exit (-1);
				} //switch
			} else {
				if (state->starting) { //!no commands in queue,starting (but not started yet)
//   D3(fprintf (debug_file,"%s:%d: format=%d, set_format=%d\n",__FILE__,__LINE__, state->format, state->set_format));

//!retry starting
					switch ((rslt=-camogm_start(1))) {
					case 0:
						break; //! file started OK, nothing to do
					case CAMOGM_FRAME_NOT_READY:        //!  just wait for the frame to appear at the current pointer
//! we'll wait for a frame, not to waste resources. But if the compressor is stopped this program will not respond to any commands
//! TODO - add another wait with (short) timeout?
//                lseek(state->fd_circ,LSEEK_CIRC_WAIT,SEEK_END);
					case CAMOGM_FRAME_CHANGED:   //! frame parameters have changed
					case CAMOGM_FRAME_NEXTFILE:
					case CAMOGM_FRAME_INVALID:   //! invalid frame pointer
					case CAMOGM_FRAME_BROKEN:    //! frame broken (buffer overrun)
//     D3(fprintf (debug_file,"%s:line %d - camogm_start() returned -%d,  state->cirbuf_rp= 0x%x\n", __FILE__,__LINE__,rslt,state->cirbuf_rp));
						usleep( COMMAND_LOOP_DELAY) ; //! it should be not too long so empty buffer will not be overrun
						break;
					case CAMOGM_FRAME_FILE_ERR:  //! error with file I/O
					case CAMOGM_FRAME_OTHER:     //! other errors
						D0(fprintf (debug_file,"%s:line %d - error=%d\n",__FILE__,__LINE__,rslt));
						break;
					default:
						D0(fprintf (debug_file,"%s:line %d - should not get here (rslt=%d)\n",__FILE__,__LINE__,rslt));
						exit (-1);
					} //switch
				} else { //! not running, not starting
					usleep( COMMAND_LOOP_DELAY) ; //! make it longer but interruptible by signals?
				}
			}
		} 
	}
	return 0;
}
/**
 * @brief Read either G_* parameter (these are 'regular' values defined by number) or P_* parameter
 *         (it can be read for up to 6 frames ahead, but current interface only allows to read last/current value)
 * @param GPNumber parameter number (as defined in c313a.h), G_* parameters have numbers above FRAMEPAR_GLOBALS, P_* - below)
 * @return parameter value
 */
unsigned long getGPValue(unsigned long GPNumber) {
   return (GPNumber>=FRAMEPAR_GLOBALS)?
              GLOBALPARS(GPNumber):
              framePars[GLOBALPARS(G_THIS_FRAME) & PARS_FRAMES_MASK].pars[GPNumber];
}

/**
 * @brief Set value of the specified global (G_*) parameter
 * @param GNumber - parameter number (as defined in c313a.h)
 * @param value  - value to set
 */
void setGValue(unsigned long  GNumber,   unsigned long value) {
    GLOBALPARS(GNumber)=value;
}

/**
 * @brief check if this application is enabled (by appropriate bit in P_DAEMON_EN), if not - 
 * and wait until enabled (return false when enabled)
 * @param daemonBit - bit number to accept control in P_DAEMON_EN parameter
 * @return (after possible waiting) true if there was no waiting, false if there was waiting
 */
int  waitDaemonEnabled(int daemonBit) { // <0 - use default
   if ((daemonBit>=0) && (daemonBit<32)) lastDaemonBit=daemonBit;
   unsigned long this_frame=this_frame=GLOBALPARS(G_THIS_FRAME);
/// No semaphors, so it is possible to miss event and wait until the streamer will be re-enabled before sending message,
/// but it seems not so terrible
   lseek(state->fd_circ, LSEEK_DAEMON_CIRCBUF+lastDaemonBit, SEEK_END); /// 
   if (this_frame==GLOBALPARS(G_THIS_FRAME)) return 1;
   return 0;
}

/**
 * @brief check if this application is enabled (by appropriate bit in P_DAEMON_EN)
 * @param daemonBit - bit number to accept control in P_DAEMON_EN parameter
 * @return (after possible waiting) true if there was no waiting, false if there was waiting
 */
int  isDaemonEnabled(int daemonBit) { // <0 - use default
   if ((daemonBit>=0) && (daemonBit<32)) lastDaemonBit=daemonBit;
   return ((framePars[GLOBALPARS(G_THIS_FRAME) & PARS_FRAMES_MASK].pars[P_DAEMON_EN] & (1 <<lastDaemonBit))!=0);
}

/*******************************************************************************
 *
 * audio
 *
 */

void camogm_set_audio_state(char *args) {
	if(strcmp(args, "on") == 0 || strcmp(args, "enable") == 0) {
		state->set_audio_enable = 1;
	}
	if(strcmp(args, "off") == 0 || strcmp(args, "disable") == 0) {
		state->set_audio_enable = 0;
	}
DA(	fprintf(stderr, "request for audio state: %s\n", args););
}

void camogm_set_sync_state(char *args) {
	if(strcmp(args, "on") == 0 || strcmp(args, "enable") == 0) {
		state->set_allow_sync = 1;
	}
	if(strcmp(args, "off") == 0 || strcmp(args, "disable") == 0) {
		state->set_allow_sync = 0;
	}
}

void camogm_set_audio_volume(char *args) {
	int vol = atoi(args);
	if(vol > 65535)
		vol = 65535;
	if(vol < 0)
		vol = 0;
	state->set_audio_volume = vol;
}

void camogm_set_audio_format(char *args) {
	char buf[8];
	int i;
	int channels = -1;
	int rate = -1;
	char *right = strstr(args, "/");
	if(right != NULL) {
		channels = atol(&right[1]);
		if(channels < AUDIO_CHANNELS_MIN)
			channels = AUDIO_CHANNELS_MIN;
		if(channels > AUDIO_CHANNELS_MAX)
			channels = AUDIO_CHANNELS_MAX;
		state->set_audio_channels = channels;
	}
	for(i = 0; i < 7; i++) {
		if(args[i] != '\0' && args[i] != '/')
			buf[i] = args[i];
		else
			break;
	}
	buf[i] = '\0';
	if(buf[0] != '\0') {
		rate = atol(buf);
		if(rate < AUDIO_RATE_MIN)
			rate = AUDIO_RATE_MIN;
		if(rate > AUDIO_RATE_MAX)
			rate = AUDIO_RATE_MAX;
		state->set_audio_rate = rate;
	}

DA(	fprintf(stderr, "request for audio format: %s; after apply: rate == %d; channels == %d\n", args, state->set_audio_rate, state->set_audio_channels););
}

void camogm_audio_initialize(void) {
	ctx_a.audio_present = 0;
	ctx_a.audio_set = 0;
	ctx_a.capture_handle = NULL;
	ctx_a.sbuffer = NULL;
	ctx_a.sample_time = 200; // ms
}

void audio_process(void) {
	if(ctx_a.audio_present == 0)
		return;
	if(state->format == CAMOGM_FORMAT_JPEG)
		return;

	int counter = 0;
	void *_buf;
	int _buf_len;
	struct timeval tv_sys;
	if(state->format == CAMOGM_FORMAT_OGM || state->format == CAMOGM_FORMAT_MOV) {
		int slen;
//		int _len = 0;
		snd_pcm_status_t *status;
		snd_pcm_status_alloca(&status);
		snd_timestamp_t ts;
		for(;;) {
			counter++;
			long avail = 0;
//			avail = snd_pcm_avail_update(ctx_a.capture_handle);
			gettimeofday(&tv_sys, NULL);
			tv_sys.tv_usec += ctx_a.sample_time;
			time_normalize(&tv_sys);
			snd_pcm_status(ctx_a.capture_handle, status);
			snd_pcm_status_get_tstamp(status, &ts);
			avail = snd_pcm_status_get_avail(status);
//fprintf(stderr, "avail == %d\n", avail);
//			avail = snd_pcm_status_get_avail_max(status);
//			int av = snd_pcm_status_get_avail(status);
			int to_push_flag = 0;
			int to_read = ctx_a.sbuffer_len; // length in the samples...
			if(ctx_a.to_finish_samples < 0)
				ctx_a.to_finish_samples = 0;
			if(avail >= ctx_a.sbuffer_len && ctx_a.to_finish_samples == 0)
				to_push_flag = 1;
			if(ctx_a.to_finish_samples > 0) {
				if(ctx_a.to_finish_samples > ctx_a.sbuffer_len) {
					if(avail >= ctx_a.sbuffer_len) {
						to_read = ctx_a.sbuffer_len;
						ctx_a.to_finish_samples -= ctx_a.sbuffer_len;
						to_push_flag = 2;
					}
				} else {
					if(avail >= ctx_a.to_finish_samples) {
						to_read = ctx_a.to_finish_samples;
						ctx_a.to_finish_samples = 0;
						to_push_flag = 2;
					}
				}
			}
			if(to_push_flag) {
				slen = snd_pcm_readi(ctx_a.capture_handle, (void *)(ctx_a.sbuffer + AUDIO_SBUFFER_PREFIX), to_read);
/*
				slen = snd_pcm_mmap_readi(ctx_a.capture_handle, (void *)(ctx_a.sbuffer + AUDIO_SBUFFER_PREFIX), to_read);
*/
//				snd_pcm_status(ctx_a.capture_handle, status);
//				snd_pcm_status_get_tstamp(status, &ts);
				if(slen > 0) {
					int flag = 1;
					long offset = 0;
					// check the length of the movie and sound track
					if(to_push_flag == 1) {
						struct timeval sl = ctx_a.time_last;
						sl.tv_usec += ctx_a.sample_time;
						time_normalize(&sl);
						struct timeval m_end;
						m_end = state->ts_video;
						m_end.tv_usec += state->frame_period / 2;
						time_normalize(&m_end);
						struct timeval m_len;
						if(m_len.tv_sec != ctx_a.movie_start.tv_sec) {
							m_len.tv_sec = m_end.tv_sec - 1;
							m_len.tv_usec = m_end.tv_usec + 1000000;
						}
						m_len.tv_sec -= ctx_a.movie_start.tv_sec;
						m_len.tv_usec -= ctx_a.movie_start.tv_usec;
						time_normalize(&m_len);
						if(time_comp(&sl, &m_len) > 0) {
							// sound too early - skip this sequence
							break;
						}
					}
					// we need to skip some samples in a new session
					// but if we just will switch the frames - we need to split new samples in the buffer for two parts - for the previous file,
					// and the next one...
					// so we can just save in the first file new data, and for the next - use "skip_samples" field
					if(ctx_a.audio_skip_samples != 0) {
DA(fprintf(stderr, "skip_samples == %d\n", ctx_a.audio_skip_samples, slen););
						if(ctx_a.audio_skip_samples >= slen) {
							ctx_a.audio_skip_samples -= slen;
							flag = 0;
						} else {
							offset = ctx_a.audio_skip_samples;
							ctx_a.audio_skip_samples = 0;
						}
					} 
					if(flag) {
						long samples = slen - offset;
						ctx_a.audio_count += samples;
						_buf = (void *)(ctx_a.sbuffer + AUDIO_SBUFFER_PREFIX);
						_buf = (void *)((char *)_buf + offset * 2 * state->audio_channels);
						_buf_len = samples * 2 * state->audio_channels;
						if(state->format == CAMOGM_FORMAT_OGM)
							camogm_audio_ogm_push(_buf, _buf_len, samples, AUDIO_SBUFFER_PREFIX);
						if(state->format == CAMOGM_FORMAT_MOV)
							camogm_audio_mov(_buf, _buf_len, samples);
						float tr = 1.0 / state->audio_rate;
						float l = tr * ctx_a.audio_count;
						unsigned long s = (unsigned long)l;
						l -= s;
						l *= 1000000;
						unsigned long us = (unsigned long)l;
						ctx_a.time_last.tv_sec = s;
						ctx_a.time_last.tv_usec = us;
DA(						fprintf(stderr, "%d: sound time %d:%06d, at %d:%06d; and the sound time is: %d:%06d, samples: %d\n", counter, s, us, tv_sys.tv_sec, tv_sys.tv_usec, ctx_a.time_last.tv_sec, ctx_a.time_last.tv_usec, samples););
					}
				}
			} else {
				break;
			}
		}
	}
}

void audio_init(by_command) {
	// apply settings - i.e. just copy to current
	state->audio_enable = state->set_audio_enable;
	state->audio_rate = state->set_audio_rate;
	state->audio_channels = state->set_audio_channels;
	state->audio_volume = state->set_audio_volume;
	state->allow_sync = state->set_allow_sync;

//	ctx_a.audio_count = 0;
//	ctx_a.audio_skip_samples = 0;

//	int slen;
	snd_timestamp_t audio_ts;
	if(state->audio_enable == 0) {
		return;
	}
	if(state->format == CAMOGM_FORMAT_JPEG)
		return;
	// set up audio device
	if(by_command && !ctx_a.audio_set) {
		snd_pcm_hw_params_t *hw_params;
		ctx_a.audio_set = 1;
//		sbuffer_len = state->audio_rate * 2 * state->audio_channels * sample_time;
//		ctx_a.sbuffer_len = state->audio_rate * state->audio_channels * ctx_a.sample_time;
		ctx_a.sbuffer_len = state->audio_rate * ctx_a.sample_time;
		ctx_a.sbuffer_len /= 1000;
		ctx_a.sbuffer_len -= ctx_a.sbuffer_len % 2;

DA(		fprintf(stderr, "sbuffer_len == %d\n", ctx_a.sbuffer_len););

//		state->audio_format = SND_PCM_FORMAT_S16_LE;
		int audio_format = SND_PCM_FORMAT_S16_LE;

		int err;
		ctx_a.sbuffer = (void *)malloc(ctx_a.sbuffer_len * 8 + AUDIO_SBUFFER_PREFIX);
//		if((err = snd_pcm_open(&ctx_a.capture_handle, "default", SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK)) < 0)
		if((err = snd_pcm_open(&ctx_a.capture_handle, "default", SND_PCM_STREAM_CAPTURE, 0)) < 0)
			return;
		snd_pcm_hw_params_alloca(&hw_params);
//		if((err = snd_pcm_hw_params_malloc(&hw_params)) < 0)
//			return;
		if((err = snd_pcm_hw_params_any(ctx_a.capture_handle, hw_params)) < 0)
			return;
		if((err = snd_pcm_hw_params_set_access(ctx_a.capture_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0)
			return;
/*
		if((err = snd_pcm_hw_params_set_access(ctx_a.capture_handle, hw_params, SND_PCM_ACCESS_MMAP_NONINTERLEAVED)) < 0)
			return;
*/
/*
		snd_pcm_access_mask_t *mask = alloca(snd_pcm_access_mask_sizeof());
		snd_pcm_access_mask_none(mask);
		snd_pcm_access_mask_set(mask, SND_PCM_ACCESS_MMAP_NONINTERLEAVED);
		if((err = snd_pcm_hw_params_set_access_mask(ctx_a.capture_handle, hw_params, mask)) < 0)
			return;
*/
//		if((err = snd_pcm_hw_params_set_format(ctx_a.capture_handle, hw_params, state->audio_format)) < 0)
		if((err = snd_pcm_hw_params_set_format(ctx_a.capture_handle, hw_params, audio_format)) < 0)
			return;
		unsigned int t = state->audio_rate;
		if((err = snd_pcm_hw_params_set_rate_near(ctx_a.capture_handle, hw_params, &t, 0)) < 0)
			return;
		state->audio_rate = t;
		if((err = snd_pcm_hw_params_set_channels(ctx_a.capture_handle, hw_params, state->audio_channels)) < 0)
			return;
		unsigned int period_time = 40 * 1000;
		if((err = snd_pcm_hw_params_set_period_time_near(ctx_a.capture_handle, hw_params, &period_time, 0)) < 0)
			return;
		unsigned int buffer_time = 2000000;
		if((err = snd_pcm_hw_params_set_buffer_time_near(ctx_a.capture_handle, hw_params, &buffer_time, 0)) < 0)
			return;
//		fprintf(stderr, "buffer_time is == %d\n", buffer_time);
		if((err = snd_pcm_hw_params(ctx_a.capture_handle, hw_params)) < 0)
			return;

/*
		snd_pcm_sw_params_t *sw_params;
		snd_pcm_sw_params_malloc(&sw_params);
		err = snd_pcm_sw_params_current(ctx_a.capture_handle, sw_params);
		if(err < 0)	return;
		err = snd_pcm_sw_params_set_tstamp_mode(ctx_a.capture_handle, sw_params, SND_PCM_TSTAMP_MMAP);
		if(err < 0)	return;
		err = snd_pcm_sw_params(ctx_a.capture_handle, sw_params);
		if(err < 0)	return;
		snd_pcm_sw_params_free(sw_params);
*/
		
//		snd_pcm_hw_params_free(hw_params);
//		if((err = snd_pcm_prepare(ctx_a.capture_handle)) < 0)
//			return;

		ctx_a.audio_present = 1;
		ctx_a.begin_of_stream_with_audio = 1;
		ctx_a.begin_of_stream_with_audio_trigger = 1;

		ctx_a.audio_count = 0;
		ctx_a.audio_skip_samples = 0;

		unsigned long rtc[4];
		audio_get_sys_fpga_times(rtc);
/*
		int fd_stream = open("/dev/ext353", O_RDWR);
		unsigned long rtc[4];
		ioctl(fd_stream, EXT_GET_TIME_SYNC, (void *)&rtc[0]);
		close(fd_stream);
*/
DA(		fprintf(stderr, "SYNC; sys time == %d:%06d; ", rtc[0], rtc[1]););
DA(		fprintf(stderr, "FPGA time == %d:%06d\n", rtc[2], rtc[3]););

DA(		fprintf(stderr, "audio_init(%d)... OK\n", by_command););
		snd_pcm_status_t *status;
		snd_pcm_status_alloca(&status);
/*
		while((slen = snd_pcm_mmap_readi(ctx_a.capture_handle, (void *)(ctx_a.sbuffer), 2)) < 0)
			snd_pcm_prepare(ctx_a.capture_handle);
*/
		snd_pcm_prepare(ctx_a.capture_handle);
		snd_pcm_reset(ctx_a.capture_handle);
		audio_set_volume(state->audio_volume);
		snd_pcm_readi(ctx_a.capture_handle, (void *)ctx_a.sbuffer, 8);

		snd_pcm_status(ctx_a.capture_handle, status);
		snd_pcm_status_get_tstamp(status, &audio_ts);
//DA(		fprintf(stderr, "audio: device is open...   !!!!!!!!!!!!!!!!!!!; slen == %d\n", slen););

		struct timeval d; // SYS time - FPGA time;
		d.tv_sec = rtc[0] - 1;
		d.tv_usec = rtc[1] + 1000000;
		d.tv_sec -= rtc[2];
		d.tv_usec -= rtc[3];
		time_normalize(&d);
		struct timeval tv;
		tv = audio_ts;
		tv.tv_sec -= 1;
		tv.tv_usec += 1000000;
		tv.tv_sec -= d.tv_sec;
		tv.tv_usec -= d.tv_usec;
		time_normalize(&tv);
		state->ts_audio = tv;

		gettimeofday(&tv, NULL);
DA(		fprintf(stderr, "audio_init(%d) - at %d:%06d", by_command, tv.tv_sec, (unsigned)tv.tv_usec););
DA(		fprintf(stderr, "status at %d:%06d\n", audio_ts.tv_sec, audio_ts.tv_usec););
	}
}

void audio_set_volume(int nvolume) {
	snd_mixer_t *mixer;
	snd_mixer_elem_t *elem;
	snd_mixer_selem_id_t *sid;
	// allocated at stack, so can't free it!
	snd_mixer_selem_id_alloca(&sid);

	snd_mixer_open(&mixer, 0);
	snd_mixer_attach(mixer, "default");
	snd_mixer_selem_register(mixer, NULL, NULL);
	snd_mixer_load(mixer);
			
	for(elem = snd_mixer_first_elem(mixer); elem; elem = snd_mixer_elem_next(elem)) {
		snd_mixer_selem_get_id(elem, sid);
		if(!snd_mixer_selem_is_active(elem))
			continue;
		// set volume at percents for capture elements
		snd_mixer_elem_t *selem = snd_mixer_find_selem(mixer, sid);
		if(selem == NULL) {
			break;
		}
		long volume_min = 0;
		long volume_max = 0;
		if(snd_mixer_selem_get_capture_volume_range(selem, &volume_min, &volume_max) == 0) {
			// set volume only for capture
			if(nvolume > 65535)
				nvolume = 65535;
			if(nvolume < 0)
				nvolume = 0;
			long long vol_new = volume_max;
			vol_new *= nvolume;
			vol_new /= 65535;
			long vol = 0;
			snd_mixer_selem_get_capture_volume(selem, SND_MIXER_SCHN_FRONT_LEFT, &vol);
			snd_mixer_selem_set_capture_volume_all(selem, vol_new);
DA(			fprintf(stderr, "element %s - OLD min vol == %d; max vol == %d; volume == %d\n", snd_mixer_selem_id_get_name(sid), volume_min, volume_max, vol););
DA(			snd_mixer_selem_get_capture_volume(selem, SND_MIXER_SCHN_FRONT_LEFT, &vol););
DA(			fprintf(stderr, "element %s - NEW min vol == %d; max vol == %d; volume == %d\n", snd_mixer_selem_id_get_name(sid), volume_min, volume_max, vol););
		}
	}
	snd_mixer_close(mixer);
}

void audio_start(int by_command) {
	state->audio_frameno = 0;
	state->audio_samples = 0;
	ctx_a.to_finish_samples = 0;
	ctx_a.time_last.tv_sec = 0;
	ctx_a.time_last.tv_usec = 0;
	ctx_a.audio_count = 0;
	if(state->format == CAMOGM_FORMAT_JPEG)
		return;
/*
	if(state->format == CAMOGM_FORMAT_JPEG) {
		if(ctx_a.audio_present) {
			char fname[256];
			sprintf(fname, "%s_01.wav", state->path_prefix);
			// open WAV file...
			fprintf(stderr, "open WAV file, name == \"%s\"\n", fname);
			wave_begin(fname);
		}
	}
*/
	struct timeval tv;
	gettimeofday(&tv, NULL);
DA(	fprintf(stderr, "audio_start(%d) - at %d:%06d\n", by_command, tv.tv_sec, (unsigned)tv.tv_usec););
}

void audio_end(int by_command) {
	if(ctx_a.audio_present) {
		if(state->format == CAMOGM_FORMAT_JPEG)
			return;
/*
		if(state->format == CAMOGM_FORMAT_JPEG) {
			// close WAV file...
			fprintf(stderr, "close WAV file\n");
			wave_end();
		}
*/
	}
	// set down audio device
	if(by_command && ctx_a.audio_set) {
		ctx_a.audio_present = 0;
		ctx_a.audio_set = 0;
		snd_pcm_drop(ctx_a.capture_handle);
		snd_pcm_close(ctx_a.capture_handle);
DA(		fprintf(stderr, "audio: device is closed... !!!!!!!!!!!!!!!!!!!\n"););
		free((void *)ctx_a.sbuffer);
		ctx_a.sbuffer = NULL;
	}
//	fprintf(stderr, "audio_end(%d)\n", by_command);
	struct timeval tv;
	gettimeofday(&tv, NULL);
DA(	fprintf(stderr, "audio_end(%d) - at %d:%06d\n", by_command, tv.tv_sec, (unsigned)tv.tv_usec););
}

void audio_free(void) {
//	fprintf(stderr, "audio_free()\n");
}

void audio_finish(int by_command) {
	// finalize audio stream to file here
	// TODO - check the time correctly.
DA(	fprintf(stderr, "here must be last push of the audio to the file\n"););
	struct timeval m_end;
DA(	fprintf(stderr, "  movie start at: %d:%06d\n", ctx_a.movie_start.tv_sec, ctx_a.movie_start.tv_usec););
	m_end = state->ts_video;
	m_end.tv_usec += state->frame_period / 2;
	time_normalize(&m_end);
DA(	fprintf(stderr, "    movie end at: %d:%06d\n", m_end.tv_sec, m_end.tv_usec););

	struct timeval m_len;
	if(m_len.tv_sec != ctx_a.movie_start.tv_sec) {
		m_len.tv_sec = m_end.tv_sec - 1;
		m_len.tv_usec = m_end.tv_usec + 1000000;
	}
	m_len.tv_sec -= ctx_a.movie_start.tv_sec;
	m_len.tv_usec -= ctx_a.movie_start.tv_usec;
	time_normalize(&m_len);
DA(	fprintf(stderr, "    movie length: %d:%06d\n", m_len.tv_sec, m_len.tv_usec););
	ctx_a.movie_start = m_end;
	// calculate how many samples we need to save now for the end
	unsigned long rtc[4];
	audio_get_sys_fpga_times(rtc);
/*
	int fd_stream = open("/dev/ext353", O_RDWR);
	unsigned long rtc[4];
	ioctl(fd_stream, EXT_GET_TIME_SYNC, (void *)&rtc[0]);
	close(fd_stream);
*/
DA(	fprintf(stderr, "_________________ END ____________________\n"););
DA(	fprintf(stderr, "       sys time == %d:%06d\n", rtc[0], rtc[1]););
DA(	fprintf(stderr, "      FPGA time == %d:%06d\n", rtc[2], rtc[3]););
DA(	fprintf(stderr, "AUDIO  sys time == %d:%06d\n", ctx_a.time_last.tv_sec, ctx_a.time_last.tv_usec););
//	int flag = 0;
	struct timeval audio_last;
	audio_last = ctx_a.time_last;
	if(m_len.tv_sec > audio_last.tv_sec) {
		m_len.tv_sec--;
		m_len.tv_usec += 1000000;
	}
	m_len.tv_sec -= audio_last.tv_sec;
	m_len.tv_usec -= audio_last.tv_usec;
	time_normalize(&m_len);
	long to_finish_us = m_len.tv_usec + 1000000 * m_len.tv_sec;
DA(	fprintf(stderr, "... and now we need to save audio for this time: %d:%06d - i.e. %06d usecs\n", m_len.tv_sec, m_len.tv_usec, to_finish_us););
	double s = state->audio_rate;
	s /= 1000.0;
	s *= to_finish_us;
	s /= 1000.0;
	ctx_a.to_finish_samples = (long)s;
	// from the state->tv_video_start to ctx_a.time_last (with FPGA time recalculation)
	do {
		audio_process();
		if(ctx_a.to_finish_samples > 0)
			sched_yield();
	} while(ctx_a.to_finish_samples > 0);
	audio_end(by_command);
}

int GPWrite(unsigned long *data, int length) {
	return write(state->fd_fparmsall, (void *)data, length);
}

//! return system (rtc[0]:rtc[1]) and FPGA (rtc[2]:rtc[3]) time
void audio_get_sys_fpga_times(unsigned long *rtc) {
	unsigned long write_data[6];
	write_data[0] = FRAMEPARS_GETFPGATIME;
	write_data[1] = 0;
	GPWrite(write_data, sizeof(unsigned long) * 2);
	rtc[2] = getGPValue(G_SECONDS);
	rtc[3] = getGPValue(G_MICROSECONDS);

	struct timeval tv_sys;
	gettimeofday(&tv_sys, NULL);
	rtc[0] = tv_sys.tv_sec;
	rtc[1] = tv_sys.tv_usec;
}
