#!/bin/sh
echo "********************************************************" ;
echo "*  Applying patches to Xilinx unisims library to make  *" ;
echo "*  it work with iverilog simulator.                    *" ;
echo "*  NOTE: Only files used in the design are modified,   *" ;
echo "*           more will likely be added later.           *" ;
echo "*  WARNING: As Xilinx can modify the files at any time,*" ;
echo "*           patchs may get broken - please contact     *" ;
echo "*           Elphel support (support@elphel.com) in     *" ;
echo "*           case.                                      *" ;
echo "*           Tested with WebPack 10.1.03,               *" ;
echo "*  UPDATE:                                             *" ;
echo "********************************************************" ;
echo
echo "Applying patch to DCM.v to fix iverilog errors:"
echo "../unisims/DCM.v:762: sorry: non-constant delays not supported here: (period)/('sd2)";
echo "../unisims/DCM.v:381: syntax error";
echo "../unisims/DCM.v:381: error: Invalid module instantiation";
echo "../unisims/DCM.v:1049: syntax error";
echo "../unisims/DCM.v:1049: error: invalid module item. Did you forget an initial or always?";
echo "../unisims/DCM.v:1322: syntax error";
echo "../unisims/DCM.v:1322: error: invalid module item. Did you forget an initial or always?";

patch -N unisims/DCM.v << 'EOP'
--- DCM.v	2008-01-19 23:23:09.000000000 -0700
+++ DCM_mod.v	2008-01-19 23:38:50.000000000 -0700

@@ -380,3 +380,3 @@
-
+/*
 generate if (SIM_MODE !== "FAST")
-begin
+begin*/
@@ -796,34 +796,24 @@
     end
 end
 
-  assign #(period/2) lock_period_dly = lock_period;
-  assign lock_period_pulse = (lock_period==1 && lock_period_dly==0) ? 1 : 0;
-
 //
 // determine clock delay
 //
 
-//always @(posedge lock_period or posedge rst_in) 
-always @(posedge lock_ps_dly or posedge rst_in) 
-  if (rst_in) begin
-    fb_delay  <= 0;
-    fb_delay_found <= 0;
-  end
-  else begin
-    if (lock_period && clkfb_type != 2'b00) begin
-	if (clkfb_type == 2'b01) begin
+always @(posedge lock_period) begin
+    if (lock_period && clkfb_type != 0) begin
+	if (clkfb_type == 1) begin
 	    @(posedge CLK0 or rst_in)
 		delay_edge = $time;
 	end
-	else if (clkfb_type == 2'b10) begin
+	else if (clkfb_type == 2) begin
 	    @(posedge CLK2X or rst_in)
 		delay_edge = $time;
 	end
-	@(posedge clkfb_in or rst_in) begin
-	 fb_delay <= ($time - delay_edge) % period_orig;
-         fb_delay_found <= 1;
-       end
+	@(posedge clkfb_in or rst_in)
+	fb_delay = ($time - delay_edge) % period_orig;
     end
+    fb_delay_found = 1;
   end
 
 //
@@ -1049,2 +1049,2 @@ 
-end 
+/*end 
-endgenerate
+endgenerate
@@ -1323,1 +1323,1 @@
-endgenerate
+endgenerate*/
EOP

echo "Applying patch to SRL16.v to fix simulation error in iverilog"
echo "Vector of 15'b0 selected by undefined index should return 0";
patch -N unisims/SRL16.v << 'EOP'
--- unisims/SRL16.v	   2008-01-19 23:23:16.000000000 -0700
+++ unisims/SRL16_mod.v	2008-01-19 23:28:49.000000000 -0700
@@ -31,7 +31,7 @@
     reg  [15:0] data;
 
 
-    assign Q = data[{A3, A2, A1, A0}];
+    assign Q = (data[15:0]==16'h0000)?1'b0:((data[15:0]==16'hffff)?1'b1:data[{A3, A2, A1, A0}]);
 
     initial
     begin
EOP

echo "Applying patch to SRL16_1.v to fix simulation error in iverilog"
echo "Vector of 15'b0 selected by undefined index should return 0";
patch -N unisims/SRL16_1.v << 'EOP'
--- unisims/SRL16_1.v   	2008-01-19 23:23:16.000000000 -0700
+++ unisims/SRL16_1_mod.v	2008-01-19 23:29:43.000000000 -0700
@@ -32,7 +32,7 @@
     wire	clk_;
 
 
-    assign Q = data[{A3, A2, A1, A0}];
+    assign Q = (data[15:0]==16'h0000)?1'b0:((data[15:0]==16'hffff)?1'b1:data[{A3, A2, A1, A0}]);
     assign clk_ = ~CLK;
 
     initial
EOP
exit 0
