/*
** -----------------------------------------------------------------------------**
** x324_1.tf
**
** Testbench file for simulating x324
**
** Copyright (C) 2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: x347_1.tf,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.2  2006/09/24 16:45:51  elphel
**  added real time clock
**
*/
`timescale 1ns/1ps


module testbench();
   parameter CLK0_PER = 8;	//125MHz
//   parameter SERIAL_PER = 200;
   parameter I2C_PER = 120;

parameter LOOP=      16'h0000;
parameter ELOOP=     16'h1000;
parameter JUMPOUT=   16'h1400;
parameter JUMP=      16'h1800;
parameter CJUMP=     16'h1c00;
parameter RESET=     16'h2000;
parameter TGLEXPOS=  16'h3000;
parameter TGLFRAME=  16'h4000;
parameter TGLLINE=   16'h5000;
parameter TGLFD=     16'h6000;
parameter TGLSHUTTER=16'h7000;
parameter VPH0=      16'h8000;
parameter VPH1=      16'h9000;
parameter VPH2=      16'ha000;
parameter VPH3=      16'hb000;
parameter VPH4=      16'hc000;
parameter VPH5=      16'hd000;
parameter VPH6=      16'he000;
parameter VPH7=      16'hf000;




// real values

//  parameter NLINES=2672;
//  parameter NPIX=  2040;	// total pixels in a half-line (with black ones)
//  parameter T_HD=   100;   //  4.0 us (3.0/3.5/10.0), HCCD delay
//  parameter T_VCCD= 100;   //  4.0 us (3.0/3.5/20.0), VCCD transfer time
//  parameter T_V3RD= 250;   // 10.0 us (8.0/10.0/15.0), Photodiode transfer time
//  parameter T_V3P= 1500;   // 60.0 us (50.0/60.0/80.0), Photodiode transfer time
//  parameter T_V3D=  500;   // 20.0 us (15.0/20.0/80.0), VCCD delay
//  parameter T_S=    125;   //  5.0 us (3.0/4.0/10.0), Shutter pulse time
//  parameter T_SD=    50;   //  2.0 us (1.0/1.5/10.0), Shutter delay time
//  parameter T_FD=    25;   //  1.0 us (0.5/?/?) Fast Dump Gate delay
//  parameter T_SFD=  2500;   //100.0 us (slow "fast dump") to reduce power

// for simulation

  parameter NLINES=  10;   //2672;
  parameter NPIX=   120;   //2040;	// total pixels in a half-line (with black ones)
  parameter T_HD=   5;  //100;   //  4.0 us (3.0/3.5/10.0), HCCD delay
  parameter T_VCCD= 5;  //100;   //  4.0 us (3.0/3.5/20.0), VCCD transfer time
  parameter T_V3RD= 25;  //250;   // 10.0 us (8.0/10.0/15.0), Photodiode transfer time
  parameter T_V3P=  15;  //1500;   // 60.0 us (50.0/60.0/80.0), Photodiode transfer time
  parameter T_V3D=  10;  // 500;   // 20.0 us (15.0/20.0/80.0), VCCD delay
  parameter T_S=    12;  //125;   //  5.0 us (3.0/4.0/10.0), Shutter pulse time
  parameter T_SD=    5;  //50;   //  2.0 us (1.0/1.5/10.0), Shutter delay time
  parameter T_FD=    4;  //25;   //  1.0 us (0.5/?/?) Fast Dump Gate delay
  parameter T_SFD=  250;  //100.0 us (slow "fast dump") to reduce power
  parameter T_MS=   25;  // 125;   //  5.0 us - rather arbitrary - mechanical shutter trigger duration


  reg [15:0]  SERIAL_READ;

// Inputs
    reg CLK0;
 //   reg CLK1;
 //   reg CLK2;
 //   reg CLK3;
    reg [13:0] ADA;
    reg [13:0] ADB;
    reg SA0;
    reg SB0;
    reg SA1;
    reg SB1;
    reg SA2;
    reg SB2;
    reg SA3;
    reg SB3;

    reg MRST;
    reg ARO;
    reg ARST;


// Outputs
    wire [13:0] DA;
//    wire [13:0] DB;
    wire FSA;
//    wire FSB;
    wire DCLKA;
    wire [9:0] DV;
    wire DVW1;
    wire DVW2;
    wire DVLD;
    wire DAC8C;
    wire DAC8D;
    wire DAC8FS;
    wire DAC8LD;
    wire DAC8R;
    wire SHTR;
    wire FDTTL;
    wire CSRG;
    wire CSH1;
    wire CSHL;
    wire CSH2;
    wire VHSCL;
    wire VHSDI;
    wire H1TTL;
    wire HLTTL;
    wire H2TTL;
    wire RGTTL;
    wire ENAPWR;
    wire [7:0] GP;
    wire ADCCLK;
    wire SHPA;
    wire SHDA;
    wire SHPB;
    wire SHDB;
    wire CLPOB;
    wire CLPDM;
    wire PBLK;
    wire STB;
    wire RUN;
    wire SCL0;
    wire SDA0;

    wire FAN;       // Fan control (on/off)
    wire LENSPWR;   // enable lens power
    wire LENSCLKO;  // lens clock output
    wire LENSDO;    // lens data output
    reg  LENSCLKI;  // lens clock input
    reg  LENSDI;    // lens data input
    

// test registers/wires
   reg I2C_CLK;
   reg HOST_SDA;   
   reg HOST_SCL;   
   reg HOST_SDA_EN;   
   reg HOST_SCL_EN;
   reg [15:0] HOST_READ_DATA;   
   
   assign SDA0= HOST_SDA_EN ? HOST_SDA : 'bz;  
   assign SCL0= HOST_SCL_EN ? HOST_SCL : 'bz;  


// Instantiate the DUT

    x347 i_347 (

            .PXD(DA[13:4]), // [9:0] - host interface, will be redefined for 14-bit data
            .DCLK(FSA),
            .BPF(DCLKA),
            .HACT(DA[0]),
            .VACT(DA[1]),
            .MRST(MRST),
            .ARO(ARO),
            .ARST(ARST),
            .SCL0(SCL0),
            .SDA0(SDA0),
            .CNVSYNC(DA[2]),
            .CNVCLK(DA[3]),
            .RUN(RUN),       // Low when the FPGA is programmed and running
            .CLK0(CLK0),      // 125MHz clock
            .ADCCLK(ADCCLK),    // ADC clock
            .SHPA(SHPA),      // other ADC control signals (outputs)
            .SHDA(SHDA),
            .SHPB(SHPB),
            .SHDB(SHDB),
            .CLPOB(CLPOB),
            .CLPDM(CLPDM),
            .PBLK(PBLK),
            .STB(STB),
            .ADB(ADB[13:0]),       // ADC channel B data  (14 bits)
            .ADA(ADA[13:0]),       // ADC channel A data  (14 bits)
            .DV(DV[9:0]),        // DAC data (10 bits)
            .DV1(DVW1),       // DAC channel 1 data strobe
            .DV2(DVW2),       // DAC channel 2 data strobe
            .DVLD(DVLD),      // DAC channels (both) output update to data
            .FDTTL(FDTTL),     // CCD Fast Drain (TTL)
            .SHTR(SHTR),      // CCD Electronic Shutter
            .CSRG(CSRG),      // CS for CCD RG DAC
            .CSH2(CSH2),      // CS for CCD Hor Ph 2 DAC
            .CSH1(CSH1),      // CS for CCD Hor Ph 1 DAC
            .CSHL(CSHL),      // CS for CCD Hor Ph Last DAC
            .VHSCL(VHSCL),     // SCL  (serial clock) for horizontal phases amplitudes DACs
            .VHSDIN(VHSDI),    // SDIN (serial data in) for horizontal phases amplitudes DACs
            .RGTTL(RGTTL),     // CCD Reset Gate (TTL)
            .H2TTL(H2TTL),     // CCD Horisontal Phase 2 (TTL)
            .HLTTL(HLTTL),     // CCD Horisontal Phase Last (TTL)
            .H1TTL(H1TTL),     // CCD Horisontal Phase 1 (TTL)
            .DAC8D(DAC8D),        // 8-channel DAC (DC biases, horizontal phases biases) serial data in
            .DAC8C(DAC8C),        // 8-channel DAC serial clock     
            .DAC8FS(DAC8FS),      // 8-channel DAC frame strobe
            .DAC8LD(DAC8LD),      // 8-channel DAC outputs update
            .DAC8R(DAC8R),        // 8-channel DAC reset
            .ENAPWR(ENAPWR),      // enable analog power suppies
            .FAN(FAN),            // Fan control (on/off)
            .LENSPWR(LENSPWR),    // enable lens power
            .LENSCLKO(LENSCLKO),  // lens clock output
            .LENSDO(LENSDO),      // lens data output
            .LENSCLKI(LENSCLKI),  // lens clock input
            .LENSDI(LENSDI),      // lens data input
            .GP(GP[7:0]));        // 8-bit GPIO (shutter control) 

	 wire [9:0] DAC2900_A;
    wire [9:0] DAC2900_B;
	 real DAC2900_RA;
	 real DAC2900_RB;
//	 assign DAC2900_RA=DAC2900_A;
//	 assign DAC2900_RB=DAC2900_B;
     always @ (posedge CLK0) begin
	  DAC2900_RA=DAC2900_A;
	  DAC2900_RB=DAC2900_B;
     end
   IC_DAC2900 i_ICDAC2900(.DA(DV[9:0]),
	                       .DB({DV[0],DV[1],DV[2],DV[3],DV[4],DV[5],DV[6],DV[7],DV[8],DV[9]}),
								  .WA(DVW1),
								  .WB(DVW2),
								  .LDA(DVLD),
								  .LDB(DVLD),
								  .QA(DAC2900_A),
								  .QB(DAC2900_B));

always @ (posedge ADCCLK) begin
 #5;
 ADA[13:0] <=ADA[13:0]+1;
 ADB[13:0] <=ADB[13:0]-1;
end


  always #(CLK0_PER/2)	CLK0 =	~CLK0;
  always #(I2C_PER/2)	I2C_CLK =	~I2C_CLK;

    	initial begin
//    $dumpfile("x324.vcd");
    $dumpfile("x347.lxt");
    $dumpvars(0,testbench);
            CLK0 = 0;
            ADA = 0;
            ADB = 0;
            MRST = 1;
            ARO = 1;
            ARST = 1;
// test registers/wires
      I2C_CLK = 0;
      I2C_CLK =0;
      HOST_SDA=1;   
      HOST_SCL=1;   
      HOST_SDA_EN=0;
      HOST_SCL_EN=0;   
#1000;      
		glbl.GSR_int = 1'b1;
		glbl.GTS_int = 1'b1;
		#20;
		glbl.GSR_int = 1'b0;
		#20;
		glbl.GTS_int = 1'b0;
		#50;
      send ('h16, 'h03); // use fan output for mechanical shutter trigger
// test RTC      
      send ('ha0, 'h1234); // seconds high
      send ('ha1, 'h5678); // seconds low
      send ('ha2, 'h9   ); // useconds high
      send ('ha3, 'habcd); // useconds low, set clock
      send ('ha4, 'h7fff); // max positive correction
//  send_a(0,0);
//program fine timing

//      send('h70,'h19);
//      send('h71,'h02);
//      send('h72,'h10);
//      send('h73,'h02);
//      send('h74,'h10);

// watchdog and on/off delays 5.0sec, 0.05 sec, 0.01 sec
//hello -nw 80 1387 31 9
//      send('h80,'h1387); //
      send('h80,'h0300); //
      send('h81,'h031); //
      send('h82,'h009); //

      send('h84,'h0028); //
      send('h85,'h001e); //
      send('h86,'h0003); //
      send('h87,'h000a); //


// programmuing DAC register 9
      send('h29,'h0f);
// programmuing DAC register 0 with data 0
      send('h20,'h0);
// set ld on
      send('h2d,'h0);
// set ld off
      send('h2c,'h0);
// write max5438
      send('h30,'haa);
      send('h31,'h0); // shortest
      send('h32,'h5);
      send('h33,'h55);
// power enable
      send('h11,'h0);
//      send('h10,'h0);

// just test - zero dac2900 out address
      send('h2f,'h0);
// fill ramdac
      send('h40,'h200);
      send('h41,'h340);
      send('h42,'h340);
      send('h43,'h200);
      send('h44,'h200);
      send('h45,'h200);
      send('h46,'h340);
      send('h47,'hc0);
      send('h48,'h340);
      send('h49,'h340);
      send('h4a,'h200);
      send('h4b,'h3ff);
      send('h4c,'h3ff);
      send('h4d,'h200);
      send('h4e,'h200);
      send('h4f,'h340);

// first erase after power on (~10000 lines?)
      send  ('h100,RESET          ); // reset toggle values
      send  ('h101,RESET          ); // just as nop
      send  ('h102,VPH0+T_FD      );  // V1M, V2L
      send  ('h103,TGLFD+T_FD     ); // turn fast dump on
      send  ('h104,LOOP+'h0a      );  // outer loop - 10 times
//       send ('h105,LOOP+'h3e8   );  // inner loop - 1000 times
       send ('h105,LOOP+'h02      );  // inner loop - 2 times
        send('h106,VPH1+T_VCCD    );  // V1L, V2M
        send('h107,VPH0+T_HD      );  // V1M, V2L - NO POWER TO DO THAT !!!

        send('h108,TGLLINE+'h4    ); // PH1, PH2 on
        send('h109,TGLLINE+T_HD   ); // PH1, PH2 off. Not sure in timing

       send ('h10a,ELOOP          );
      send  ('h10b,ELOOP          );
      send  ('h10c,TGLFD+T_HD     ); // turn fast dump off
// Here starts the main loop:
      send  ('h10d,LOOP           );  // loop forever
       send ('h10e,RESET          ); // reset toggle values
       send ('h10f,VPH0+T_FD      );  // V1M, V2L
       send ('h110,TGLFD+T_FD     ); // turn fast dump on
       send ('h111,LOOP           ); // forever until start
        send('h112,VPH1+T_VCCD    );  // V1L, V2M
        send('h113,CJUMP+'h19     );  //LABEL1
        send('h114,VPH0+T_HD      );  // V1M, V2L, slow

        send('h115,TGLLINE+T_SFD  ); // PH1, PH2 on
        send('h116,TGLLINE+T_HD   ); // PH1, PH2 off. Not sure in timing

        send('h117,CJUMP+'h1a     );  //LABEL2
       send ('h118,ELOOP          );
       send ('h119,VPH0+T_FD      );  // V1M, V2L, LABEL1
       // inserting mechanical shutter
       send ('h11a,TGLSHUTTER+T_MS);  // turn mechanical shutter pulse on (5usec)
       send ('h11b,TGLSHUTTER );      // turn mechanical shutter pulse off
//       send ('h11c,LOOP+40        );  // shutter delay *** external address **** (2715 cycles, 108.6usec)
       send ('h11c,LOOP+4        );  // shutter delay *** external address **** (2715 cycles, 108.6usec)
       send('h11d,VPH1+T_VCCD    );  // V1L, V2M  - 100 cycles (4usec)
       send('h11e,VPH0+T_SFD     );  // V1M, V2L, slow - 2500 (100usec)
       send('h11f,TGLLINE+10     );  // PH1, PH2 on  10 cyc - 0.4 usec//flash just a little of horizontal register with each line
       send('h120,TGLLINE+T_HD   );  // PH1, PH2 off. 100 cyc - 4 sec
       send ('h121,ELOOP          );  // end of shutter delay
       // continue with old code.
       send ('h122,TGLFD+T_HD     ); // turn fast dump off (label2)
       send ('h123,TGLSHUTTER+T_S ); // turn shutter on
       send ('h124,TGLEXPOS       ); // turn exposure on
       send ('h125,TGLSHUTTER+T_SD); // turn shutter off
       send ('h126,LOOP+2         );  //exposure, outer loop ***************
        send('h127,VPH0+2         );  //exposure - step duration 2499=100usec
       send ('h128,ELOOP          );
       send ('h129,VPH2+T_V3P     );  // V1M, V2M
       send ('h12a,VPH3+T_V3RD    );  // V1L, V2H
       send ('h12b,TGLEXPOS       ); // turn exposure off
       send ('h12c,VPH2+T_V3D     );  // V1M, V2M
       send ('h12d,VPH0+T_VCCD    );  // V1M, V2L, different delay
       send ('h12e,TGLFRAME       ); // turn frame on
       send ('h12f,LOOP+NLINES    );  // shift out frame - try extra lines **
        send('h130,VPH1+T_VCCD    );  // V1L, V2M
        send('h131,VPH0+T_HD      );  // V1M, V2L
        send('h132,TGLLINE+NPIX   ); // PH1, PH2 on
        send('h133,TGLLINE+T_HD   ); // PH1, PH2 off. Not sure in timing
       send ('h134,ELOOP          );
       send ('h135,TGLFRAME       ); // turn frame off
      send  ('h136,ELOOP          ); // end outer (forever) loop - restart acquisition

// Programming shutter sequencer      
  /* Simple sequencer, no loops, up to 128 commands.
   Each command is 32-bit word (uses 2 of 16-bit ones from the interface - little endian):
   0XXX XXXX XXXX XXXX XXXX XXXX XXXX XXXX - end sequencer
   1XX0 XXXX XXXX NNNN NNNN NNNN NNNN NNNN - nop, just wait N..N clocks
   1X1X XXXX XXXX NNNN NNNN NNNN NNNN NNNN - nop, just wait N..N clocks
   11XX XXXX XXXX NNNN NNNN NNNN NNNN NNNN - nop, just wait N..N clocks
   1001 BBBB BBBB NNNN NNNN NNNN NNNN NNNN - set output to B..B, wait N..N clocks
   N..N==0 - command takes one cycle
   1 cycle is made 5 input clock cycles (5MHz), it is possible to reduce to 4 cycles ( if needed )
*/
      send  ('h200,'h5           ); send  ('h201,'h8000        ); // 80000005 : 5 clocks - 1 usec
      send  ('h202,'h0           ); send  ('h203,'h9010        ); // 90100000 : output[0], no delay
      send  ('h204,'h0           ); send  ('h205,'h9020        ); // 90200000 : output[1], no delay
      send  ('h206,'h1           ); send  ('h207,'h9040        ); // 90400000 : output[2], 1 extra delay
      send  ('h208,'h2           ); send  ('h209,'h9080        ); // 90800000 : output[3], 2 extra delay
      send  ('h20a,'h3           ); send  ('h20b,'h9100        ); // 91000000 : output[4], 3 extra delay
      send  ('h20c,'h4           ); send  ('h20d,'h9200        ); // 92000000 : output[5], 4 extra delay
      send  ('h20e,'h5           ); send  ('h20f,'h9400        ); // 94000000 : output[6], 5 extra delay
      send  ('h210,'h6           ); send  ('h211,'h9800        ); // 98000000 : output[7], 6 extra delay
      send  ('h212,'h0           ); send  ('h213,'h0000        ); // 00000000 : stop



      
      send  ('h11,'h00); // pre-enable power

//  assign cs_seqen =  (addr[9:1]== 9'h9 ); // disable (0x12) and enable (0x13) sequencer 
//  assign cs_expose=  (addr[9:0]==10'h14); // 0x14 - start exposure
// subtract (before multiplying)
      send  ('h50,'h0);  send  ('h51,'h0);  send  ('h52,'h0);  send  ('h53,'h0);
      send  ('h54,'h0);  send  ('h55,'h0);  send  ('h56,'h0);  send  ('h57,'h0);

// program gains
      send  ('h58,'h1000);  send  ('h59,'h1000);  send  ('h5a,'h1000);  send  ('h5b,'h1000);
      send  ('h5c,'h1000);  send  ('h5d,'h1000);  send  ('h5e,'h1000);  send  ('h5f,'h1000);
// window ((top-1)/(height+1)/(left-1)/(width-1))
      send  ('h60,3-1);  send  ('h61,4+1);  send  ('h62, 5-1);  send  ('h63,100-1);

      send  ('h64,1-1);  send  ('h65,1+1);  send  ('h66, 40-1);  send  ('h67,30-1);
      send  ('h68,126-2); // half line (LSB - shift)

// programming timing - periond=5, make same as before
					      // 0 -  write period (default = 4 =>5 cycles, maximum 15->16 cycles). For foolproof everything less than 4 will add 4
							// 1 -  write PH inactive (for period ==5 should be 3)
							// 2 - ADCCLK active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 3 - ADCCLK active at (64-71)(68-75)(72-79) ... (120..127) ns
							// 4 - SHPA   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 5 - SHPA   active at (64-71)(68-75)(72-79) ... (120..127) ns
							// 6 - SHDA   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 7 - SHDA   active at (64-71)(68-75)(72-79) ... (120..127) ns
							// 8 - SHPB   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 9 - SHPB   active at (64-71)(68-75)(72-79) ... (120..127) ns
							// a - SHDB   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// b - SHDB   active at (64-71)(68-75)(72-79) ... (120..127) ns
                     // c - clk_adc active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) 
                     // d - clk_adc active at (64-71)(68-75)(72-79) ... (120..127) ns
                     // e - clk_2x  active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
                     // f - clk_2x  active at (64-71)(68-75)(72-79) ... (120..127) ns
/*
      send('h70,'h0004); // pixel period= 5 clk (40ns)
      send('h71,'h0003); // write PH inactive (for period ==5 should be 3)
      send('h72,'h0105); // ADCCLK  active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h73,'h0000); // ADCCLK  active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h74,'h0040); // SHPA    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h75,'h0000); // SHPA    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h76,'h0001); // SHDA    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h77,'h0000); // SHDA    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h78,'h0040); // SHPB    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h79,'h0000); // SHPB    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h7a,'h0001); // SHDB    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h7b,'h0000); // SHDB    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h7c,'h0105); // clk_adc active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h7d,'h0000); // clk_adc active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h7e,'h0042); // clk2x   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h7f,'h0000); // clk2x   active at (64-71)(68-75)(72-79) ... (120..127) ns
*/
      send('h70,'h0008); // pixel period= 5 clk (40ns)
      send('h71,'h0005); // write PH inactive (for period ==5 should be 3)
      send('h72,'h3540); // ADCCLK  active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h73,'h0000); // ADCCLK  active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h74,'h0400); // SHPA    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h75,'h0000); // SHPA    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h76,'h0001); // SHDA    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h77,'h0000); // SHDA    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h78,'h0400); // SHPB    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h79,'h0000); // SHPB    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h7a,'h0001); // SHDB    active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h7b,'h0000); // SHDB    active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h7c,'h3540); // clk_adc active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h7d,'h0000); // clk_adc active at (64-71)(68-75)(72-79) ... (120..127) ns
      send('h7e,'h2010); // clk2x   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
      send('h7f,'h0000); // clk2x   active at (64-71)(68-75)(72-79) ... (120..127) ns
      
      #10000;
      send  ('h01,'h01); // monitor slavewaiting (undefined)
      #10000;
      send  ('h02,'h01); // set slavewaiting, (will reset monitor)
      #10000;
      send  ('h01,'h01); // re-enable monitor slavewaiting
      #10000;
      send  ('h02,'h00); // reset slavewaiting, (will reset monitor)
      #10000;
      send  ('h01,'h01); // re-enable monitor slavewaiting
      #10000;


      send  ('h14,'h00); // start exposure
		#500000;
      send  ('h14,'h00); // start exposure
		#500000;
      send  ('h14,'h00); // start exposure
// power disable
      #130000;
      send('h10,'h0);
//test RTC      
      
      send ('ha5, 'h0); // latch RTC data
      receive ('h0b0);
      receive ('h0b1);
      receive ('h0b2);
      receive ('h0b3);
      receive ('h0b4);
      receive ('h0b5);
      receive ('h0b6);
      receive ('h0b7);
      receive ('h0b8);
      receive ('h0b9);
      receive ('h0ba);
      receive ('h0bb);
      receive ('h0bc);
      receive ('h0bd);
      receive ('h0be);
      receive ('h0bf);
      
      $finish;

end

/*
parameter LOOP=      16'h0000;
parameter ELOOP=     16'h1000;
parameter JUMPOUT=   16'h1400;
parameter JUMP=      16'h1800;
parameter CJUMP=     16'h1c00;
parameter RESET=     16'h2000;
parameter TGLEXPOS=  16'h3000;
parameter TGLFRAME=  16'h4000;
parameter TGLLINE=   16'h5000;
parameter TGLFD=     16'h6000;
parameter TGLSHUTTER=16'h7000;
parameter VPH0=      16'h8000;
parameter VPH1=      16'h9000;
parameter VPH2=      16'ha000;
parameter VPH3=      16'hb000;
parameter VPH4=      16'hc000;
parameter VPH5=      16'hd000;
parameter VPH6=      16'he000;
parameter VPH7=      16'hf000;
  parameter NLINES=  5;   //2672;
  parameter NPIX=   30;   //2040;	// total pixels in a half-line (with black ones)
  parameter T_HD=   10;  //100;   //  4.0 us (3.0/3.5/10.0), HCCD delay
  parameter T_VCCD= 10;  //100;   //  4.0 us (3.0/3.5/20.0), VCCD transfer time
  parameter T_V3RD= 25;  //250;   // 10.0 us (8.0/10.0/15.0), Photodiode transfer time
  parameter T_V3P=  15;  //1500;   // 60.0 us (50.0/60.0/80.0), Photodiode transfer time
  parameter T_V3D=  50;  // 500;   // 20.0 us (15.0/20.0/80.0), VCCD delay
  parameter T_S=    12;  //125;   //  5.0 us (3.0/4.0/10.0), Shutter pulse time
  parameter T_SD=    5;  //50;   //  2.0 us (1.0/1.5/10.0), Shutter delay time
  parameter T_FD=    4;  //25;   //  1.0 us (0.5/?/?) Fast Dump Gate delay
  parameter T_SD=  250;  //100.0 us (slow "fast dump") to reduce power


command table:
   0000 NNNN NNNN NNNN - begin repeat loop (up to 4095 times). 0000 0000 0000 - forever
	0001 00xx xxxx xxxx - end loop
	0001 01xx aaaa aaaa - jump (aaaa aaaa), pop loop stack (jump out of the loop)
	0001 10xx aaaa aaaa - jump (aaaa aaaa)
	0001 11xx aaaa aaaa - jump (aaaa aaaa) if start pending, pop loop stack
	0010 wwww wwww wwww - reset toggle signals, wait (wwww wwww wwww) cycles
	0011 wwww wwww wwww - toggle "expos", wait
	0100 wwww wwww wwww - toggle "frame", wait
	0101 wwww wwww wwww - toggle "line", wait
	0110 wwww wwww wwww - toggle "fd", wait
	0110 wwww wwww wwww - toggle "shutter", wait
   1000 wwww wwww wwww - set vertical phase #0 V1M V2L
   1001 wwww wwww wwww - set vertical phase #1 V1L V2M
   1010 wwww wwww wwww - set vertical phase #2 V1M V2M
   1011 wwww wwww wwww - set vertical phase #3 V1L V2H
   1100 wwww wwww wwww - set vertical phase #4 V1L V2L
   1101 wwww wwww wwww - set vertical phase #5 V1M V2A
   1110 wwww wwww wwww - set vertical phase #6 V1A V2M
   1111 wwww wwww wwww - set vertical phase #7 ?   ?
*/

//i2c tasks

   task i2c_start;  // SCL is supposed to be 1, SDA - released
     begin
       $display ("i2c START at %t", $time);
       HOST_SDA=1;
       HOST_SCL=1;
       HOST_SDA_EN=0;
       HOST_SCL_EN=1;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SDA_EN=1;
       HOST_SDA=0;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SCL=0;
     end
   endtask   

   task i2c_stop;  // SCL is supposed to be 0, SDA - any
     begin
       $display ("i2c STOP at %t", $time);
       HOST_SCL=0;
       HOST_SCL_EN=1;
       wait (~I2C_CLK);wait (I2C_CLK); // wait for bus turnover (it was likely reading ACKN)
       HOST_SDA=0;
       HOST_SDA_EN=1;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SCL=1;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SDA=1;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SDA_EN=0;
     end
   endtask
   
   task i2c_restart;  // SCL is supposed to be 0, SDA - any
     begin
       $display ("i2c RESTART at %t", $time);
       HOST_SCL=0;
       HOST_SCL_EN=1;
       wait (~I2C_CLK);wait (I2C_CLK); // wait for bus turnover (it was likely reading ACKN)
       HOST_SDA=1;
       HOST_SDA_EN=1;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SCL=1;
       HOST_SDA_EN=0;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SDA=0;
       HOST_SDA_EN=1;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SCL=0;
     end
   endtask

   task i2c_sendbyte;  // SCL is supposed to be 0, SDA - any
     input [7:0] b;
     reg ackn;
     integer i;
     begin
       $display ("i2c send (%x) at %t", b,$time);
       HOST_SCL=0;
       HOST_SCL_EN=1;
       wait (~I2C_CLK);wait (I2C_CLK); // wait for bus turnover (it was reading ACKN)
       for (i=7;i>=0;i=i-1) begin
         HOST_SDA=b[i];
         HOST_SDA_EN=1;
         wait (~I2C_CLK);wait (I2C_CLK);
         HOST_SCL=1;
         if (HOST_SDA) begin
           HOST_SDA_EN=0;
         end
         wait (~I2C_CLK);wait (I2C_CLK);
         HOST_SCL=0;
       end
       HOST_SDA_EN=0; // float for acknowledge
       wait (~I2C_CLK);wait (I2C_CLK);
       ackn=SDA0;
       #1;
       $display ("i2c acknowledge (%x)", ackn);
       HOST_SCL=1;
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SCL=0;
     end
   endtask   

   task i2c_readbyte;  // SCL is supposed to be 0, SDA - any (more)
     input more;
     reg [7:0] d;
     integer i;
     begin
       $display ("i2c read (more=%x) at %t", more, $time);
       HOST_SCL=0;
       HOST_SCL_EN=1;
       HOST_SDA_EN=0;
       for (i=7;i>=0;i=i-1) begin
         wait (~I2C_CLK);wait (I2C_CLK);
         d[i]=SDA0;
         #1;
         HOST_SCL=1;
         wait (~I2C_CLK);wait (I2C_CLK);
         HOST_SCL=0;
       end
       wait (~I2C_CLK);wait (I2C_CLK); // wait for bus turnover (it was reading)
       HOST_SDA=~more;
       HOST_SDA_EN=1; 
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SCL=1;
       if (HOST_SDA) begin
         HOST_SDA_EN=0;
       end
       wait (~I2C_CLK);wait (I2C_CLK);
       HOST_SCL=0;
       $display ("i2c read returned %x", d);
       HOST_READ_DATA=d;   

     end
   endtask   



// non-i2c tasks

/*

   task	send_dibit;
	 input	[ 1:0]	d;
	 begin
//     $display ("    at %t senddibit %x",$time,d);
	   # (SERIAL_PER);
		SB0 <= SB0 ^  d[1]; 
		SA0 <= SA0 ^ ~d[1]; 
	   # (SERIAL_PER);
		SB0 <= SB0 ^  d[0]; 
		SA0 <= SA0 ^ ~d[0]; 
	 end
	endtask

	task send_a;
	 input	w;
	 input	[9:0]	a;
	 reg     [9:0]  a0;
	 reg            first;
	 begin
//     $display ("at %t send_a (%x, %x)",$time,w, a);

	   send_dibit(0);	// start
	   send_dibit(w?2:1);	// w/r
		first <=1'b1;
		a0[9:0]<= a[9:0];
		#1;
//	   $display ("at %t first=%x a0[9:0]=%x", $time,first,a0);
		while (first || (a0!=10'b0)) begin
//	     $display ("  at %t a0= %x",$time,a0);
		  send_dibit(a0[0]?2'b10:2'b01);
		  #1 {first,a0[9:0]}<={2'b0,a0[9:1]};
		  #1;
		end
		if (w) send_dibit(3); // act
	 end
	endtask

	task send_d;
	 input	[15:0]  d;
	 reg     [15:0]  d0;
	 begin
//     $display ("at %t send_d (%x)",$time,d);
      d0<= d;
		#1;
		while (d0!=16'b0) begin
//	     $display ("  at %t d0= %x",$time,d0);
		  send_dibit(d0[0]?2'b10:2'b01);
		  #1 d0[15:0]<={1'b0,d0[15:1]};
		  #1;
		end


		send_dibit(3); // act
	 end
	endtask
   
	task send;
	 input	[9:0] a;
	 input  [15:0]	d;
    begin
     send_a(1,a);
	  send_d(d);
	 end
	endtask

   task receive;
    input   [9:0] a;
    reg [11:0] d;
    begin
     send_a(0,a);
     send_dibit(3); // activate
      
     # (SERIAL_PER);
     d [0] <= SI0;  
     SA0 <= ~SA0; //dibit=1, toggle A
     # (SERIAL_PER);
     d [1] <= SI0;  
     SB0 <= ~SB0; //dibit=1, toggle B
     # (SERIAL_PER);
     d [2] <= SI0;  
     SA0 <= ~SA0; //dibit=1, toggle A
     # (SERIAL_PER);
     d [3] <= SI0;  
     SB0 <= ~SB0; //dibit=1, toggle B
     # (SERIAL_PER);
     d [4] <= SI0;  
     SA0 <= ~SA0; //dibit=1, toggle A
     # (SERIAL_PER);
     d [5] <= SI0;  
     SB0 <= ~SB0; //dibit=1, toggle B
     # (SERIAL_PER);
     d [6] <= SI0;  
     SA0 <= ~SA0; //dibit=1, toggle A
     # (SERIAL_PER);
     d [7] <= SI0;  
     SB0 <= ~SB0; //dibit=1, toggle B
     # (SERIAL_PER);
     d [8] <= SI0;  
     SA0 <= ~SA0; //dibit=1, toggle A
     # (SERIAL_PER);
     d [9] <= SI0;  
     SB0 <= ~SB0; //dibit=1, toggle B
     # (SERIAL_PER);
     d [10] <= SI0;  
     SA0 <= ~SA0; //dibit=1, toggle A
     # (SERIAL_PER);
     d [11] <= SI0;  
     SB0 <= ~SB0; //dibit=1, toggle B
     # (SERIAL_PER);
     SERIAL_READ[11:0] <= d[11:0];
    end
   endtask
   
   
   
   
      i2c_start;
      i2c_sendbyte ('h10);
      i2c_sendbyte ('h01);
      i2c_sendbyte ('h12);
      i2c_sendbyte ('h35);
      i2c_sendbyte ('h56);
      i2c_sendbyte ('h78);
      i2c_stop;
#2000;
      i2c_start;
      i2c_sendbyte ('h10);
      i2c_sendbyte ('h01);
      i2c_restart;      
      i2c_sendbyte ('h11);
      i2c_readbyte (1);
      i2c_readbyte (1);
      i2c_readbyte (1);
      i2c_readbyte (0);
      i2c_stop;
   
   
*/


// send data to register over i2c
	task send;
	 input	[9:0] a;
	 input  [15:0]	d;
    begin
      i2c_start;
      i2c_sendbyte ('h20 | {a[9:8],1'b0});
      i2c_sendbyte (a[7:0]);
      i2c_sendbyte (d[15:8]);
      i2c_sendbyte (d[7:0]);
      i2c_stop;
	 end
	endtask

   task receive;
    input   [9:0] a;
    reg [15:0] d;
    begin
      i2c_start;
      i2c_sendbyte ('h20 | {a[9:8],1'b0});
      i2c_sendbyte (a[7:0]);
      i2c_restart;      
      i2c_sendbyte ('h20 | {a[9:8],1'b1});
      i2c_readbyte (1);
      d[15:8] <= HOST_READ_DATA[7:0];
      i2c_readbyte (0);
      d[ 7:0] <= HOST_READ_DATA[7:0];
      i2c_stop;
     SERIAL_READ[15:0] <= d[15:0];
     end
   endtask
endmodule

