//x353.h
// fpga definitions (should match those of Verilog sources)
  #define X313_MINMODREV      0x03534015  // minimal fpga model/rev that should work with current driver
  #define X313_MAXMODREV      0x035340ff  // maximal fpga model/rev that should work with current driver

#define X313__RA__STATUS	0x10    // read status register
#define X313__RA__IRQS  	0x11    // read interrupt register
#define X313__RA__TRIGPH    0x12    // read trigger phase (line/pixel of readout when trigger came)
#define X313__RA__MODEL     0x13    // read model number/revision of FPGA
#define X313__RA__TABLE     0x14    // readback FPGA tables (not all of them, may be removed later)

#define X313__RA__XFERCNTR  0x14  // readback compressor transfer counter (24 bits, counts 32-byutes chunks)
#define X313__RA__HIGHFREQ  0x15  // 32-bit accumulated value oif hi-frequency components (filter is table defined) - (since 03533014)

  
//  #define X313__RA__USB     0x60    // read USB data
#define X313__RA__IOPINS  0x70    // read state of 12 I/O pins
    // bit 0 - SCL1
    // bit 1 - SDA1
    // bit 2 - XRST
    // bit 3 - AUXCLK
    // bit 4 - EXPS
    // bit 5 - TRIG
#define X313__RA__SENSFPGA  0x74 // - currently will read the same s0x70. Bit 16 (0x10000) is multiplexed state of 1 of 3 pins
#define SFPGA_RD_BIT       16 // Obsolete????
#define I2C_FRAME_NUMBER   0x16 // Read 3 LSB of frame number as seen by I2C controller and i2c busy (0x10000)


#define X313__RA__SDCH0     0x20
#define X313__RA__SDCH1     0x24
#define X313__RA__SDCH2     0x28
#define X313__RA__SDCH3     0x2c
#define X313__RA__SDBUF3	0x30    // 30.37 - data window for PIO access (channel3)

  #define X313_SR__CLK_LOCKED      28
  #define X313_SR__SENS_DCM_OVFL   27
  #define X313_SR__SENS_DCM_LOCKED 26
  #define X313_SR__SENS_DCM_RDY    25

  #define X313_SR__SENS_DCM_EARLY  24
  #define X313_SR__SENS_DCM_LATE   23
// new meaning:
  #define X313_SR__SENS_DCM_ERROR  24 // if 0, the data strobe is in a middle of >= T/2 stable data (never the case with 96MHz at slow drivers)
                                      // if 1 - at least one of the before/after strobes return different data than the actual strobe
  #define X313_SR__SENS_DCM_LATE   23 // "1" - needs phase increase ( 80 for 90 degrees, 20 - for fine)

  #define X313_SENSOR_PHASE ((port_csp0_addr[X313__RA__STATUS] >> X313_SR__SENS_DCM_LATE ) & 3)


  #define X313_SR__DCM_OVFL   22
  #define X313_SR__DCM_LOCKED 21
  #define X313_SR__DCM_RDY    20
  #define X313_SR__DCM_EARLY  19
  #define X313_SR__DCM_LATE   18

  #define X313_SR__DMA_EMPTY  17 //  Dma buffer empty (wait after compressor is done)
  #define X313_SR__DONE_CMPRS 16 // Compressor done (IRQ available)
  #define X313_SR__DONE_CI    15  // Compressor finished reading SDRAM (IRQ available)
  #define X313_SR__DCCRDY     14  // Next 128 (or just the last in frame) DC components are ready
                               // to be read out.
  #define X313_SR__DONE       13  // reset by writing to X313_WA_TRIG
  // 00 - off
  // 01 - waiting fo frame sync to start acquisition
  // 10 - waiting for trigger
  // 11 - acquisition in progr
  #define X313_SR__SENST1     12
  #define X313_SR__SENST0     11
  #define X313_SR__NXTFR3     10
  #define X313_SR__NXTFR2      9
  #define X313_SR__NXTFR1      8
  #define X313_SR__NXTFR0      7
  #define X313_SR__PIOWEMPTY   6
  #define X313_SR__PIORDY      5
  #define X313_SR__CH2RDY      4
  #define X313_SR__CH1RDY      3
  #define X313_SR__CH0RDY      2
  #define X313_SR__SCL0        1
  #define X313_SR__SDA0        0
// Will chnage
  #define X313_PIOR__SCL1    0
  #define X313_PIOR__SDA1    1
  #define X313_PIOR__XRST    2
  #define X313_PIOR__AUXCLK  3
  #define X313_PIOR__EXPS    4
  #define X313_PIOR__TRIG    5



// valid for model 323
#define X313_SR__X323_SI	    0



#define X313_SR(x) ((port_csp0_addr[X313__RA__STATUS] >> X313_SR__##x ) & 1)
#define X313_PIOR(x) ((port_csp0_addr[X313__RA__IOPINS] >> X313_PIOR__##x ) & 1)


#define X313_IR__VACT          0 // start of VACT pulse
#define X313_IR__XINT          1 // external enterrupt
#define X313_IR__XFEROVR       2 // DMA xfer over
#define X313_IR__DONE          3 // DMA xfer over, persistent till reset through...
#define X313_IR__EOT           4
#define X313_IR__DCC           5 // Next 128 16-bit DC components (and HF too) are ready (or the partial block - last in frame).
                                 // Reset when all 128 are read out or DCC mode is reset (0 is written to bit 12 register X313_WA_SDCH3_CTL0 - 0x2c)
#define X313_IR__DONE_INPUT    6     // persistent till compressor reset/restarted
#define X313_IR__DONE_COMPRESS 7  // persistent till compressor reset/restarted
#define X313_IR__SMART         8  // Single cycle, does not need restart, configurable to wait for VACT, always at VACT if no compression is underway

 #define X313_IR(x) ((port_csp0_addr[X313__RA__IRQS] >> ( X313_IR__##x ) + 8) & 1)

 #define EN_INTERRUPT(x)  port_csp0_addr[X313_WA_IRQ_ENA]= (1<< ( X313_IR__##x ))
 #define DIS_INTERRUPT(x) {port_csp0_addr[X313_WA_IRQ_DIS]= (1<< ( X313_IR__##x ) ); port_csp0_addr[X313_WA_IRQ_RST]= (1<< ( X313_IR__##x ) );}
 #define DIS_INTERRUPTS   {port_csp0_addr[X313_WA_IRQ_DIS]= 0xffff; port_csp0_addr[X313_WA_IRQ_RST]= 0xffff;}


/* peripherial write addresses (relative to csp0, in long words) */
#define X313_WA_WCTL	    0 // write control register - 32
#define X313_WA_DMACR	    1 // DMA control register:
                              // 17:    DMA enable
                              // 16:    0- raw, 1 - JPEG
                              // 15-10: not used
                              // 0-9:   line length in long words
//!UPDATED 8.0.10
                             // 5 - enable update DMA1 config (bits 4,3)
                             // 4 - PIO mode on DMA1
                             // 3 - enable DMA1
                             // 2 - enable update DMA0 config (bits 4,3)
                             // 1 - PIO mode on DMA0
                             // 0 - enable DMA0

#define X313_WA_SENSFPN	    2 // Sensor/FPN control register:
                              // 31-11: not used
                                       // [10] - testmode - if 1 generates gradient data (as Zoran chips) where pixel value = horizontal position
                                       // [9:7] - submode - subtract background (8 bits) mode (use di[7:0]):
                                       // 000 - no subtraction;
                                       // 001 - (finest) subtract 8 bit bkgnd from 12 bits pixels
                                       // 010 - shift 8bit bkgnd 1 bit  left before applying
                                       // 011 - shift 8bit bkgnd 2 bits left before applying
                                       // 100 - shift 8bit bkgnd 2 bits left before applying
                                       // 101 - shift 8bit bkgnd 2 bits left before applying
                                       // fpn data to subtract should be a little less to add a "fat zero" 
                                       // [6:4] - mpymode sensitivity correction mode (use di[15:8]):
                                       // 000 - no correction
                                       // 001 - fine correction (+/-3.125%)
                                       // 010 - fine correction (+/-6.25%)
                                       // 011 - fine correction (+/-12.5%)
                                       // 100 - +/- 25%
                                       // 101 - +/- 50%
                                       // [3] - wdth - word width: 0 - 8 bit, 1 - 16 bit (5 MSB == 0)
                                       // [2:0] scaling of 11 bit FPN result to fit in 8bit output:
                                       // 00 - default - use [9:1]
                                       // 01 - use [10:2] - to protect from saturation after applying mpymode
                                       //      nominal range - 0..127
                                       // 10 - use [7:0] before saturation, "digital gain" == 4 (maximal)
                                       // 11 - use [8:1] before saturation, "digital gain" == 2
                              // Result data is saturated by 255

#define X313_SENSFPN_D(t,s,m,d,l)       (((t) & 1) << 10) | \
                                        (((s) & 7) << 7) | \
                                        (((m) & 7) << 4) | \
                                        (((d) & 1) << 3) | \
                                        (((l) & 7) << 0)

#define X313_WA_VIRTTRIG   3  // Virtual trigger threshold
                              //   31-22  not used
                              //   21-0  Trigger will fire if sum of pixels in a line
                              //         (after FPN processig) is more than this value.
                              //         Disabled if 0 - selected real (electrical) external trigger
#define X313_WA_TRIG       4  // Sensor triggrring
                              // 31-3: not used
                              //    2: Enable sensor (0 - abort at once)
                              //    1: External Trigger (0 - internal)
                              //    0: continuous acquisition (0 - single "frame" - actually number of lines specified)
#define X313_WA_NLINES     5  //  number of lines to acquire (afer trigger)
                              // 31-16: not used
                              // 15 - select photofinish mode frames in one (1), nummber of lines (0)
                              // 14 - 0 - write nilines/photofinish frames, 1 - write line width
                              // if photofinish  (([15]==0) && ([14]==0)) FIXME: change control in the driver, now FF mode is written separately - FIXED
                              // 13-0 : number of frames to combine it one (decimate frame sync pulses)
                              // else (([15]==0) && ([14]==0))
                              // 13-0 : number of lines to acquire (in a frame or after the external trigger)
                              // else ([14]==1)
                              // 13-0 : number of pixels to acquire if internal line width is used.
//
#define X313_WA_WCTL24     6  // write control register - same as X313_WA_WCTL, but only 24 lower bits are affected 

#define X313_WA_DCDC       7  // sensor DC-DC converter frequency settings
                              // 31-05: not used
                              //   4-0  0 - use internal clock (not sync)
                              //        1..31 use divided pixel clock
                              //        for 20MHz use 5'h10,
                              //        else - N= (Fpix[MHz]/1.2)-1, if Fpix=20MHz,
                              //        N= 15.7->16=5'h10

  #define X313_WA_DCM      8  // SDRAM clock phase shift
//  #define X3X3_RSTSENSDCM     {port_csp0_addr[X313_WA_DCM]=0x30;} // needed after changing frequency to restart sensor DCM
  #define X3X3_RSTSENSDCM     {port_csp0_addr[X313_WA_DCM]=0xf0;} // reset 90 degrees here too
  #define X3X3_SENSDCM_DEC90  {port_csp0_addr[X313_WA_DCM]=0x80;} // switch clock phase in precise 90-degree increments (approximately same as 100 steps of 0x20)
  #define X3X3_SENSDCM_INC90  {port_csp0_addr[X313_WA_DCM]=0x40;} // switch clock phase in precise 90-degree increments (approximately same as 100 steps of 0x10)
// NOTE fine steps range is limited, combine with 90-degree steps when needed
/// NOTE resetting DCM causes skip in output clock for several cycles!
  #define X3X3_SENSDCM_DEC    {port_csp0_addr[X313_WA_DCM]=0x20;} // switch clock phase in fine steps (~110 for 90 degrees @96MHz)
  #define X3X3_SENSDCM_INC    {port_csp0_addr[X313_WA_DCM]=0x10;} // switch clock phase in fine steps (~110 for 90 degrees @96MHz)
  #define X3X3_SENSDCM_NOP    {port_csp0_addr[X313_WA_DCM]=0x0;} // resets DCM errors (before measuring sensor phase)
/// Control for HACT/VACT delay with respect to PXD (-270..+360 degrees in 90-degree steps - 8 steps cycle)
  #define X3X3_SENSDCM_HACT_LATE90  {port_csp0_addr[X313_WA_DCM]=0x100;} // delay measuring HACT by 90 degrees
  #define X3X3_SENSDCM_HACT_EARLY90 {port_csp0_addr[X313_WA_DCM]=0x200;} // advance measuring HACT by 90 degrees
  #define X3X3_SENSDCM_HACT_ZERO    {port_csp0_addr[X313_WA_DCM]=0x300;} // reset HACT strobe to PXD (zero delay)
  #define X3X3_SENSDCM_CLK2X_RESET  {port_csp0_addr[X313_WA_DCM]=0x400;} // reset pclk2x DCM (used fro histograms, needed after changing pclk)

#define	X313_WA_COLOR_SAT  9  // bits  9: 0 - Cb (blue) color saturation. 0x90 (default) for saturation =1.0 
                              // bits 21:12 - Cr (red) color saturation. 0xb6 (default) for saturation =1.0 
#define DEFAULT_COLOR_SATURATION_BLUE 0x90 // 100*realtive saturation blue
#define DEFAULT_COLOR_SATURATION_RED  0xb6 // 100*realtive saturation red

#define X313_WA_FRAMESYNC_DLY     0x0a  // Delay frame sync interrupt by number of lines (default - 0)
#define X313_WA_QUANTIZER_MODE    0x0b  // Quantizer tuning - 0..7 - zero bin, 15:8 - quantizer bias

#define	X313_WA_COMP_CMD   0x0c
// rev 7.2 control bits (written to X313_WA_COMP_CMD)
// [23] ==1 - set focus mode
// [22:21] 0 - none
//         1 - replace
//         2 - combine for all image
//         3 - combine in window only
 #define COMPCMD_FOCUS(x) ((1<<23) | (((x) & 3) << 21))

// [20] ==1 - set bayer shift (from the gammas)
// [19:18]  - bayer shift value (from the gammas)
 #define COMPCMD_BAYERSHIFT(x) ((1<<20) | (((x) & 3) << 18))

// [17] == 1 - set processed tile position in the 20x20 tile got from memory
// [16:14] 0 - top left alignment
//         1 - 1 pixel  right and 1 - down from the top left alignment
//         2 - 2 pixels right and 2 - down from the top left alignment
//         3 - 3 pixels right and 3 - down from the top left alignment
//         4 - 4 pixels right and 4 - down from the top left alignment
 #define COMPCMD_TILESHIFT(x) ((1<<17) | (((x) & 7) << 14))

// [13]==1 - enable color modes
// [12:9]== 0 - monochrome, (4:2:0),
//          1 - color, 4:2:0, 18x18(old)
//          2 - jp4, original (4:2:0),
//          3 - jp4, dc -improved (4:2:0),
//          4 - color, 4:2:0, 20x20, middle of the tile (not yet implemented)
//          5 - jp4, 4 blocks, (legacy)
//          6 - jp4, 4 blocks, dc -improved
//          7 - jp4, 4 blocks, differential red := (R-G1), blue:=(B-G1), green=G1, green2 (G2-G1). G1 is defined by Bayer shift, any pixel can be used
//          8 - jp4, 4 blocks, differential HDR: red := (R-G1), blue:=(B-G1), green=G1, green2 (high gain)=G2) (G1 and G2 - diagonally opposite)
//          9 - jp4, 4 blocks, differential, divide differences by 2: red := (R-G1)/2, blue:=(B-G1)/2, green=G1, green2 (G2-G1)/2
//         10 - jp4, 4 blocks, differential HDR: red := (R-G1)/2, blue:=(B-G1)/2, green=G1, green2 (high gain)=G2), 
//         11-13 - reserved
//         14 - mono, 4 blocks
//         15 - reserved

 #define COMPCMD_DEMOS(x) ((1<<13) | (((x) & 0x0f) << 9))
 //  6 blocks output per macroblock:
 #define DEMOS_MONO6     0 // original monochrome YCbCr 4:2:0 with zeroed color components
 #define DEMOS_COLOR18   1 // original color YCbCr 4:2:0, 3x3 demosaic (18x18 tiles) 
 #define DEMOS_JP46      2 // original jp4, (4:2:0, zero color), decoded by regular JPEG decoder
 #define DEMOS_JP46DC    3 // dc-improved: same as DEMOS_JP46, but DC difference separate for each component
 #define DEMOS_COLOR20   4 // color YCbCr 4:2:0, 5x5 demosaic (20x20 tiles) - not yet implemented
//  4 blocks output per macroblock:
 #define DEMOS_JP4       5 // similar to DEMOS_JP46 ,   but zero color components are not output
 #define DEMOS_JP4DC     6 // similar to DEMOS_JP46DC , but zero color components are not output
 #define DEMOS_JP4DIFF   7 // differential red := (R-G1), blue:=(B-G1), green=G1, green2 (G2-G1). G1 is defined by Bayer shift, any pixel can be used
 #define DEMOS_JP4HDR    8 // similar to DEMOS_JP4DIFF, but second green (opposite from the reference one) is encoded without subtracting:
                           // red := (R-G1), blue:=(B-G1), green=G1, green2 (high gain)=G2) (G1 and G2 - diagonally opposite)
 #define DEMOS_JP4DIFF2  9 // similar to DEMOS_JP4DIFF, but all differences are divided by 2 to fit into 8 bit range:
                           // red := (R-G1)/2, blue:=(B-G1)/2, green=G1, green2 (G2-G1)/2
 #define DEMOS_JP4HDR2  10 // red := (R-G1)/2, blue:=(B-G1)/2, green=G1, green2 (high gain)=G2), 
 #define DEMOS_MONO4    14 // monochrome, but the block scan order is still the same as in YCbCr 4:2:0 (macroblocks in scan order, block in 2x2 macroblock in scan order)

// [8:7] == 0,1 - NOP, 2 - disable, 3 - enable subtracting of average value (DC component), bypassing DCT

 #define COMPCMD_DCSUB(x) ((1<<8) | (((x) & 1) << 7))  

// [6] == 1 - enable quantization bank select, 0 - disregard bits [5:3]
// [5:3] = quantization page number (0..7)

 #define COMPCMD_QTAB(x) ((1<<6) | (((x) & 7) << 3))  
 #define FPGA_NQTAB     8 // total number of quantization table pairs

 #define COMPCMD_RESET  4 // reset compressor, stop immediately
 #define COMPCMD_STOP   5 // enable compressor, disable repetitive mode (will finish current frame)
 #define COMPCMD_SINGLE 6 // enable compressor, acquire single frame (will disable repetitive mode if it was set)
 #define COMPCMD_RUN    7 // enable repetitive mode
// [2]== 1 - enable on/off control:
// [1:0]== 0 - reset compressor, stop immediately
//         1 - enable compressor, disable repetitive mode
//         2 - enable compressor, compress single frame
//         3 - enable compressor, enable repetitive mode
//

  #define	X313_WA_COMP_TA	   0x0e
  #define	X313_WA_COMP_TD	   0x0f
  #define	X313_WA_MCUNUM	   0x0d
///FIXME: (in FPGA) - now if "smart" mode is disabled, VACT will not cause interrupt on smart bit. So smart mode will be always enabled for now
  #define   X313_WA_SMART_IRQ 0x1a /// configure smart IRQ (3-en/2-dis "smart" mode - waiting for next frame sync for early compression done)
                                   /// 0xc - en/0x8 - dis waiting for DMA FIFO empty
                                   /// 0x8000 - reset circuitry
  #define   X313_WA_DCM_RST   0x1b // async reset of the system DCM
  #define   X313_WA_IRQ_RST   0x1c // reset selected interrupt bits
  #define   X313_WA_IRQ_DIS   0x1d // disable selected interrupt bits (mask)
  #define   X313_WA_IRQ_ENA   0x1e // enable selected interrupt bits
  #define   X313_WA_IRQ_WVECT 0x1f // write vector number (in bits [0:7], [11:8] - interrupt number (0..15)

 #define EN_INTERRUPT(x)  port_csp0_addr[X313_WA_IRQ_ENA]= (1<< ( X313_IR__##x ))
 #define DIS_INTERRUPT(x) {port_csp0_addr[X313_WA_IRQ_DIS]= (1<< ( X313_IR__##x ) ); port_csp0_addr[X313_WA_IRQ_RST]= (1<< ( X313_IR__##x ) );}
 #define DIS_INTERRUPTS   {port_csp0_addr[X313_WA_IRQ_DIS]= 0xffff; port_csp0_addr[X313_WA_IRQ_RST]= 0xffff;}
 

  // changed for rev 03533001f
  #define   X313_WA_HIST_LEFT   0x40
  #define   X313_WA_HIST_TOP    0x41
  #define   X313_WA_HIST_WIDTH  0x42
  #define   X313_WA_HIST_HEIGHT 0x43
  #define   X313_WA_HIST_ADDR   0x44
  #define   X313_RA_HIST_DATA   0x45  // use CSP4 with wait cycles to have a pulse (is timing for CSP4 now the same?)

  #define   X313_WA_RTC_USEC  0x48
  #define   X313_WA_RTC_SEC   0x49  // sets both seconds and microseconds, so should be written second
  #define   X313_WA_RTC_CORR  0x4a
  #define   X313_WA_RTC_LATCH 0x4b
  #define   X313_RA_RTC_USEC  0x48
  #define   X313_RA_RTC_SEC   0x49
  #define   X313_WA_TIMESTAMP 0x4c  // 0 - no, 1 - normal frames, 2 - photofinish
  #define   X313_TIMESTAMPLEN 28 // pixels used for timestamp


  #define   X313_WA_DCR0      0x4e
  #define   X313_WA_DCR1      0x4f
/// FPGA control bits/field can be modified without keeping shadow register, it is important when using "write only" sequencer,
/// that just writes data at appropriate time, so CPU does not know the current shadow
/// The following macro  returns data to be written to X313_WA_DCR0 or X313_WA_DCR1 (same register can be OR-ed) to
/// change control bits for x to the value y
#define X353_DCR0(x,y)  (((((y) & ((1 << X353DCR0__##x##__WIDTH)-1))) | (1 << X353DCR0__##x##__WIDTH) ) << X353DCR0__##x##__BITNM) 
#define X353_DCR1(x,y)  (((((y) & ((1 << X353DCR1__##x##__WIDTH)-1))) | (1 << X353DCR1__##x##__WIDTH) ) << X353DCR1__##x##__BITNM) 


///============ X313_WA_DCR0  ==============
#define  X353DCR0__BAYER_PHASE__BITNM 0
#define  X353DCR0__BAYER_PHASE__WIDTH 2

//#define  X353DCR0__FILLFACTORY__BITNM 3
//#define  X353DCR0__FILLFACTORY__WIDTH 1

#define  X353DCR0__HACT_REGEN__BITNM 3 // use HACT duration from the memory channel0 (0 - use one from the sensor)
#define  X353DCR0__HACT_REGEN__WIDTH 1


//cb_hact_regen
///#define  X353DCR0__DLYHOR__BITNM      5 // removed - not needed, variable Bayer shift replaces it
///#define  X353DCR0__DLYHOR__WIDTH      1


/// RESET_MCONTR may cause infrequent interrupt loss
#define  X353DCR0__RESET_MCONTR__BITNM      5 // 1 - enable reset memory controllers (channles 0,1,2) at each frame
#define  X353DCR0__RESET_MCONTR__WIDTH      1


#define  X353DCR0__NEGRST__BITNM      7 // "zoran"
#define  X353DCR0__NEGRST__WIDTH      1

///#define  X353DCR0__SKIPLINEL__BITNM   9 // removed - not needed, variable Bayer shift replaces it
///#define  X353DCR0__SKIPLINE__WIDTH    1

#define  X353DCR0__USE_SENSOR_CLK__BITNM   9 // 1 - use clock output from the senor as a reference for the sensor PXD/HACT/VACT
#define  X353DCR0__USE_SENSOR_CLK__WIDTH    1

#define  X353DCR0__XT_POL__BITNM     11
#define  X353DCR0__XT_POL__WIDTH      1
#define  X353DCR0__ARST__BITNM       13
#define  X353DCR0__ARST__WIDTH        1
#define  X353DCR0__ARO__BITNM        15
#define  X353DCR0__ARO__WIDTH         1

// for 5MPix set - old sensors, reset - new ones
#define  X353DCR0__CNVEN__BITNM      17
#define  X353DCR0__CNVEN__WIDTH       1

#define  X353DCR0__SENSTRIGEN__BITNM 19
#define  X353DCR0__SENSTRIGEN__WIDTH  1

#define  X353DCR0__ENDFRAMES__BITNM  21
#define  X353DCR0__ENDFRAMES__WIDTH   1


///============ X313_WA_DCR1  ==============
#define  X353DCR1__MRST__BITNM        0
#define  X353DCR1__MRST__WIDTH        1

// use start of trigger as a timestamp (in async mode to prevent timestamp jitter)
#define  X353DCR1__EXTERNALTS__BITNM   2 // use external timestamp if available
#define  X353DCR1__EXTERNALTS__WIDTH   1

//DCLKMODE -  0 - DCLK - clock to sensor, 1 - DCLK - input of conposite sync (from 10347)
#define  X353DCR1__DCLKMODE__BITNM    4
#define  X353DCR1__DCLKMODE__WIDTH    1

//PXD14 - 1 - 14-bit data from sensor 
#define  X353DCR1__PXD14__BITNM       6
#define  X353DCR1__PXD14__WIDTH       1

#define  X353DCR1__HACT_PHASE__BITNM  8
#define  X353DCR1__HACT_PHASE__WIDTH  2

// source of pixel clock. Now 0 - internal (CLK1), 1,2,3 - external (bpf)
#define  X353DCR1__PCLKSRC__BITNM    11
#define  X353DCR1__PCLKSRC__WIDTH     2

#define  X353DCR1__HFCOMP__BITNM     14
#define  X353DCR1__HFCOMP__WIDTH      3

#define  X353DCR1__BLOCKVSYNC__BITNM 18
#define  X353DCR1__BLOCKVSYNC__WIDTH  1

#define  X353DCR1__OUTPUTTS__BITNM 20 // output timestamp, not just pulse
#define  X353DCR1__OUTPUTTS__WIDTH  1


/// Macros that replace old ones for the control register in cc313.h
/// these macros write directly, bypassing the sequencer
/// For sequencer (preferred way in most cases) - use something like  X3X3_SEQ_SEND1(fpga_addr,  X313_WA_DCR0, X353_DCR0(SENSTRIGEN,async);
//#define  X353DCR0__RESET_MCONTR__BITNM      5 // 1 - enable reset memory controllers (channles 0,1,2) at each frame
//#define  X353DCR0__RESET_MCONTR__WIDTH      1
#define CCAM_RESET_MCONTR_ON  port_csp0_addr[X313_WA_DCR0]=X353_DCR0(RESET_MCONTR,1)
#define CCAM_RESET_MCONTR_OFF port_csp0_addr[X313_WA_DCR0]=X353_DCR0(RESET_MCONTR,0)

#define CCAM_ARO_ON   port_csp0_addr[X313_WA_DCR0]=X353_DCR0(ARO,1)
#define CCAM_ARO_OFF  port_csp0_addr[X313_WA_DCR0]=X353_DCR0(ARO,0)
/// turn clock to sensor on/off (default - on)
#define CCAM_DCLK_OFF port_csp0_addr[X313_WA_DCR1]=X353_DCR1(DCLKMODE,1)
#define CCAM_DCLK_ON  port_csp0_addr[X313_WA_DCR1]=X353_DCR1(DCLKMODE,0)
#define CCAM_ARST_OFF port_csp0_addr[X313_WA_DCR0]=X353_DCR0(ARST,1)
#define CCAM_ARST_ON  port_csp0_addr[X313_WA_DCR0]=X353_DCR0(ARST,0)
#define CCAM_MRST_OFF port_csp0_addr[X313_WA_DCR1]=X353_DCR1(MRST,1)
#define CCAM_MRST_ON  port_csp0_addr[X313_WA_DCR1]=X353_DCR1(MRST,0)
#define CCAM_NEGRST   port_csp0_addr[X313_WA_DCR0]=X353_DCR0(NEGRST,1)
#define CCAM_POSRST   port_csp0_addr[X313_WA_DCR0]=X353_DCR0(NEGRST,0)
/// enable power converter control. *** Only if it is not MT9P001 !!! ***
#define CCAM_CNVEN_ON  port_csp0_addr[X313_WA_DCR0]=X353_DCR0(CNVEN,1)
#define CCAM_CNVEN_OFF port_csp0_addr[X313_WA_DCR0]=X353_DCR0(CNVEN,0)
//! Sensor (MT9X001) trigger source selection (internal/external)
#define CCAM_TRIG_INT port_csp0_addr[X313_WA_DCR0]=X353_DCR0(SENSTRIGEN,0)
#define CCAM_TRIG_EXT port_csp0_addr[X313_WA_DCR0]=X353_DCR0(SENSTRIGEN,1)

//! Enable ending frame being compressed if no more data will be available
//! (frame ended before specified number of blocks compressed). If 0 - old functionality

#define CCAM_ENDFRAMES_DIS port_csp0_addr[X313_WA_DCR0]=X353_DCR0(ENDFRAMES,0)
#define CCAM_ENDFRAMES_EN  port_csp0_addr[X313_WA_DCR0]=X353_DCR0(ENDFRAMES,1)


//! early/normal timestamp in async mode
#define CCAM_EXTERNALTS_DIS  port_csp0_addr[X313_WA_DCR1]=X353_DCR1(EXTERNALTS,0)
#define CCAM_EXTERNALTS_EN   port_csp0_addr[X313_WA_DCR1]=X353_DCR1(EXTERNALTS,1)
/// Set HACT phase (90 degrees increments - needed for MT9P001
#define CCAM_SET_HACT_PHASE(x) port_csp0_addr[X313_WA_DCR1]=X353_DCR1(HACT_PHASE,(x))

/// disable (CCAM_VSYNC_OFF) and enable (CCAM_VSYNC_ON) sensor vertical sync pulses
/// to i2c and command sequencer to prevent their mismatch (it also disable "smart" interrupts
/// so before enabling (from reset state) these two sequencer - disable vsync and
/// reenable it after both sequencers are enabled
#define CCAM_VSYNC_ON  port_csp0_addr[X313_WA_DCR1]=X353_DCR1(BLOCKVSYNC,0)
#define CCAM_VSYNC_OFF port_csp0_addr[X313_WA_DCR1]=X353_DCR1(BLOCKVSYNC,1)

#define CCAM_OUTPUTTS_DIS  port_csp0_addr[X313_WA_DCR1]=X353_DCR1(OUTPUTTS,0)
#define CCAM_OUTPUTTS_EN   port_csp0_addr[X313_WA_DCR1]=X353_DCR1(OUTPUTTS,1)




/// #define IS_KAI11000 (ccam_cr_shadow & X313_MASK(PXD14))

///=============================

// following registers accept 32-bit data words to be sent as 4 (or less) i2c command, that consists of
// start, slave address (MSB), register address (bits 16..23), and 2 bytes of data (MSB first).
// For single-byte registers LSB of the data word is not used, data byte uses bits 8..15
// Example 0x902b0010 - write 0x0010 to register 0x2b of slave 0x90
// Writes use relative (to the current frame) and absolute (modulo 8) addresses (readable through I2C_FRAME_NUMBER)
// When using absolute addressing writing to the frame one before current will effectively write to the current frame,
// so if new frame happens during writing - data will not be lost - it will transferred to the next frame.
// Each frame buffer can hold 63 commands, current frame can process unlimited number of commands as long as FIFO does not have
// more than 63 at any given moment.


  #define   X313_I2C_FRAME0   0x50 // write command to be sent at frame number 0 (modulo 8) - see I2C_FRAME_NUMBER, 3 LSBs
  #define   X313_I2C_FRAME1   0x51 // same for frame number 1
  #define   X313_I2C_FRAME2   0x52 // same for frame number 1
  #define   X313_I2C_FRAME3   0x53 // same for frame number 1
  #define   X313_I2C_FRAME4   0x54 // same for frame number 1
  #define   X313_I2C_FRAME5   0x55 // same for frame number 1
  #define   X313_I2C_FRAME6   0x56 // same for frame number 1
  #define   X313_I2C_FRAME7   0x57 // same for frame number 1
  #define   X313_I2C_ASAP     0x58 // write command to be sent to sensor ASAP
  #define   X313_I2C_NEXT     0x59 // write command to be sent after next frame starts
  #define   X313_I2C_NEXT2    0x5a // write command to be sent after next 2 frames start
  #define   X313_I2C_NEXT3    0x5b // write command to be sent after next 2 frames start
  #define   X313_I2C_NEXT4    0x5c // write command to be sent after next 2 frames start
  #define   X313_I2C_NEXT5    0x5d // write command to be sent after next 2 frames start
  #define   X313_I2C_NEXT6    0x5e // write command to be sent after next 2 frames start
  #define   X313_I2C_CMD      0x5f // write command to i2c controller, command consists of several independent bit fields
// below data that can be written to port_csp0_addr[X313_I2C_CMD], OR-ed
  #define X3X3_SET_I2C_DLY(x)   (0x100 |  ((x) & 0xff))
  #define X3X3_SET_I2C_BYTES(x) (0x800 | (((x)<<9) & 0x600))
  #define X3X3_I2C_RUN_BITS     0x3000 // can not be combined with X3X3_I2C_RESET
  #define X3X3_I2C_STOP_BITS    0x2000
  #define X3X3_I2C_RESET_BITS   0x4000 
/// software control of SDA0, SCL0 lines (when hardware i2c is off)
  #define X3X3_I2C_SCL_0_BITS   0x10000 
  #define X3X3_I2C_SCL_1_BITS   0x20000 
  #define X3X3_I2C_SCL_Z_BITS   0x30000 
  #define X3X3_I2C_SDA_0_BITS   0x40000 
  #define X3X3_I2C_SDA_1_BITS   0x80000 
  #define X3X3_I2C_SDA_Z_BITS   0xc0000 

  #define X3X3_I2C_IS_BUSY   (((port_csp0_addr[I2C_FRAME_NUMBER] | \
                                port_csp0_addr[I2C_FRAME_NUMBER] | \
                                port_csp0_addr[I2C_FRAME_NUMBER] | \
                                port_csp0_addr[I2C_FRAME_NUMBER] | \
                                port_csp0_addr[I2C_FRAME_NUMBER] ) & 0x10000)?1:0)
  #define X3X3_I2C_FRAME       (port_csp0_addr[I2C_FRAME_NUMBER]  & 0x7)
  #define X3X3_I2C_SEND2(a,s,r,d) {port_csp0_addr[a] = ((s)<<24) | (((r) & 0xff) << 16) | ((d) & 0xffff) ; X3X3_AFTERWRITE ;} /// is there any need for X3X3_AFTERWRITE???
  #define X3X3_I2C_SEND1(a,s,r,d) {port_csp0_addr[a] = ((s)<<24) | (((r) & 0xff) << 16) | (((d) & 0xff) << 8) ; X3X3_AFTERWRITE ;}

  #define X3X3_GAMMA_PAGE     ((port_csp0_addr[I2C_FRAME_NUMBER]  & 0x20000)?1:0) // gamma table page (0/1) currently used
/// X3X3_I2C_STOP_WAIT does not stop frame counter - it still counts frame pulses
  #define X3X3_I2C_STOP_WAIT  {port_csp0_addr[X313_I2C_CMD]=X3X3_I2C_STOP_BITS; while (X3X3_I2C_IS_BUSY) ; }
  #define X3X3_I2C_RUN        {port_csp0_addr[X313_I2C_CMD]=X3X3_I2C_RUN_BITS ; X3X3_AFTERWRITE ;}
  #define X3X3_I2C_RESET_WAIT {port_csp0_addr[X313_I2C_CMD]=X3X3_I2C_RESET_BITS; while (X3X3_I2C_IS_BUSY) ; }


  #define   X313_SEQ_FRAME0   0x60 // write command to be sent at frame number 0 (modulo 8) - see CMD_FRAME_NUMBER, LSBs, bits [5:3]
  #define   X313_SEQ_FRAME1   0x61 // same for frame number 1
  #define   X313_SEQ_FRAME2   0x62 // same for frame number 2
  #define   X313_SEQ_FRAME3   0x63 // same for frame number 3
  #define   X313_SEQ_FRAME4   0x64 // same for frame number 4
  #define   X313_SEQ_FRAME5   0x65 // same for frame number 5
  #define   X313_SEQ_FRAME6   0x66 // same for frame number 6
  #define   X313_SEQ_FRAME7   0x67 // same for frame number 7
  #define   X313_SEQ_ASAP     0x68 // write command to be sent  ASAP
  #define   X313_SEQ_NEXT     0x69 // write command to be sent after next frame starts
  #define   X313_SEQ_NEXT2    0x6a // write command to be sent after next 2 frames start
  #define   X313_SEQ_NEXT3    0x6b // write command to be sent after next 3 frames start
  #define   X313_SEQ_NEXT4    0x6c // write command to be sent after next 4 frames start
  #define   X313_SEQ_NEXT5    0x6d // write command to be sent after next 5 frames start
  #define   X313_SEQ_NEXT6    0x6e // write command to be sent after next 6 frames start
  #define   X313_SEQ_CMD      0x6f // write command to i2c controller, command consists of several independent bit fields

  #define X3X3_SEQ_RUN_BITS     0x3000 // can not be combined with X3X3_I2C_RESET
  #define X3X3_SEQ_STOP_BITS    0x2000
  #define X3X3_SEQ_RESET_BITS   0x4000 

  #define X3X3_SEQ_SEND1(f,a,d) {port_csp0_addr[f] = ((a)<<24) | ((d) & 0xffffff) ; X3X3_AFTERWRITE ;} /// is there any need for X3X3_AFTERWRITE???

  #define X3X3_SEQ_STOP  {port_csp0_addr[X313_SEQ_CMD]=X3X3_SEQ_STOP_BITS;}
  #define X3X3_SEQ_RUN   {port_csp0_addr[X313_SEQ_CMD]=X3X3_SEQ_RUN_BITS ; X3X3_AFTERWRITE ;}
  #define X3X3_SEQ_RESET {port_csp0_addr[X313_SEQ_CMD]=X3X3_SEQ_RESET_BITS; X3X3_AFTERWRITE ; X3X3_AFTERWRITE ; X3X3_AFTERWRITE ; X3X3_AFTERWRITE ; }


//     [14] -   reset all FIFO (takes 16 clock pulses), stop i2c until run command (can not be combined with "run")
//     [13] - if 1 - use [12] to run/stop i2c controller (needed to be stopped and confirmed not busy before software i2c)
//     [12] - run/stop i2c (when [13] == 1
//     [11] -   if 1, use [10:9] to set command bytes to send after slave address (0..3)
//     [10:9] - number of bytes to send, valid if [11] is set. For Micron sensors - use 3
//     [8]    - set duration of quarter i2c cycle in system clock cycles - nominal value 100 (0x64) for 160MHz
//     [7:0]  - duration of quarter i2c cycle (applied if [8] is set)


//  #define   X313_WA_USB       0x60  // ... 0x6f
  #define   X313_WA_IOPINS    0x70  // bits [31:24] - enable channels (channel 0 -software, enabled at FPGA init)
  #define     X313_WA_IOPINS_EN_TRIG_OUT  0xc000000
  #define     X313_WA_IOPINS_DIS_TRIG_OUT 0x8000000

/// channel 3 has highest priority, 0 - lowest. If channel wants this pin as output, the (enabled) channel with highest number will get it
/// bits 31,30 control channel 3 (0x00000000, 0x40000000 - no change, 0x80000000 - disable, 0xc0000000 - enable)
/// bits 29,28 control channel 2 (0x00000000, 0x10000000 - no change, 0x20000000 - disable, 0x30000000 - enable) - 10364 board
/// bits 27,26 control channel 1 (0x00000000, 0x04000000 - no change, 0x08000000 - disable, 0x0c000000 - enable) - 10369 board
/// bits 25,24 control channel 0 (0x00000000, 0x01000000 - no change, 0x02000000 - disable, 0x03000000 - enable) - 'software' set/reset
/// lower bits control 'software' settings for the bits (channel 0). They can be disabled by bits 25.24, also other channels overwrite them
/// bits 23,22 contorl IO11:     (0x00000000 - no change,  0x00400000 - out 0, 0x00800000 - out 1 , 0x00c00000 - tristated (input))
/// bits 21,20 contorl IO10:     (0x00000000 - no change,  0x00100000 - out 0, 0x00200000 - out 1 , 0x00300000 - tristated (input))
/// bits 19,18 contorl IO09:     (0x00000000 - no change,  0x00040000 - out 0, 0x00080000 - out 1 , 0x000c0000 - tristated (input))
/// bits 17,16 contorl IO08:     (0x00000000 - no change,  0x00010000 - out 0, 0x00020000 - out 1 , 0x00030000 - tristated (input))
/// bits 15,14 contorl IO07:     (0x00000000 - no change,  0x00004000 - out 0, 0x00008000 - out 1 , 0x0000c000 - tristated (input))
/// bits 13,12 contorl IO06:     (0x00000000 - no change,  0x00001000 - out 0, 0x00002000 - out 1 , 0x00003000 - tristated (input))
/// bits 11,10 contorl IO05:     (0x00000000 - no change,  0x00000400 - out 0, 0x00000800 - out 1 , 0x00000c00 - tristated (input))
/// bits 09,08 contorl IO04:     (0x00000000 - no change,  0x00000100 - out 0, 0x00000200 - out 1 , 0x00000300 - tristated (input))
/// bits 07,06 contorl IO03:     (0x00000000 - no change,  0x00000040 - out 0, 0x00000080 - out 1 , 0x000000c0 - tristated (input))
/// bits 05,04 contorl IO02:     (0x00000000 - no change,  0x00000010 - out 0, 0x00000020 - out 1 , 0x00000030 - tristated (input))
/// bits 03,02 contorl IO01:     (0x00000000 - no change,  0x00000004 - out 0, 0x00000008 - out 1 , 0x0000000c - tristated (input))
/// bits 01,00 contorl IO00:     (0x00000000 - no change,  0x00000001 - out 0, 0x00000002 - out 1 , 0x00000003 - tristated (input))
/// I/O pins
/// Pin number     10369                  10364
/// 00           I2C SCL                 Encoder1A
/// 01           I2C SDA                 Encoder1B
/// 02    Granddaughter board J9         Encoder2A
/// 03    Granddaughter board J9         Encoder2B
/// 04    Granddaughter board J10        Encoder3A
/// 05    Granddaughter board J10        Encoder3B
/// 06    Granddaughter board J8,J11     Motor1A
/// 07    Granddaughter board J8,J11     Motor1B
/// 08    Int sync I/O (optoisolated)    Motor2A
/// 09    Int sync Out (high current)    Motor2B
/// 10    Ext sync I/O (optoisolated)    Motor3A
/// 11    Ext sync Out (high current)    Motor3B

// to eliminate the need for a shadow registers made that a
//    dibit=0 - no change
//    dibit=1 - set en=1, d=0
//    dibit=2 - set en=1, d=1
//    dibit=3 - set en=0, d=0
        // 0 - use bits [25:0] of the control word (0 - data0, 1 - enable out0, 2 - data1, ..., 11 - enable out5)
        // 1 - use channel A (USB)?
        // 2 - use channel B (tbd)
        // 3 - use channel C (tbd)
  #define X313_WA_SENSFPGA  0x74 //Control programming of external FPGA on the sensor/sensor multiplexor board
  #define X313_WA_CAMSYNCTRIG 0x78 // trigger condition, 0 - internal, else dibits ((use<<1) | level) for each GPIO[11:0] pin
  #define X313_WA_CAMSYNCDLY  0x79 // trigger delay, 32 bits in pixel clocks
  #define X313_WA_CAMSYNCOUT  0x7a // trigger output to GPIO, dibits ((use << 1) | level_when_active). Bit 24 - test mode, when GPIO[11:10] are controlled by other internal signals
  #define X313_WA_CAMSYNCPER  0x7b // output sync period (32 bits, in pixel clocks). 0- stop. 1..256 - single, >=256 repetitive with specified period.
  #define SFPGA_TDI_BIT       0x0
  #define SFPGA_TMS_BIT       0x2
  #define SFPGA_TCK_BIT       0x4
  #define SFPGA_PROG_BIT      0x6
  #define SFPGA_PGMEN_BIT     0x8
  #define SFPGA_RD_SENSPGMPIN 0x80000
  #define SFPGA_RD_TDO        0x90000
  #define SFPGA_RD_DONE       0xa0000



        //19:16 - 0xb..0xf - no changes
        // Mulptiplex status signals into a single line
        //      - 0xa - select xfpgadone
        //      - 0x9 - select xfpgatdo
        //      - 0x8 - select senspgmin (default)
        //      - 0x0..0x7 - no changes
        //15:10 - not used
        // 9: 8 - 3 - set xpgmen,
        //      - 2 - reset xpgmen,  
        //      - 0, 1 - no changes to xpgmen
        // 7: 6 - 3 - set xfpgaprog,
        //      - 2 - reset xfpgaprog,  
        //      - 0, 1 - no changes to xfpgaprog
        // 5: 4 - 3 - set xfpgatck,
        //      - 2 - reset xfpgatck,  
        //      - 0, 1 - no changes to xfpgatck
        // 3: 2 - 3 - set xfpgatms,
        //      - 2 - reset xfpgatms,
        //      - 0, 1 - no changes to xfpgatms
        // 1: 0 - 3 - set xfpgatdi,
        //      - 2 - reset xfpgatdi,
        //      - 0, 1 - no changes to xfpgatdi


// NOTE: Important note (issue came out with faster ETRAX FS). Due to write pipe and shared r/w addresses
//       it is not possible to read from registers 0x20.0x2f TWO cycles after any write to FPGA
//       The following dummy read will take 2 bus cycles
#define X3X3_AFTERWRITE {if (!port_csp0_addr[X313__RA__MODEL]) printk ("model=0");} //just to be sure this read will not be optimized out
#define x3x3_DELAY(x) {int iiii; for (iiii=0; iiii < (x); iiii++) X3X3_AFTERWRITE ; }

#define X313_WA_SDCH0_CTL0 0x20   // SDRAM control for Channel 0 - writing causes initialization of the channel
//---353---
//    20   |       | Channel 0 (16 bit data sensor->SDRAM)               | Initialize channel:            |  (mode 0)                         |
//         | 31-16 | not used                                            | 1) set address to startAddres  | 31-30 | nbuf[1:0]                 |
//         |    15 | mode: 0 - 256x1 16-bit (or 512x1 bytes)             | (word at address 01 with other | 29-18 | ntileY[11:0]              |
//         |       |       1 -  18x9 16 bit (or 18x18 bytes)             | startAddress bits should be    | 17-16 | ntileX[8:7]               |
//         |       | mode should be set to 0 for channel 0               | written earlier;               |       |                           |
//         |    14 | WnR - Write/not read. Should be 1 for channel 0     | 2) set nbuf[1:0] to 0          | 15-00 | Same as written           |
//         |    13 | depend. Should be set to 1 if needed to synchronize |                                |       |                           |
//         |       | reading from SDRAM to writing from sensor           |                                |       |                           |
//         | 12-00 | frame start address startAddr[20:8]. Addressed are  |                                |       |                           |
//         |       | 16-bit words, low 8 bits are 0 (aligned to 256 word |                                |       |                           |
//         |       | page boundary                                       |                                |       |                           |
//---333---
//    20   |       | Channel 0 (16 bit data sensor->SDRAM)               | Initialize channel:            |  (mode 0)                         |
//         | 31-16 | not used                                            | 1) set address to startAddres  | 31-30 | nbuf[1:0]                 |
//         |    15 | mode: 0 - 256x1 16-bit (or 512x1 bytes)             | (word at address 01 with other | 29-18 | ntileY[11:0]              |
//         |       |       1 -  18x9 16 bit (or 18x18 bytes)             | startAddress bits should be    | 17-16 | ntileX[8:7]               |
//         |       | mode should be set to 0 for channel 0               | written earlier;               |       |                           |
//         |    14 | WnR - Write/not read. Should be 1 for channel 0     | 2) set nbuf[1:0] to 0          | 15-00 | Same as written           |
//         |    13 | depend. Should be set to 1 if needed to synchronize |                                |       |                           |
//         |       | reading from SDRAM to writing from sensor           |                                |       |                           |
//         |    12 | not used                                            |                                |       |                           |
//         | 11-00 | frame start address startAddr[19:8]. Addressed are  |                                |       |                           |
//         |       | 16-bit words, low 8 bits are 0 (aligned to 256 word |                                |       |                           |
//         |       | page boundary                                       |                                |       |                           |
//---313/323---
// 31-16 | not used                                            | 1) set address to startAddres  |    31 | nbuf (buffer page number)
//    15 | mode: 0 - 128x1 16-bit (or 256x1 bytes)             | (word at address 01 with other |       |   for ping-pong operation
//       |       1 -  16x8 16 bit (or 16x16 bytes)             | startAddress bits should be    | 30-19 | tileY (12 bits) if mode=0
//       | mode should be set to 0 for channel 0               | written earlier;               | 18-16 | tileX ( 2 bits) if mode=0
//    14 | WnR - Write/not read. Should be 1 for channel 0     | 2) set nbuf to 0               |READ 0x21 instead !!!        
//    13 | depend. Should be set to 1 if needed to synchronize |                                | 15-00 | Same as written
//       | reading from SDRAM to writing from sensor           |
//    12 | should be 0                                         |
// 11-00 | frame start address startAddr[18:7]. Addressed are  |
//       | 16-bit words, low 7 bits are 0 (aligned to 128 word |
//       | page boundary                                       |
#define X313_WA_SDCH0_CTL1 0x21   //       | Channel 0 (16 bit data sensor->SDRAM)               | none                           |    31 | nbuf (buffer page number)
//---353---
//    21   |       | Channel 0 (16 bit data sensor->SDRAM)               | none                           | 31-30 | nbuf[1:0]                 |
//         | 31-14 | not used                                            |                                | 29-18 | ntileY[11:0]              |
//         | 13-04 | nTileX[9:0]. For mode=0 nTileX[9:5] specifies number|                                |    17 | 0                         |
//         |       | of the last 8x16 block to read in a line.           |                                |    16 | ntileX[4]                 |
//         |       | Actual number of 8x16 blocks in a line is nTileX+1  |                                | 15-00 | Same as written           |
//         |       | For mode0 last 128x16 may be partial if             |                                |       |                           |
//         |       | nTileX[9:5] != 0x1F                                 |                                |       |                           |
//         | 03-00 | startAddr[24:21] - used with other bits specified   |                                |       |                           |
//         |       | at address 20                                       |                                |       |                           |
//---333---
//    21   |       | Channel 0 (16 bit data sensor->SDRAM)               | none                           | 31-30 | nbuf[1:0]                 |
//         | 31-14 | not used                                            |                                | 29-18 | ntileY[11:0]              |
//         | 13-04 | nTileX[9:0]. For mode=0 nTileX[9:5] specifies number|                                |    17 | 0                         |
//         |       | of the last 8x16 block to read in a line.           |                                |    16 | ntileX[4]                 |
//         |       | Actual number of 8x16 blocks in a line is nTileX+1  |                                | 15-00 | Same as written           |
//         |       | For mode0 last 128x16 may be partial if             |                                |       |                           |
//         |       | nTileX[9:5] != 0x1F                                 |                                |       |                           |
//         | 03-00 | startAddr[23:20] - used with other bits specified   |                                |       |                           |
//         |       | at address 20                                       |                                |       |                           |
//---313/323---
// 31-13 | not used                                            |                                |       |   for ping-pong operation
// 12-04 | nTileX[8:0]. For mode=0 nTileX[8:4] specify number  |                                | 30-19 | tileY (12 bits) if mode=0
//       | of the last 8x16 block to read in a line.           |                                | 18-14 | tileX (4 bits) if mode=0
//       | Actual number of 8x16 blocks in a line is nTileX+1  |                                |       |
//       | For mode0 last 128x16 may be partial if             |                                | 13-00 | Same as written
//       | nTileX[3:0] != 0xF                                  |                                |       |
// 03-00 | startAddr[22:19] - used with other bits specified   |                                |       |
//       | ar address 00                                       |                                |       |
#define X313_WA_SDCH0_CTL2 0x22
//---333---
//    22   |       | Channel 0 (16 bit data sensor->SDRAM)               | none                           | 31-30 | nbuf[1:0]                 |
//         | 31-12 | not used                                            |                                | 29-18 | ntileY[11:0]              |
//         | 11-00 | nTileY[11:0] - (for mode=0) - number of the last    |                                |    17 | 0                         |
//         |       |   line in in a frame. Total number of lines in      |                                | 16-12 | ntileX[4:0]               |
//         |       | a frame is nTileY[11:0]+1 (mode=0) or               |                                | 11-00 | Same as written           |
//         |       | (nTileY[11:0] & 12'hff0)+18 if mode =1              |                                |       |                           |
//         |       | (nTileY >> 4) +1 of rows of 18x9x16 (18x18x8) tiles |                                |       |                           |
//---313/323---
// Channel 0 (16 bit data sensor->SDRAM)                       | none                           | 31-16 | same as for address=20
// 31-11 | not used                                            |                                | 15-00 | same as written
// 10-00 | nTileY[10:0] - (for mode=0) - number of the last    |                                |       |
//       |   line in in a frame. Total number of lines in      |                                |       |
//       | a frame is nTileY[10:0]+1 (mode=0) or               |                                |       |
//       | (nTileY[10:0] & 11'h7f0)+16 if mode =1              |                                |       |
//       | (nTileY >> 4) +1 of rows of 16x8x16 (16x16x8) tiles |                                |       |
#define X313_WA_SD_MANCMD 0x23

//       | Manual command to SDRAM for initialization. Works   | Perform one-cycle SDRAM manual | 31-16 | same as for address=20
//       | (and is recommended) if SDRAM is disabled (see re-  | command by driving             | 15-00 | same as written
//       | gister 27)                                          | RAS,CAS,WE,BA[1:0],A[11/12:0]  |       |
//---333/353 ---
//         | 31-18 | not used                                            |                                |       |                           |
//         |    17 | RAS                                                 |                                |       |                           |
//         |    16 | CAS                                                 |                                |       |                           |
//         |    15 | WE                                                  |                                |       |                           |
//         | 14-13 | bank address [1:0]                                  |                                |       |                           |
//         | 12-00 | address[12:0]                                       |                                |       |                           |
//---313/323---
// 31-16 | not used                                            | according to data bits         |       |
// 15-14 | 00 - write mode reg (RAS=L, CAS=L, WE=L)            |                                |       |
//       | 01 - refresh        (RAS=L, CAS=L, WE=H)            |                                |       |
//       | 10 - precharge      (RAS=L, CAS=H, WE=L)            |                                |       |
//       | 11 - nop            (RAS=H, CAS=H, WE=H)            |                                |       |
// 13-12 | bank address [1:0]                                  |                                |       |
// 11-00 | address[11:0]                                       |                                |       |
#define X313_WA_SDCH1_CTL0 0x24
//---333---
//    24   |       | Channel 1 (16 bit FPN data from SDRAM)              | Initialize channel:            |    31 | nbuf (buffer page number) |
//         | 31-16 | not used                                            | 1) set address to startAddres  |       |   for ping-pong operation |
//         |    15 | mode: 0 - 128x1 16-bit (or 256x1 bytes)             | (word at address 01 with other | 30-20 | tileY (11 bits) if mode=0 |
//         |       |       1 -  16x8 16 bit (or 16x16 bytes)             | startAddress bits should be    | 19-16 | tileX ( 4 bits) if mode=0 |
//         |       | mode should be set to 0 for channel 0               | written earlier;               |       |                           |
//         |    14 | WnR - Write/not read. Should be 0 for channel 1     | 2) set nbuf to 0               | 15-00 | Same as written           |
//         |    13 | depend. Should be set to 0 as FPN data is supposed  | 3) start reading SDRAM as WnR  |       |                           |
//         |       | to be always ready.                                 | is 0 if channel 1 is enabled   |       |                           |
//         |    12 | not used                                            |                                |       |                           |
//         | 11-00 | frame start address startAddr[18:7]. Addressed are  |                                |       |                           |
//         |       | 16-bit words, low 7 bits are 0 (aligned to 128 word |                                |       |                           |
//         |       | page boundary                                       |                                |       |                           |
//---313/323---
//       | Channel 1 (16 bit FPN data from SDRAM)              | Initialize channel:            |    31 | nbuf (buffer page number)
// 31-16 | not used                                            | 1) set address to startAddres  |       |   for ping-pong operation
//    15 | mode: 0 - 128x1 16-bit (or 256x1 bytes)             | (word at address 01 with other | 30-20 | tileY (11 bits) if mode=0
//       |       1 -  16x8 16 bit (or 16x16 bytes)             | startAddress bits should be    | 19-16 | tileX ( 4 bits) if mode=0
//       | mode should be set to 0 for channel 0               | written earlier;               |       |
//    14 | WnR - Write/not read. Should be 0 for channel 1     | 2) set nbuf to 0               | 15-00 | Same as written
//    13 | depend. Should be set to 0 as FPN data is supposed  | 3) start reading SDRAM as WnR  |       |
//       | to be always ready.                                 | is 0 if channel 1 is enabled   |       |
//    12 | not used                                            |                                |       |
// 11-00 | frame start address startAddr[18:7]. Addressed are  |                                |       |
//       | 16-bit words, low 7 bits are 0 (aligned to 128 word |                                |       |
//       | page boundary                                       |                                |       |
#define X313_WA_SDCH1_CTL1 0x25
//---333---
//    25   |       | Channel 1 (16 bit FPN data from SDRAM)              | none                           | 31-16 | same as for address=24    |
//         | 31-14 | not used                                            |                                | 15-00 | same as written           |
//         | 13-04 | nTileX[6:0]. For mode=0 nTileX[6:4] specify number  |                                |       |                           |
//         |       | of the last 8x16 block to read in a line.           |                                |       |                           |
//         |       | Actual number of 8x16 blocks in a line is nTileX+1  |                                |       |                           |
//         |       | For mode0 last 128x16 may be partial if             |                                |       |                           |
//         |       | nTileX[3:0] != 0xF                                  |                                |       |                           |
//         | 03-00 | startAddr[22:19] - used with other bits specified   |                                |       |                           |
//         |       | ar address 24                                       |                                |       |                           |
//---313/323---
//       | Channel 1 (16 bit FPN data from SDRAM)              | none                           | 31-16 | same as for address=24
// 31-13 | not used                                            |                                | 15-00 | same as written
// 12-04 | nTileX[8:0]. For mode=0 nTileX[8:4] specify number  |                                |       |
//       | of the last 8x16 block to read in a line.           |                                |       |
//       | Actual number of 8x16 blocks in a line is nTileX+1  |                                |       |
//       | For mode0 last 128x16 may be partial if             |                                |       |
//       | nTileX[3:0] != 0xF                                  |                                |       |
// 03-00 | startAddr[22:19] - used with other bits specified   |                                |       |
//       | ar address 00                                       |                                |       |
#define X313_WA_SDCH1_CTL2 0x26
//---333---
//    26   |       | Channel 1 (16 bit FPN data from SDRAM)              | none                           | 31-16 | same as for address=24    |
//         | 31-11 | not used                                            |                                | 15-00 | same as written           |
//         | 10-00 | nTileY[10:0] - (for mode=0) - number of the last    |                                |       |                           |
//         |       |   line in in a frame. Total number of lines in      |                                |       |                           |
//         |       | a frame is nTileY[10:0]+1 (mode=0) or               |                                |       |                           |
//         |       | (nTileY[10:0] & 11'h7f0)+16 if mode =1              |                                |       |                           |
//         |       | (nTileY >> 4) +1 of rows of 16x8x16 (16x16x8) tiles |                                |       |                           |
//---313/323---
//       | Channel 1 (16 bit FPN data from SDRAM)              | none                           | 31-16 | same as for address=24
// 31-11 | not used                                            |                                | 15-00 | same as written
// 10-00 | nTileY[10:0] - (for mode=0) - number of the last    |                                |       |
//       |   line in in a frame. Total number of lines in      |                                |       |
//       | a frame is nTileY[10:0]+1 (mode=0) or               |                                |       |
//       | (nTileY[10:0] & 11'h7f0)+16 if mode =1              |                                |       |
//       | (nTileY >> 4) +1 of rows of 16x8x16 (16x16x8) tiles |                                |       |

// with ETRAX FS there was a problem reading this register right after write to 0x20..0x2f range
#define X313_WA_SD_MODE    0x27
/// rev 0x03533022 - now 0x27 used 12-bit input, each internal bit is controlled by an input  dibit (0x - don't change, 10 - reset, 11 - set)
/// reading back 0x27 returns "old" data (non-masked enable bits) in 5:0, and masked (by other factors, like waiting for linked channel)
/// in bits 11:6 
///      {4'h0,enXfer[3:0],enRefresh,enSDRAM,enXfer0[3:0],enRefresh0,enSDRAM0}:

//---333---
//same as 313/323
//---313/323---
//       | Set SDRAM operation mode                            | change SDRAM mode by enabling  | 31-16 | same as for address=24
// 31-06 | not used                                            | channels specified in data     | 15-00 | same as written
//    05 | enXfer[3] - enable SDRAM transfers for Channel 3    |                                |       |
//       |    CPU <-> SDRAM through PIO                        |                                |       |
//    04 | enXfer[2] - enable SDRAM transfers for Channel 2    |                                |       |
//       |    SDRAM -> (compressor) -> CPU through DMA         |                                |       |
//    03 | enXfer[1] - enable SDRAM transfers for Channel 1    |                                |       |
//       |    SDRAM-> sensor data processor (FPN correction)   |                                |       |
//    02 | enXfer[0] - enable SDRAM transfers for Channel 0    |                                |       |
//       |    sensor ->(FPN correction) ->SDRAM                |                                |       |
//    01 | enRefresh - enable SDRAM automatic refresh as a     |                                |       |
//       |    background process                               |                                |       |
//    00 | enSDRAM - enable SDRAM controller. If set to 0      |                                |       |
//       |    only manual commands are allowed (see register   |                                |       |
//       |    at address 03)                                   |                                |       |
#define X313_WA_SDCH2_CTL0 0x28
//---333---
//    28   |       | Channel 2 (8 bit data to JPEG encoder or just DMA)  | Initialize channel:            |    31 | nbuf (buffer page number) |
//         | 31-16 | not used                                            | 1) set address to startAddres  |       |   for ping-pong operation |
//         |    15 | mode: 0 - 128x1 16-bit (or 256x1 bytes) - N/A       | (word at address 01 with other | 30-20 | tileY (11 bits) if mode=0 |
//         |       |       1 -  16x8 16 bit (or 16x16 bytes)  ******     | startAddress bits should be    | 19-16 | tileX ( 4 bits) if mode=0 |
//         |       | mode maybe set to either 0 for sequential access    | written earlier;               |       |                           |
//         |       |  to raw data, or to 1 for JPEG 18x18 MCU access     |                                | 30-24 | tileY ( 7 bits) if mode=1 |
//         |    14 | WnR - Write/not read. Should be 0 for channel 2     | 2) set nbuf to 0               | 23-16 | tileX ( 8 bits) if mode=1 |
//         |    13 | depend. May be set to 1 to force encoder/DMA wait   | 3) start reading SDRAM as WnR  |       |                           |
//         |       | for sensor data available. Should work correctly    | is 0 if channel 2 is enabled   | 15-00 | Same as written           |
//         |       | for sequentional write and tiled read.              |                                |       |                           |
//         |    12 | not used                                            |                                |       |                           |
//         | 11-00 | frame start address startAddr[19:8]. Addressed are  |                                |       |                           |
//         |       | 16-bit words, low 8 bits are 0 (aligned to 256 word |                                |       |                           |
//         |       | page boundary                                       |                                |       |                           |
//---313/323---
//       | Channel 2 (8 bit data to JPEG encoder or just DMA)  | Initialize channel:            |    31 | nbuf (buffer page number)
// 31-16 | not used                                            | 1) set address to startAddres  |       |   for ping-pong operation
//    15 | mode: 0 - 128x1 16-bit (or 256x1 bytes)             | (word at address 01 with other | 30-20 | tileY (11 bits) if mode=0
//       |       1 -  16x8 16 bit (or 16x16 bytes)             | startAddress bits should be    | 19-16 | tileX ( 4 bits) if mode=0
//       | mode maybe set to either 0 for sequential access    | written earlier;               |       |
//       |  to raw data, or to 1 for JPEG 16x16 MCU access     |                                | 30-24 | tileY ( 7 bits) if mode=1
//    14 | WnR - Write/not read. Should be 0 for channel 2     | 2) set nbuf to 0               | 23-16 | tileX ( 8 bits) if mode=1
//    13 | depend. May be set to 1 to force encoder/DMA wait   | 3) start reading SDRAM as WnR  |       |
//       | for sensor data available. Should work correctly    | is 0 if channel 2 is enabled   | 15-00 | Same as written
//       | for sequentional write and tiled read.              |                                |       |
//    12 | not used                                            |                                |       |
// 11-00 | frame start address startAddr[18:7]. Addressed are  |                                |       |
//       | 16-bit words, low 7 bits are 0 (aligned to 128 word |                                |       |
//       | page boundary                                       |                                |       |
#define X313_WA_SDCH2_CTL1 0x29
//---333---
//    29   |       | Channel 2 (8 bit data to JPEG encoder or just DMA)  | none                           |    31 | nbuf[1]                   |
//         | 31-14 | not used                                            |                                |    30 | nbuf[0]                   |
//         | 13-04 | nTileX[9:0]. Mode is supposed to be 1               |                                | 29-22 | ntyleY[11:4]              |
//         |       | Actual number of 8x16 blocks in a line is nTileX+1  |                                | 21-16 | ntileX[9:4]               |
//         |       |                                                     |                                |       |                           |
//         | 03-00 | startAddr[23:20] - used with other bits specified   |                                |       |                           |
//         |       | at address 28                                       |                                |       |                           |
//---313/323---
//       | Channel 2 (8 bit data to JPEG encoder or just DMA)  | none                           | 31-16 | same as for address=28
// 31-13 | not used                                            |                                | 15-00 | same as written
// 12-04 | nTileX[8:0]. For mode=0 nTileX[8:4] specify number  |                                |       |
//       | of the last 8x16 block to read in a line.           |                                |       |
//       | Actual number of 8x16 blocks in a line is nTileX+1  |                                |       |
//       | For mode0 last 128x16 may be partial if             |                                |       |
//       | nTileX[3:0] != 0xF                                  |                                |       |
// 03-00 | startAddr[22:19] - used with other bits specified   |                                |       |
//       | ar address 00                                       |                                |       |
#define X313_WA_SDCH2_CTL2 0x2a
//---333---
//    2a   |       | Channel 2 (8 bit data to JPEG encoder or just DMA)  | none                           |    31 | nbuf[1]                   |
//         | 31-12 | not used                                            |                                |    30 | nbuf[0]                   |
//         | 11-00 | nTileY[11:0] - (for mode=0) - number of the last    |                                | 29-22 | ntyleY[11:4]              |
//         |       |   line in in a frame. Total number of lines in      |                                | 21-12 | ntileX[9:0]               |
//         |       | a frame is nTileY[10:0]+1 (mode=0) or               |                                | 11-00 | Same as written           |
//         |       | (nTileY[10:0] & 11'h7f0)+16 if mode =1              |                                |       |                           |
//         |       | (nTileY >> 4) +1 of rows of 16x8x16 (16x16x8) tiles |                                |       |                           |
//---313/323---
//       | Channel 2 (8 bit data to JPEG encoder or just DMA)  | none                           | 31-16 | same as for address=28
// 31-11 | not used                                            |                                | 15-00 | same as written
// 10-00 | nTileY[10:0] - (for mode=0) - number of the last    |                                |       |
//       |   line in in a frame. Total number of lines in      |                                |       |
//       | a frame is nTileY[10:0]+1 (mode=0) or               |                                |       |
//       | (nTileY[10:0] & 11'h7f0)+16 if mode =1              |                                |       |
//       | (nTileY >> 4) +1 of rows of 16x8x16 (16x16x8) tiles |                                |       |
#define X313_WA_SDCH3_CTL0 0x2c
//---333---
//    2c   |       |  Channel 3 (CPU PIO <-> SDRAM)                      | Initialize channel:            |    (mode 0)                       |
//         | 31-16 | not used                                            | 1) set address to startAddres  | 31-30 | nbuf[1:0]                 |
//         |    15 | mode: 0 - 256x1 16-bit (or 512x1 bytes)             | (word at address 01 with other | 29-18 | ntileY[11:0]              |
//         |       |       1 -  18x9 16 bit (or 18x18 bytes)             | startAddress bits should be    | 17-16 | ntileX[8:7]               |
//         |       | mode should be set to 0 for channel 0               | written earlier;               |       |                           |
//         |    14 | WnR - Write/not read. Should be 1 for channel 0     | 2) set nbuf[1:0] to 0          | 15-00 | Same as written           |
//         |    13 | depend. Should be set to 1 if needed to synchronize |                                |       |                           |
//         |       | reading from SDRAM to writing from sensor           |                                |       |                           |
//         |    12 | not used                                            |                                |       |                           |
//         | 11-00 | frame start address startAddr[19:8]. Addressed are  |                                |       |                           |
//         |       | 16-bit words, low 8 bits are 0 (aligned to 256 word |                                |       |                           |
//         |       | page boundary                                       |                                |       |                           |
//---313/323---
//       |  Channel 3 (CPU PIO <-> SDRAM)                      | Initialize channel:            |    31 | nbuf (buffer page number)
// 31-16 | not used                                            | 1) set address to startAddres  |       |   for ping-pong operation
//    15 | mode: 0 - 128x1 16-bit (or 256x1 bytes)             | (word at address 01 with other | 30-20 | tileY (11 bits) if mode=0
//       |       1 -  16x8 16 bit (or 16x16 bytes)             | startAddress bits should be    | 19-16 | tileX ( 4 bits) if mode=0
//       | mode maybe set to either 0 for sequential access    | written earlier;               |       |
//       |  to raw data, or to 1 16x16x8 tiled                 |                                | 30-24 | tileY ( 7 bits) if mode=1
//    14 | WnR - Write/not read. May be set to any value for   | 2) set nbuf to 0               | 23-16 | tileX ( 8 bits) if mode=1
//       | Channel 3 - it works both ways                      |                                |       |
//    13 | depend. May be set to 1 to force the channel wait   | 3) start reading SDRAM if WnR  |       |
//       | for sensor data available. (Read through status)    | is 0 and channel 3 is enabled  | 15-00 | Same as written
//    12 | read DC coimponents instead of SDRAM                |                                |       |
// 11-00 | frame start address startAddr[18:7]. Addressed are  |                                |       |
//       | 16-bit words, low 7 bits are 0 (aligned to 128 word |                                |       |
//       | page boundary                                       |                                |       |
#define X313_WA_SDCH3_CTL1 0x2d
//---333---
//    2d   |       |  Channel 3 (CPU PIO <-> SDRAM)                      | none                           | 31-30 | nbuf[1:0]                 |
//         | 31-14 | not used                                            |                                | 29-18 | ntileY[11:0]              |
//         | 13-04 | nTileX[9:0]. For mode=0 nTileX[9:5] specifies number|                                |    17 | 0                         |
//         |       | of the 8x16 blocks to read in a line.               |                                |    16 | ntileX[4]                 |
//         |       | Actual number of 8x16 blocks in a line is nTileX+1  |                                | 15-00 | Same as written           |
//         |       | For mode0 last 256x16 may be partial if             |                                |       |                           |
//         |       | nTileX[9:5] != 0x00                                 |                                |       |                           |
//         | 03-00 | startAddr[23:20] - used with other bits specified   |                                |       |                           |
//         |       | at address 2c                                       |                                |       |                           |
//---313/323---
//       |  Channel 3 (CPU PIO <-> SDRAM)                      | none                           | 31-16 | same as for address=2c
// 31-13 | not used                                            |                                | 15-00 | same as written
// 12-04 | nTileX[8:0]. For mode=0 nTileX[8:4] specify number  |                                |       |
//       | of the last 8x16 block to read in a line.           |                                |       |
//       | Actual number of 8x16 blocks in a line is nTileX+1  |                                |       |
//       | For mode0 last 128x16 may be partial if             |                                |       |
//       | nTileX[3:0] != 0xF                                  |                                |       |
// 03-00 | startAddr[22:19] - used with other bits specified   |                                |       |
//       | ar address 00                                       |                                |       |
#define X313_WA_SDCH3_CTL2 0x2e
//---333---
//    2e   |       |  Channel 3 (CPU PIO <-> SDRAM)                      | none                           | 31-30 | nbuf[1:0]                 |
//         | 31-12 | not used                                            |                                | 29-18 | ntileY[11:0]              |
//         | 11-00 | nTileY[11:0] - (for mode=0) - number of the last    |                                |    17 | 0                         |
//         |       |   line in in a frame. Total number of lines in      |                                | 16-12 | ntileX[4:0]               |
//         |       | a frame is nTileY[11:0]+1 (mode=0) or               |                                | 11-00 | Same as written           |
//         |       | (nTileY[11:0] & 12'hff0)+18 if mode =1              |                                |       |                           |
//         |       | (nTileY >> 4) +1 of rows of 18x9x16 (18x18x8) tiles |                                |       |                           |
//---313/323---
//       |  Channel 3 (CPU PIO <-> SDRAM)                      | none                           | 31-16 | same as for address=2c
// 31-11 | not used                                            |                                | 15-00 | same as written
// 10-00 | nTileY[10:0] - (for mode=0) - number of the last    |                                |       |
//       |   line in in a frame. Total number of lines in      |                                |       |
//       | a frame is nTileY[10:0]+1 (mode=0) or               |                                |       |
//       | (nTileY[10:0] & 11'h7f0)+16 if mode =1              |                                |       |
//       | (nTileY >> 4) +1 of rows of 16x8x16 (16x16x8) tiles |                                |       |
#define X313_WA_SDPIO_NEXT  0x2f
//---333---
// similar to 313/333 but will open access to next 128x32bit page, resets address inside page
//---313/323---
//       |  Channel 3 (CPU PIO <-> SDRAM)                      | Next page (8x32 bit window)    | 31-16 | same as for address=2c
// 31-00 | not used                                            | for CPU PIO access. Check      | 15-00 | same as written
//       |                                                     | status bit to see ifbuffer is  |       |
//       |                                                     | ready                          |       |
#define X313_WA_SD_PIOWIN  0x30

#define X313_WA_LENSCORR   0x31

#define X313_LENS_AX(x)           (     0x0 | ((x)&  0x7ffff))
#define X313_LENS_AY(x)           ( 0x80000 | ((x)&  0x7ffff))
#define X313_LENS_C(x)            (0x100000 | ((x)&  0x7ffff))
#define X313_LENS_BX(x)           (0x200000 | ((x)& 0x1fffff))
#define X313_LENS_BY(x)           (0x400000 | ((x)& 0x1fffff))
#define X313_LENS_SCALES(color,x) (0x600000 | (((color)&3)<<17) | ((x)& 0x1ffff))
#define X313_LENS_FATZERO_IN(x)   (0x680000 | ((x)& 0xffff))
#define X313_LENS_FATZERO_OUT(x)  (0x690000 | ((x)& 0xffff))
#define X313_LENS_POSTSCALE(x)    (0x6a0000 | ((x)& 0x7))


/*!*******************************************************************************************************
*! Moved all references to FPGA access to memory-control registers here to simplify code maitenance
*! when FPGA changes.
*!
*! Split SDARM channel initailization in 3 macros
*! X313_PREINIT_SDCHAN(num,mode,wnr,dep,sa,ntilex,ntiley) - writes two (of 3) registers (not yet starting the channel)
*!                                                          returns value foo the 3-rd (command) register
*! X313_PREINIT_SDCHAN(num,cmd)                           - writes the channel command register, starting it
*!                                                          waits 2 cycles after (if ETRAX FS) to make next reads safe
*! X313_INIT_SDCHAN(num,mode,wnr,dep,sa,ntilex,ntiley)    - combination of the 2 above, works as before
*!
*!/ added X3X3_AFTERWRITE to be able to read FPGA after that macro (w/o -= failed in ETRAX FS)
*/

#define RD_SD_PIOWIN port_csp4_addr[X313_WA_SD_PIOWIN]

/// X313_PREINIT_SDCHAN and X313_POSTINIT_SDCHAN are used only here
/// X3X3_AFTERWRITE ised here and in fpga_io.c
#define X313_PREINIT_SDCHAN(num,mode,wnr,dep,sa,ntilex,ntiley) \
          ((port_csp0_addr[X313_WA_SDCH0_CTL0+1+((num)<<2)] = (0x8000 | (((ntilex) & 0x3ff) << 4) | (((sa) >> 21) & 0xf))), \
           (port_csp0_addr[X313_WA_SDCH0_CTL0+2+((num)<<2)] =            ((ntiley) & 0x3fff)), \
           (((mode) & 1) << 15) | (((wnr) & 1) << 14) | (((dep) & 1) << 13) | (((sa) >> 8) & 0x1fff))
#define X313_POSTINIT_SDCHAN(num,cmd) {port_csp0_addr[X313_WA_SDCH0_CTL0+0+((num)<<2)] = cmd;   X3X3_AFTERWRITE}

#define X313_INIT_SDCHAN(num,mode,wnr,dep,sa,ntilex,ntiley) \
        {X313_POSTINIT_SDCHAN ( num, X313_PREINIT_SDCHAN ( num,mode,wnr,dep,sa,ntilex,ntiley ))}

/// data to be programmed to the memory registers through the sequencer (all arguments are the same, not all are used for all registers)
#define X313_SDCHAN_REG0(mode,wnr,dep,sa,ntilex,ntiley) ((((mode) & 1) << 15) | (((wnr) & 1) << 14) | (((dep) & 1) << 13) | (((sa) >> 8) & 0x1fff))
#define X313_SDCHAN_REG1(mode,wnr,dep,sa,ntilex,ntiley) (0x8000 | (((ntilex) & 0x3ff) << 4) | (((sa) >> 21) & 0xf))
#define X313_SDCHAN_REG2(mode,wnr,dep,sa,ntilex,ntiley) ((ntiley) & 0x3fff)
/// mode=1, wnr=1 - photofinish


#define X313_CHN_EN_D(x)   ((0x30 << (((x)& 3)<<1)) | 0xf) // enable selected channel (and memory+refresh, just in case too)
#define X313_CHN_DIS_D(x)   (0x20 << (((x)& 3)<<1)) // disable selected channel
#define X313_CHN_DISALL_D  0xaa0 // disable all channels, do not modify refresh or memory as a whole
#define X313_SDRAM_OFF_D   0xaaa // disable all channels, refresh and memory itself
#define X313_SDRAM_ON_D    0xaaf // will disable all channels but refresh

#define X313_CHN_EN(x)   {X3X3_AFTERWRITE ; port_csp0_addr[X313_WA_SD_MODE] = X313_CHN_EN_D(x); } // added | 0xf so SDRAM will be enabled too
#define X313_CHN_DIS(x)  {X3X3_AFTERWRITE ; port_csp0_addr[X313_WA_SD_MODE] = X313_CHN_DIS_D(x); }
#define X313_CHN_DISALL  {X3X3_AFTERWRITE ; port_csp0_addr[X313_WA_SD_MODE] = X313_CHN_DISALL_D ; }
#define X313_SDRAM_OFF   {port_csp0_addr[X313_WA_SD_MODE]= X313_SDRAM_OFF_D;} // no need to wait here
#define X313_SDRAM_ON    {port_csp0_addr[X313_WA_SD_MODE]= X313_SDRAM_ON_D;} // no need to wait here - will disable all channels but refresh

#define X313_IS_SDRAM_ON (port_csp0_addr[X313_WA_SD_MODE],((port_csp0_addr[X313_WA_SD_MODE] & 3)==3)) // first - dummy read
#define X313_CHN0_BOUND  (port_csp0_addr[X313_WA_SDCH2_CTL0],(port_csp0_addr[X313_WA_SDCH2_CTL0] & 0x2000))

#define X313_XFERCNTR (port_csp0_addr[X313__RA__XFERCNTR],port_csp0_addr[X313__RA__XFERCNTR])
#define X313_HIGHFREQ (port_csp0_addr[X313__RA__HIGHFREQ],port_csp0_addr[X313__RA__HIGHFREQ])
#define X313_IRQSTATE (port_csp0_addr[0x11],port_csp0_addr[0x11])
#define X313_IOPINS      (port_csp0_addr[X313__RA__IOPINS],port_csp0_addr[X313__RA__IOPINS])

//!X313_WA_SDCH0_CTL2 bits 12..15 are coming from dynamic register and do not depend on the written data
/// used in sensor_common.c
#define X313_CHN0_USED   ( (port_csp0_addr[X313_WA_SDCH0_CTL1]& 0xffff), \
                          (port_csp0_addr[X313_WA_SDCH0_CTL1] & 0xffff) )

/// used only here?
#define X313_CHN0_SET_USED     { port_csp0_addr[X313_WA_SDCH0_CTL1]=0x8000; }
#define X313_CHN0_SET_UNUSED   { port_csp0_addr[X313_WA_SDCH0_CTL1]=0;      }


#define X313_SET_FPGA_TIME(x,y) { port_csp0_addr[X313_WA_RTC_USEC]= ( y ); port_csp0_addr[X313_WA_RTC_SEC]= ( x ); } 
#define X313_GET_FPGA_TIME(x,y) { port_csp0_addr[X313_WA_RTC_LATCH]= 0; X3X3_AFTERWRITE ; x = port_csp0_addr[X313_WA_RTC_SEC]; y = port_csp0_addr[X313_WA_RTC_USEC];} 
#define X313_GET_FPGA_SECONDS(x) { port_csp0_addr[X313_WA_RTC_LATCH]= 0; X3X3_AFTERWRITE ; x = port_csp0_addr[X313_WA_RTC_SEC]} 



// As I will use application, not driver to access frame memory, I'll put here SDRAM memory map
// The window size is designed to fit 1280x1024x16 - FPN, twice as much for image
 #define X313_MAP_FPN   0
 #define X313_SDRAM_SIZE 0x4000000
 #define X313_MAXWIDTH  4096 // multiple of 128
 #define X313_MAXHEIGHT 16384 // multiple of 16 - unsafe - not enough room for black level subtraction
 #define X313_MAXHEIGHT_SAFE 4096 // multiple of 16  OK for black level subtraction TODO: disable black level if unsafe
 #define X313_MAP_FRAME ((X313_MAP_FPN) + (X313_MAXWIDTH) * (X313_MAXHEIGHT_SAFE))
 #define X313_MARGINS 4
 #define X313_TILEHOR 16
 #define X313_TILEVERT 16
