/*****************************************************************************
*!
*! FILE NAME  : head.S
*!
*! DESCRIPTION: Taken from  arch/cris/arch-v32/boot/rescue/head.S,
*! from Linux 2.6.19 (IR2_6_19-3). Modified for (NAND) boot loader.
*!
*! ---------------------------------------------------------------------------
*! (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
*!
*! This program is free software; you can redistribute it and/or modify
*! it under the terms of the GNU General Public License version 2 as 
*! published by the Free Software Foundation.
*!
*! This program is distributed in the hope that it will be useful,
*! but WITHOUT ANY WARRANTY; without even the implied warranty of
*! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*! GNU General Public License for more details.
*!
*! You should have received a copy of the GNU General Public License
*! along with this program; if not, write to the Free Software
*! Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*!
*!***************************************************************************/

#ifdef INCLUDE_FWINFO
#include "fwinfo_data.h" /* local version, originally in libs/fwinfo */
#endif
#include "bootconfig.h"
#include <asm/arch/hwregs/asm/reg_map_asm.h>
#include <asm/arch/hwregs/asm/config_defs_asm.h>

#include <asm/arch/hwregs/asm/ser_defs_asm.h>
#include <asm/arch/hwregs/asm/pinmux_defs_asm.h>

;; Debugging. Normally all these are set to 0.

;; set to 1 to include debug messages on LEDs
;; don't forget to configure LED port according to product in use
#define LEDS (0)

;;  set to 1 to implement binary counter on LEDs (for testing)
#define LEDTEST (0)

;; set to 1 to implement binary counter on leds after copying from nand,
;; instead of booting
#define FLASH_LEDS_INSTEAD_OF_BOOT (0)

;; set to 1 to dump internal RAM to serial port
#define SERIAL_DUMP (0)

;; select serial port for dump, 0 or 1 supported
#define SERIAL_PORT (0)

;; set to 1 to configure and enable serial receiver (when serial dump enabled)
#define SERIAL_RECEIVE (0)

#if LEDS
#include <asm/arch/hwregs/asm/gio_defs_asm.h>
#include <asm/arch/hwregs/asm/pinmux_defs_asm.h>
#include <asm/arch/hwregs/asm/bif_core_defs_asm.h>
#endif

#if SERIAL_DUMP
#include <asm/arch/hwregs/asm/ser_defs_asm.h>
#include <asm/arch/hwregs/asm/pinmux_defs_asm.h>
#endif


#if (SERIAL_PORT == 0)
#define regi_serial regi_ser0
#endif
#if (SERIAL_PORT == 1)
#define regi_serial regi_ser1
#endif

;; Macros

#if LEDS
.macro	SAY x
	orq 31, $r9
	and.d ~(\x), $r9
	move.d $r9, [$r8]
.endm
#else
.macro	SAY x
	;; nothing
.endm
#endif

#if SERIAL_DUMP
.macro	DISPLAY x
	move.d \x, $r6 	; save value
	moveq 28, $r5 ; counter / shift amount
8:
	move.d $r6, $r3 ; fetch value
	bsr nybbleout
	lsr.d $r5, $r3 ; shift nybble we want (delay slot)
	subq 4, $r5  ; count down bits
	bpl 8b ; loop
	nop
	bsr serout
	moveq 13, $r3 ; delay slot
	bsr serout
	moveq 10, $r3 ; delay slot
.endm
#else
.macro 	DISPLAY x
	;; nothing
.endm
#endif


;; Code

	.text
start:

	;; TODO: Add code for Ronny's search-for-good-block boot rom algorithm
	
	;; Start clocks for used blocks.
	move.d REG_ADDR(config, regi_config, rw_clk_ctrl), $r1
	move.d [$r1], $r0
	or.d   REG_STATE(config, rw_clk_ctrl, cpu, yes) | \
	       REG_STATE(config, rw_clk_ctrl, bif, yes) | \
	       REG_STATE(config, rw_clk_ctrl, fix_io, yes), $r0
	move.d $r0, [$r1]

;;------------==================--------------||

;; init port ser0

#undef _DEBUG

#ifdef _DEBUG
   ;; rw_xoff: chr and automatic = 0
   move.d REG_ADDR(ser, regi_ser0, rw_xoff), $r1
   move.d [$r1], $r0
   and.d ~(REG_MASK(ser, rw_xoff, automatic) | REG_MASK(ser, rw_xoff, chr)), $r0
   move.d $r0, [$r1]

   ;; tr control
   move.d REG_ADDR(ser, regi_ser0, rw_tr_ctrl), $r1
   move.d [$r1], $r0
   and.d ~(REG_MASK(ser, rw_tr_ctrl, base_freq) | REG_MASK(ser, rw_tr_ctrl, stop_bits)), $r0
   or.d REG_STATE(ser, rw_tr_ctrl, stop_bits, bits2) | REG_STATE(ser, rw_tr_ctrl, base_freq, f29_493) | REG_STATE(ser, rw_tr_ctrl, en, yes), $r0
   move.d $r0, [$r1]

   ;; tr baud
   move.d REG_ADDR(ser, regi_ser0, rw_tr_baud_div), $r1
   move.d [$r1], $r0
   move.w (29493000 / 8) / 115200, $r0
   move.d $r0, [$r1]
#endif



;; starting debug - serial port is already configured in network bootloader
;; Andrey - some debugging - send "6"
;   movu.b 0x2a , $r3
;   move.d REG_ADDR(ser, regi_ser0, rs_stat_din), $r1
;   move.d REG_ADDR(ser, regi_ser0, rw_dout), $r2
;2:
;   move.d [$r1], $r4
;   btstq REG_BIT(ser, rs_stat_din, tr_rdy), $r4
;   bpl 2b
;   nop
;   move.d $r3, [$r2] ; delay slot
;   nop
;; end of Andrey - some debugging  - send "*"

;;------------==================--------------||


#if LEDS
	;; set up for led control on PB 0..4
	move.d REG_ADDR(gio, regi_gio, rw_pb_dout), $r8 ; output reg
	move.d [$r8], $r9 ; shadow

    	;; set up pinmux
	move.d REG_ADDR(pinmux, regi_pinmux, rw_pb_gio), $r2
	move.d [$r2], $r3
        orq 31, $r3
	move.d $r3, [$r2]

	;; set up GPIO
	;; set to outputs
	move.d REG_ADDR(gio, regi_gio, rw_pb_oe), $r2
	move.d [$r2], $r3
	orq 31, $r3
	move.d $r3, [$r2]


#if LEDTEST
       	;; led test

	moveq 0, $r1 ; led 5-bit binary counter
1:
	or.d 31, $r9
	xor $r1, $r9
	move.d $r9, [$r8]
	addq 1, $r1

	move.d 100000000, $r2 ; delay loop (100e6 => 1s @200Mc)
2:
	bne 2b
	subq 1, $r2

	ba 1b ; loop
	nop
#endif

#endif


check_nand_boot:
	;; Check boot source by checking highest nybble of PC:
	;; If we're running at address 0x0XXXXXXX, we're in flash/eprom/sram
	;; If we're running at address 0x38000000, we're in internal RAM,
	;; so we're most likely coming from NAND.
	;; If we're running at address 0x40000000, we're in SDRAM,
	;; so we've most likely been started by some sort of bootstrapper
	;; e.g. fsboot, which in turn implies NAND, else we would be booting
	;; normally at 0x0XXXXXXX

	SAY 1

here:
	lapcq	., $r0		; get PC
	sub.d	(here-start), $r0 ; offset from here
	beq	normal_boot	; running at address 0 - normal boot
	move.d	$r0, $r13	; save offset for later (unused delay slot)
	lsrq	28, $r0		; get highest nybble

	SAY 2

	;; Prepare to copy 128KB of the NAND flash to internal RAM
	move.d	0x38000800, $r10 ; dest in internal RAM
	move.d	0x1f800, $r12	; #bytes
;	move.d	0x00800, $r12	; #bytes
	cmpq	4, $r0		; running in RAM => started by fsboot
	bhs	copy_from_ram
	movu.w 0x0800, $r11	; source in flash (byte address) (DELAY SLOT)

#ifdef DEBUG_
 movu.b 0x2a , $r3
 move.d REG_ADDR(ser, regi_ser0, rs_stat_din), $r1
 move.d REG_ADDR(ser, regi_ser0, rw_dout), $r2
3:
 move.d [$r1], $r4
 btstq REG_BIT(ser, rs_stat_din, tr_rdy), $r4
 bpl 3b
 nop
 move.d $r3, [$r2] ; delay slot
 nop
#endif

#include "./nand_init.S"

#ifdef DEBUG_
 movu.b 0x30 , $r3
 move.d REG_ADDR(ser, regi_ser0, rs_stat_din), $r1
 move.d REG_ADDR(ser, regi_ser0, rw_dout), $r2
4:
 move.d [$r1], $r4
 btstq REG_BIT(ser, rs_stat_din, tr_rdy), $r4
 bpl 4b
 nop
 move.d $r3, [$r2] ; delay slot
 nop
#endif

#if LEDS
	;; must set up registers again (clobbered by nand_init)
	move.d REG_ADDR(gio, regi_gio, rw_pb_dout), $r8 ; output reg
	move.d [$r8], $r9 ; shadow
#endif

	SAY 3

	ba	copy_complete
	nop

        ; Since the code above has to reside within the first 256 bytes of 
        ; NAND flash, verify that the code so far hasn't gone past this
	; limit. If you're considering removing this, you haven't 
	; properly understood the problem; see nand_init.S for details.
	.org PAGE_SIZE_ADDRESSES 	; from nand_init.S
	.org PAGE_SIZE_BYTES 		; from nand_init.S

#ifdef INCLUDE_FWINFO
	; Image info section
	; Could be done more elegantly with a separate file and its
	; own section and a change to the linker script, but why bother
	; for so little data

	.org FWINFO_LOCATION

;	.int FWINFO_MAGIC_UNINITIALIZED
	INITIALIZE_FWINFO_AREA
	.org FWINFO_LOCATION + FWINFO_MAXSIZE, 0
#endif

normal_boot:
 	SAY 4

	;; Normal NOR boot
	move.d	CONFIG_ETRAX_PTABLE_SECTOR, $r10
	jump	$r10 			; Jump to decompresser 
	nop
				
copy_from_ram:
	add.d	$r13, $r11		; mem offs + src offs -> src addr
1:
	move.d	[$r11+], $r0		; read source
	subq	4, $r12			; 4 bytes at a time
	bne	1b			; loop until done
	move.d	$r0, [$r10+]		; write dest (DELAY SLOT)
	jump 	copy_complete		; jump to internal RAM
	nop				; delay slot


copy_complete:	

#ifdef DEBUG_
 movu.b 0x31 , $r3
 move.d REG_ADDR(ser, regi_ser0, rs_stat_din), $r1
 move.d REG_ADDR(ser, regi_ser0, rw_dout), $r2
4:
 move.d [$r1], $r4
 btstq REG_BIT(ser, rs_stat_din, tr_rdy), $r4
 bpl 4b
 nop
 move.d $r3, [$r2] ; delay slot
 nop
#endif


	SAY 5

#if FLASH_LEDS_INSTEAD_OF_BOOT

flash:

       	;; led test

	moveq 0, $r1 ; led binary counter
1:
	or.d 31, $r9
	xor $r1, $r9
	move.d $r9, [$r8]
	addq 1, $r1

	move.d 100000000, $r2 ; delay loop (100e6 => 1s)
2:
	bne 2b
	subq 1, $r2

	ba 1b ; loop forever
	nop
#endif

#ifdef DEBUG_
 movu.b 0x32 , $r3
 move.d REG_ADDR(ser, regi_ser0, rs_stat_din), $r1
 move.d REG_ADDR(ser, regi_ser0, rw_dout), $r2
3:
 move.d [$r1], $r4
 btstq REG_BIT(ser, rs_stat_din, tr_rdy), $r4
 bpl 3b
 nop
 move.d $r3, [$r2] ; delay slot
 nop
#endif

#if SERIAL_DUMP
	;; dump memory to serial port

#if (SERIAL_PORT == 1)
	;; set up serial-1 pins (serial-0 are set up from reset)
	move.d REG_ADDR(pinmux, regi_pinmux, rw_hwprot), $r1
	move.d [$r1], $r0
	or.d REG_STATE(pinmux, rw_hwprot, ser1, yes), $r0
	move.d $r0, [$r1]
#endif

 	;; rw_xoff: chr and automatic = 0
	move.d REG_ADDR(ser, regi_serial, rw_xoff), $r1
	move.d [$r1], $r0
	and.d ~(REG_MASK(ser, rw_xoff, automatic) | REG_MASK(ser, rw_xoff, chr)), $r0
	move.d $r0, [$r1]

	;; tr control
	move.d REG_ADDR(ser, regi_serial, rw_tr_ctrl), $r1
	move.d [$r1], $r0
	and.d ~(REG_MASK(ser, rw_tr_ctrl, base_freq) | REG_MASK(ser, rw_tr_ctrl, stop_bits)), $r0
	or.d REG_STATE(ser, rw_tr_ctrl, stop_bits, bits2) | REG_STATE(ser, rw_tr_ctrl, base_freq, f29_493) | REG_STATE(ser, rw_tr_ctrl, en, yes), $r0
	move.d $r0, [$r1]

	;; tr baud
	move.d REG_ADDR(ser, regi_serial, rw_tr_baud_div), $r1
	move.d [$r1], $r0
	move.w (29493000 / 8) / 115200, $r0
	move.d $r0, [$r1]

#if SERIAL_RECEIVE
	;; rec control
	move.d REG_ADDR(ser, regi_serial, rw_rec_ctrl), $r1
	move.d [$r1], $r0
	and.d ~(REG_MASK(ser, rw_rec_ctrl, base_freq)), $r0
	or.d REG_STATE(ser, rw_rec_ctrl, base_freq, f29_493) | REG_STATE(ser, rw_tr_ctrl, en, yes), $r0
	move.d $r0, [$r1]

	;; rec baud
	move.d REG_ADDR(ser, regi_serial, rw_rec_baud_div), $r1
	move.d [$r1], $r0
	move.w (29493000 / 8) / 115200, $r0
	move.d $r0, [$r1]
#endif

	;; dump memory

	move.d 0x38000000, $r5 ; pointer

	bsr serout
	moveq 13, $r3
	bsr serout
	moveq 10, $r3
	bsr serout
	moveq 10, $r3

1:
	movu.b [$r5+], $r3 ; get value
	move.d $r3, $r6 ; save

	bsr nybbleout
	lsrq 4, $r3 ; high nybble (delay slot)

	move.d $r6, $r3 ; restore
	bsr nybbleout
	andq 15, $r3 ; delay slot

	movu.b 32, $r3
	bsr serout
	nop

	move.d $r5, $r3
	andq 15, $r3
	bne 1b
	nop

	bsr serout
	moveq 13, $r3 ; delay slot
	bsr serout 
	moveq 10, $r3 ; delay slot

	ba 1b ; loop forever
	nop

nybbleout:
	cmpq 10, $r3
	blo 1f
	addq 48, $r3 ; delay slot
	addq 7, $r3
1:
serout:
	move.d REG_ADDR(ser, regi_serial, rs_stat_din), $r1
	move.d REG_ADDR(ser, regi_serial, rw_dout), $r2
2:
	move.d [$r1], $r4
	btstq REG_BIT(ser, rs_stat_din, tr_rdy), $r4
	bpl 2b
	nop
	ret
	move.d $r3, [$r2] ; delay slot

#endif

;; Init DRAM

#ifdef DEBUG_
 movu.b 0x37 , $r3
 move.d REG_ADDR(ser, regi_ser0, rs_stat_din), $r1
 move.d REG_ADDR(ser, regi_ser0, rw_dout), $r2
3:
 move.d [$r1], $r4
 btstq REG_BIT(ser, rs_stat_din, tr_rdy), $r4
 bpl 3b
 nop
 move.d $r3, [$r2] ; delay slot
 nop
#endif

#include "dram_init.S"
	move.d	RAM_INIT_MAGIC, $r8 	; tell kernel boot loader dram init'd
	move.d	NAND_BOOT_MAGIC, $r12	; booted from NAND flash

;; TODO: Clear .bss (not needed yet because size of .bss is zero)
;; TODO: Change .ld script so that BSS is in DRAM?

;; Ok, time to do something. Continue with boot loader in C.
;; Must set up minimal C environment first though.

	move.d	0x38020000, $sp ; stack pointer at top of internal RAM

	move.d	bootload, $r10
	jump	$r10 			; Jump to boot loader
	nop
