/*****************************************************************************
*!
*! FILE NAME  : fwinfo_data.h
*!
*! DESCRIPTION: Data structures and definitions for firmware image info.
*!
*! ---------------------------------------------------------------------------
*! (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
*!
*! This program is free software; you can redistribute it and/or modify
*! it under the terms of the GNU General Public License as published by
*! the Free Software Foundation; either version 2 of the License, or
*! (at your option) any later version.
*!
*! This program is distributed in the hope that it will be useful,
*! but WITHOUT ANY WARRANTY; without even the implied warranty of
*! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*! GNU General Public License for more details.
*!
*! You should have received a copy of the GNU General Public License
*! along with this program; if not, write to the Free Software
*! Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*!
*!***************************************************************************/

#ifndef _FWINFO_DATA_H
#define _FWINFO_DATA_H

#define FWINFO_LOCATION 2048 /* after first NAND page in small chips */
/*#define FWINFO_LOCATION 4096*/ /* after first NAND page in small chips */
/*#define FWINFO_LOCATION 2048*/ /* after first NAND page in small chips */
/*#define FWINFO_LOCATION (2048 * 4)*/ /* after first NAND page in small chips */
#define FWINFO_MAXSIZE 128 /* arbitrary, reasonable */

#define FWINFO_MAGIC_UNINITIALIZED 0xBFD0573A
#define FWINFO_MAGIC_INITIALIZED 0xA930EBF3

#define FWINFO_MAJOR 1
#define FWINFO_MINOR 1

#define FWINFO_MEDIA_TYPE_NORFLASH 1
#define FWINFO_MEDIA_TYPE_NANDFLASH 2

#define FWINFO_HWID_LENGTH 8

#ifndef __ASSEMBLY__

struct fwinfo {
	unsigned long magic;
	unsigned long major;
	unsigned long minor;
	unsigned long bootblkpar_flash_addr;
	unsigned long bootblkpar_flash_end;
	unsigned long ptable_flash_addr;
	unsigned long ptable_img_addr;
	unsigned long media_type;
	unsigned long media_erasesize;
	unsigned long media_pagesize;
	unsigned long media_totalsize_mb;
	char hwids[FWINFO_HWID_LENGTH]; /* space padded */
};

#else

/* to use macro, -D[efine] constants on command line (or elsewhere),
 * then call macro at appropriate place in program.
 */
.macro	INITIALIZE_FWINFO_AREA
	.dword	FWINFO_MAGIC_INITIALIZED
	.dword	FWINFO_MAJOR
	.dword 	FWINFO_MINOR
	.dword	FWINFO_BBP_ADDR
	.dword	FWINFO_BBP_END
	.dword	FWINFO_PTABLE_MEDIA_ADDR
	.dword	FWINFO_PTABLE_IMG_ADDR
	.dword	FWINFO_MEDIA_TYPE
	.dword	FWINFO_MEDIA_ERASESIZE
	.dword	FWINFO_MEDIA_PAGESIZE
	.dword	FWINFO_MEDIA_TOTALSIZE_MB
1:
	.ascii FWINFO_HWID
	.org	1b + FWINFO_HWID_LENGTH, ' '
.endm

#endif /* !__ASSEMBLY__ */

#endif /* _FWINFO_DATA_H */

/******************************* END OF FILE ********************************/
