/*****************************************************************************
*!
*! FILE NAME  : lib.h
*!
*! DESCRIPTION: Small practical functions for boot loader.
*! 		malloc/free
*! 		memcpy/memset
*! 		writeb/writew/readb/readw
*! 		putc/puts/putnybble/putx
*! 		error/error2
*! 		serial_init
*!
*! ---------------------------------------------------------------------------
*!
*! malloc by Hannu Savolainen 1993 and Matthias Urlichs 1994
*! puts by Nick Holloway 1993, better puts by Martin Mares 1995
*! adoptation for Linux/CRIS Axis Communications AB, 1999
*! 
*! Other parts (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
*!
*! This program is free software; you can redistribute it and/or modify
*! it under the terms of the GNU General Public License as published by
*! the Free Software Foundation; either version 2 of the License, or
*! (at your option) any later version.
*!
*! This program is distributed in the hope that it will be useful,
*! but WITHOUT ANY WARRANTY; without even the implied warranty of
*! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*! GNU General Public License for more details.
*!
*! You should have received a copy of the GNU General Public License
*! along with this program; if not, write to the Free Software
*! Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*!
*!***************************************************************************/

#ifndef _LIB_H
#define _LIB_H

#include <linux/types.h>

/* nice stuff we need without having any library around */

void* memset(void* s, int c, size_t n);
void* memcpy(void* __dest, __const void* __src, size_t __n);

#define memzero(s, n) memset((s), 0, (n))

void *malloc(unsigned int size);
void free(void *where);
void error(const char *m) __attribute__ ((noreturn));
void error2(const char *m, int l, const char *f) __attribute__ ((noreturn));
void serial_init(void);

void putc(const char);
void puts(const char *);
void putnybble(unsigned int n);
void putx(unsigned int x);
void putnl();

#define GPIO_SYNC 0
unsigned char readb(const volatile void *addr);
unsigned short readw(const volatile void *addr);
void writeb(unsigned char b, volatile void *addr);
void writew(unsigned short b, volatile void *addr);

#endif /* _LIB_H */
