<?php
/*!*******************************************************************************
*! FILE NAME  : coring_filter_setup.php
*! DESCRIPTION: Temporary generates and writes coring function to the FPGA table
*! Copyright (C) 2010 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!  $Log: coring_filter_setup.php,v $
*!  Revision 1.1  2010/08/01 19:29:24  elphel
*!  files updated to support coring function for noise filtering
*!
*/
$TABLE='0';
$FILTER=0.5;
$DEBUG=false;
$SIMUL=false;
$SET_Y=true;
$SET_C=true;
$GENERATE_C=false;
foreach ($_GET as $key=>$value) {
  switch (strtoupper($key)) {
    case 'F':  
    case 'FILTER':   $FILTER=$value+0;break;
    case 'T':  
    case 'N':  
    case 'TABLE':    $TABLE=strtoupper($value);break;
    case 'DEBUG':    $DEBUG=true;break;
    case 'SIMUL':    $SIMUL=true;break;
    case 'C':        $GENERATE_C=true;break;

  }
}
if ($GENERATE_C) {
  generateCCode(0,10,0.1); //$min,$max,$step
  exit(0);
}

$TABLE_NUMBER=$TABLE+0;
if ($TABLE_NUMBER>8) {
  echo "TABLE_NUMBER should be 0..8 (with optional modifiers 'Y' and/or 'C'), you entered $TABLE_NUMBER";
  exit (0);
}
$SET_Y=(strpos($TABLE,'Y')!==false);
$SET_C=(strpos($TABLE,'C')!==false);
if (!$SET_Y && ! $SET_C) {
  $SET_Y=true;
  $SET_C=true;
}

if ($DEBUG) echo "<pre>\n";
  
$bin_table=generateCoringFunction($FILTER);;
//var_dump($bin_table);
if ($DEBUG) {
//  print_r($table);
  for ($i=0; $i<strlen($bin_table);$i++) {
    $d=ord(substr($bin_table,$i,1));
//echo $d;
// swap nibbles;
    $d=(($d & 0xf)<<4) | ($d>>4);
    $s=dechex($d);
    while (strlen($s)<2) $s='.'.$s;
    echo $s;
  }
  echo "\n";
}
//}
if ($DEBUG) echo "</pre>\n";

if ($SIMUL) {
  echo "<pre>\n";
  echo "\nlines for coring.dat file\n";
  for ($i=0; $i<strlen($bin_table);$i+=2) {
    if (($i & 0xf)==0) echo "\n";
    $d=ord(substr($bin_table,$i,1))+(ord(substr($bin_table,$i+1,1))<<8);
    $s=dechex($d);
    while (strlen($s)<4) $s='0'.$s;
    echo '  '.$s;
  }
  echo "\n";
  echo "</pre>\n";
}

if ($DEBUG) echo "<pre>\n";

/// Now write to FPGA
       $tables_file  = fopen('/dev/fpga_tables', 'w');
       if ($SET_Y) {
         if ($DEBUG) echo "Setting Y table $TABLE_NUMBER\n";
         fseek ($tables_file, 0x3000+0x100*$TABLE_NUMBER) ; // start of filter table, Y
         fwrite($tables_file, $bin_table);
       }
       if ($SET_C) {
         if ($DEBUG) echo "Setting C table $TABLE_NUMBER\n";
         fseek ($tables_file, 0x3080+0x100*$TABLE_NUMBER) ; // start of filter table, C
         fwrite($tables_file, $bin_table);
       }
       fclose($tables_file);
if ($DEBUG) echo "</pre>\n";

exit (0);

function generateCCode0($min,$max,$step) {
  echo "<pre>\n";
  echo "// Coring function table, 256 4-bit values per function\n";
  echo "// min=$min, max=$max, step=$step\n\n";
  for ($filter=$min; $filter<$max; $filter+=$step) {
    echo "// filter=$filter\n";
    $bin_table=generateCoringFunction($filter);
//    echo strlen($bin_table);
    for ($l=0; $l<strlen($bin_table);$l+=32) {
      $line='"';
      for ($i=$l; ($i<$l+32) && ($i<strlen($bin_table));$i++) {
        $d=ord(substr($bin_table,$i,1));
// swap nibbles;
        $d=(($d & 0xf)<<4) | ($d>>4);
        $s=dechex($d);
        while (strlen($s)<2) $s='0'.$s;
        $line.=$s;
      }
      echo $line."\"\n";
    }
  }
  echo "</pre>\n";
}

function generateCCode($min,$max,$step) {
  echo "<pre>\n";
  echo "// Coring function table, 256 4-bit values per function\n";
  echo "// min=$min, max=$max, step=$step\n";
  echo "// See coring_filter_setup.php to generate this table (with parameter '?C')\n";

  echo "  static unsigned long coring_tables[]= {\n";
  $text='';
  for ($filter=$min; $filter<$max; $filter+=$step) {
    $text.= "// filter=$filter\n";
    $bin_table=generateCoringFunction($filter);
//    echo strlen($bin_table);
    for ($l=0; $l<strlen($bin_table);$l+=32) {
      $line='';
      for ($i=$l; ($i<$l+32) && ($i<strlen($bin_table));$i+=4) {
        $num="";
        for ($j=0;$j<4;$j++) {
          $d=ord(substr($bin_table,$i+$j,1));
          $s=dechex($d);
          while (strlen($s)<2) $s='0'.$s;
          $num=$s.$num;
        }
//        $line.=' 0x'.$num.',';
        $line.=' 0x'.$num.',';
      }
      $text.='  '.$line."\n";
    }
  }
  echo substr($text,0,strlen($text)-2);
  echo "\n  };\n";
  echo "</pre>\n";
}


function generateCoringFunction($filter) {
  $fc=16*$filter/(16-$filter);
  $F2=$fc*$fc;
  $bin_table="";
  for ($i=0;$i<128;$i++) for ($n=0;$n<2;$n++) {
    $x=((2*$i+$n)/16);
    $e=1;
    if ($filter>0) {
      $x_corr=16*$x/(16-$x); // behaves like x for small x, goex to infinity for large ones
      $e=1-exp(-($x_corr*$x_corr/$F2)); // try making +16 - infinity
    }
    $y=$e*$x; 
    $v= (int) ($y+0.5);
    if ($v>15) $v=15;
    if ($n==0) $d=$v;
    else  $bin_table.=pack('C', $d|($v<<4));
  }
 return $bin_table;
}

?>
