<?php
/*!*******************************************************************************
*! FILE NAME  : fpga.php
*! DESCRIPTION: Provides control of FPGA compressor/image acquisition
*!              to the circular buffer (circbuf)
*! Copyright (C) 2007 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!  $Log: fpga.php,v $
*!  Revision 1.1.1.1  2008/11/27 20:04:03  elphel
*!
*!
*!  Revision 1.4  2008/04/29 05:50:18  elphel
*!  made it handle unsigned long >=0x80000000
*!
*!  Revision 1.3  2008/03/20 22:32:38  elphel
*!  cleanup, modified to return xml
*!
*!  Revision 1.2  2008/03/17 17:44:03  elphel
*!  updated to make use of elhel_fpga_* (extension) instead of file read/write
*!
*!  Revision 1.1  2007/12/06 19:02:50  elphel
*!  collected most of the camera PHP script into one  /packages/web/353/php_top directory
*!
*!  Revision 1.1  2007/10/27 00:55:31  elphel
*!  untested revision - need to go
*!
*!  Revision 1.1  2007/10/16 19:25:27  elphel
*!  Added php script to control driver parameters (P_XXX, as defined in c313a.h) instead of the IOCTL. It is possible to control sensor phase with it
*!
*!
*/
    $address="";
    $data="";
     foreach($_GET as $key=>$value) {
       switch($key) {
       case "a":
       case "adr":
         $address=$value+0;
         break;
       case "a4":
       case "adr4":
         $address=$value+0x2000;
         break;
       case "d":
       case "data":
         $sdata=$value;
         $data=$value+0;
// Is there easier way to handle unsigned longs in PHP?
         if ((strtoupper(substr($sdata,0,2))=="0X") && (strlen($sdata)==10)) {
           $data1=substr($sdata,0,6)+0;
           if ($data1>=32768) $data1-=65536;
           $data=$data1*65536+("0x".substr($sdata,6,4));
         }
         break;
       }
      }
    if ($address=="") {
      echo <<<USAGE
   <p>This read/writes fpga registers. Use with extreme care.</p>
   <ul>
    <li><b>fpga.php?a=<i>register_address</i></b> -  read FPGA register <i>register_address</i></li>
    <li><b>fpga.php?a=<i>register_address</i>&d=<i>data</i></b> - write <i>data</i> to FPGA register <i>register_address</i></li>
    <li><b>fpga.php?a4=<i>register_address</i></b> - read FPGA register <i>register_address</i> (csp4)</li>
  </ul>
  <p><i>Note: only reading autoincrementing registers requires csp4 mode</i></p>
USAGE;
      exit (0);
    }
    $xml = new SimpleXMLElement("<?xml version='1.0'?><fpga/>");
    $xml->addChild ('address',($address>=0x2000)?($address-0x2000):$address);
    $xml->addChild ('hex_address',sprintf("0x%x",($address>=0x2000)?($address-0x2000):$address));
    if ($data==="") {
       $data=elphel_fpga_read($address);
       $xml->addChild ('read',$data);
       $xml->addChild ('hex_read',sprintf("0x%x",$data));
    } else {
       elphel_fpga_write($address,$data);
       $xml->addChild ('written',$data);
       $xml->addChild ('hex_written',sprintf("0x%x",$data));
    }
    $rslt=$xml->asXML();
    header("Content-Type: text/xml");
    header("Content-Length: ".strlen($rslt)."\n");
    header("Pragma: no-cache\n");
    printf($rslt);
?>
