#!/usr/local/sbin/php -q
<?php
/*!
*! PHP script
*! FILE NAME  : sensor_pattern_compress.php
*! DESCRIPTION: test of SDRAM phase, SDRAM memory, FPGA JPEG compression with pattern data from sensor instead of real image
*! AUTHOR     : spectr
*! Copyright (C) 2008 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*/

require 'i2c.inc';

$boot_flag_file="/var/state/boot.done";	// wil lbe created by launcher.php as last initscript after boot

$CMD = 'check';
if($_GET['cmd'] == 'add')
	$CMD = 'add';
if($argv[1] == 'add')
	$CMD = 'add';
$_pass = FALSE;

$sensor_type = '5MPx';

// wait here if we started from prod353
$flag = false;
while(!is_file("/etc/autocampars.xml")) {
	$flag = true;
	sleep(20);
}
if($flag)
	sleep(10);

function _sensor_id() {
	global $sensor_type;

	$s_id = i2c_receive(16, 0, 0x4800, 0);
	$sensor_id = sprintf("0x%x",$s_id);
	if($sensor_id == '0xffffffff') {
		$sensor_type = '3MPx';
		$s_id = i2c_receive(16, 0, 0x5D00, 0);
		$sensor_id = sprintf("0x%x",$s_id);
	}
	return $sensor_id;
}

function _sensor_pattern_set() {
	global $sensor_type;

	if($sensor_type == '5MPx') {
//		i2c_send(16, 0, 0x48A0, 0x19, 0);
		// disable Black Level Calibration
		i2c_send(16, 0, 0x4862, 0x02, 0);
		// use other test pattern
		i2c_send(16, 0, 0x48A0, 0x41, 0);
		return;
	}
	if($sensor_type == '3MPx') {
		$addr = 0x5D00;
		i2c_send(16, 0, $addr + 0x32, 0xAAAA, 0);
		$ctrl = i2c_receive(16, 0, $addr + 0x07, 0);
		$ctrl |= 0x40;
		i2c_send(16, 0, $addr + 0x07, $ctrl, 0);
		return;
	}
}

function _sensor_pattern_unset() {
	global $sensor_type;

	if($sensor_type == '5MPx') {
		i2c_send(16, 0, 0x4862, 0x00, 0);
		i2c_send(16, 0, 0x48A0, 0x00, 0);
		return;
	}
	if($sensor_type == '3MPx') {
		$addr = 0x5D00;
		$ctrl = i2c_receive(16, 0, $addr + 0x07, 0);
		$ctrl &= 0xFFBF;
		i2c_send(16, 0, $addr + 0x07, $ctrl, 0);
		return;
	}
}

// wait here 90 secs till boot complete
$boot_flag_counter = 0;
while(!file_exists($boot_flag_file)) {
	$boot_flag_counter += 1;
	if($boot_flag_counter > 90)
		break;
	sleep(1);
}

$MD_LENGTH = 32;
//$SPC_DAT_FILE = "/etc/test/spc.dat";
$spc_xml_file = "/etc/test/spc.xml";
$md5sums = "";
if(file_exists($spc_xml_file)) {
	$md5sums = new SimpleXMLElement(file_get_contents($spc_xml_file));
//	$inits = parse_ini_file($SPC_DAT_FILE);
} else {
	$md5sums = "<?xml version='1.0' standalone='yes'?><sensor_pattern_test>\n";
	// sample of records: "<md5sum sensor_id='0x1801' fpga_rev='0x353301e'>df5bf0c75fd10b1f9de3847076e32be2</md5sum>"
	$md5sums .= "</sensor_pattern_test>";
	$md5sums = new SimpleXMLElement($md5sums);
//	$inits["test"] = "empty";
}

# initialize result
$res_xml = "<?xml version='1.0' standalone='yes'?><test name='sensor_pattern_compression'>\n";

# "online result" - i.e. runtime
function res_update($name, $result, $description) {
	global $res_xml;

	$res_xml .= "<${name} result=\"${result}\">";
	if($description != "")
		$res_xml .= $description;
	$res_xml .= "</${name}>\n";
}

// return XML responce
function _finish() {
	global $res_xml;
	global $_pass;
	if($_pass == TRUE)
		res_update("finish", "ok", "");
	else
		res_update("finish", "fail", "");

	$res_xml .= "</test>";

	$xml = new SimpleXMLElement($res_xml);
	$rslt=$xml->asXML();

	header("Content-Type: text/xml");
	header("Content-Length: ".strlen($rslt)."\n");
	header("Pragma: no-cache\n");
	printf($rslt);

	exec("sync");

	exit(0);
}

//--------------------
// FPGA revision...
$data = elphel_fpga_read(0x13);
$fpga_rev = sprintf("0x%x",$data);

//--------------------
// check SDRAM phases
$SDRAM_PHASES_LOG = "/var/log/fpga_sdram_phase";
if(is_file($SDRAM_PHASES_LOG)) {
	$ph_rez = exec("cat $SDRAM_PHASES_LOG | grep 'result='");
	if($ph_rez == "result=OK") {
		res_update("SDRAM_phase", "ok", "");
	} else {
//		$__description = "SDRAM phase incorrect";
//		$__result = "fail";
		res_update("SDRAM_phase", "fail", "SDRAM phase incorrect");
		_finish();
	}
}

//--------------------
// test SDRAM
//exec("fpcf -w 4 0"); // turn off compressor
apply_def_settings(0);
sleep(2);
$sdram_test_file = '/var/log/sdram_test';
$sdram_rez = exec('fpcf -mem 2000000 FFFF 0000 AAAA 5555 > '.$sdram_test_file);
$sdram_rez = exec('cat '.$sdram_test_file." | grep 'result='");
//exec("fpcf -w 4 0"); // turn on compressor back
if($sdram_rez == "result=OK") {
	res_update("SDRAM_memory", "ok", "");
} else {
//	$__description = "SDRAM pattern test fail";
//	$__result = "fail";
	res_update("SDRAM_memory", "fail", "SDRAM pattern test fail");
	_finish();
}

//--------------------
// initialize...
// deprecated - 8.0 firmware do it at boot time
/*
file_get_contents("http://127.0.0.1/camera_demo.php");
*/

//--------------------
// sensor number...
//$s_id = i2c_receive(16, 0, 0x4800, 0);
//$sensor_id = sprintf("0x%x",$s_id);
$sensor_id = _sensor_id();
//--------------------
// check test pattern compressed image

//echo "sensor_id == ".$sensor_id."\n";

function md5sum_find($xml_md5sum, $sensor_id, $fpga_rev) {
	$rez->sensor = false;
	$rez->fpga = false;
	$rez->md5sum = "";

	for($i = 0; $i < count($xml_md5sum->md5sum); $i++) {
		$md5 = $xml_md5sum->md5sum[$i];
		if($md5['sensor_id'] == $sensor_id)
			$rez->sensor = true;
		if($md5['fpga_rev'] == $fpga_rev)
			$rez->fpga = true;
		if($md5['sensor_id'] == $sensor_id && $md5['fpga_rev'] == $fpga_rev) {
			$rez->md5sum = $md5;
			break;
		}
	}
	return $rez;
}

$md5sum = "";
if($CMD == "check") {
	$rez = md5sum_find($md5sums, $sensor_id, $fpga_rev);
	if(!$rez->sensor) {
		res_update('SENSOR_id', 'fail', 'Unknown sensor ID: '.$sensor_id);
		_finish();
	} else
		res_update('SENSOR_id', 'ok', $sensor_id);
	if(!$rez->fpga) {
		res_update('FPGA_rev', 'fail', 'Unknown FPGA revision: '.$fpga_rev);
		_finish();
	} else
		res_update('FPGA_rev', 'ok', $fpga_rev);
	if($rez->md5sum == "") {
		res_update('FPGA_compress', 'unknown', 'control MD5SUM for sensor ID '.$sensor_id.' and FPGA revision '.$fpga_rev.'not found');
		_finish();
	} else
		$md5sum = $rez->md5sum;
}

//echo "apply image settings...\n";

// image settings...
// file_get_contents("http://127.0.0.1/imconf_load.php?conf=/etc/test/sensor_pattern_compress.conf");

// DON'T change it! Except you know what you do, and will update spc.xml
// predefined image settings for test
// 96% quality, maximum width & height
// pass 0 - prepare for SDRAM test
// pass 1 - turn off daemons
function apply_def_settings($pass) {
	$DEF_QUALITY = 96;
	$DEF_WOI_WIDTH = 10000;
	$DEF_WOI_HEIGHT = 10000;
	// sample gamma == 1.0
	$DEF_GAMMA = 6554624;
	$DEF_COLOR = 2;	// color mode - "color, YCbCr 4:2:0, 3x3 pixels"

	/// apply predefined settings with autocampars.php
	$conf_xml = simplexml_load_file("/etc/autocampars.xml");
	foreach($conf_xml->paramSets->children() as $paramSet) {
	    $numSet = (integer)$paramSet->attributes()->number;
	    if($numSet == 0) {
			if($pass == 1 || $pass == 2 || $pass == 0) {
	    	    $paramSet->QUALITY = $DEF_QUALITY;
		        $paramSet->WOI_WIDTH = $DEF_WOI_WIDTH;
    		    $paramSet->WOI_HEIGHT = $DEF_WOI_HEIGHT;
        		$paramSet->GTAB_R = $DEF_GTAB_R;
		        $paramSet->GTAB_G = $DEF_GTAB_G;
    		    $paramSet->GTAB_B = $DEF_GTAB_B;
        		$paramSet->GTAB_GB = $DEF_GTAB_GB;
		        $paramSet->SENSOR_RUN = 2;
		        $paramSet->COMPRESSOR_RUN = 2;
				$paramSet->GAINR = 0x20000;
				$paramSet->GAING = 0x20000;
				$paramSet->GAINB = 0x20000;
				$paramSet->GAINGB = 0x20000;
				$paramSet->RSCALE = 0x10000;
				$paramSet->GSCALE = 0x10000;
				$paramSet->BSCALE = 0x10000;
				$paramSet->DGAINR = 32768;
				$paramSet->DGAING = 32768;
				$paramSet->DGAINGB = 32768;
				$paramSet->DGAINB = 32768;
				$paramSet->COLOR = 1;
				$paramSet->AUTOEXP_ON = 0;
				$paramSet->GAIN_MIN = 0x10000;
				$paramSet->GAIN_MAX = 0x10000;
				$paramSet->ANA_GAIN_ENABLE = 0;
				// daemons
				$paramSet->DAEMON_EN = 0;
				$paramSet->DAEMON_EN_AUTOEXPOSURE = 0;
				$paramSet->DAEMON_EN_STREAMER = 0;
				$paramSet->DAEMON_EN_CCAMFTP = 0;
				$paramSet->DAEMON_EN_CAMOGM = 0;
				$paramSet->DAEMON_EN_AUTOCAMPARS = 0;
			}
			if($pass == 0) { // for SDRAM test
		        $paramSet->SENSOR_RUN = 0;
		        $paramSet->COMPRESSOR_RUN = 0;
			}
	    }
	}
	$z = $conf_xml->asXml();
	$f = fopen('/tmp/spc_def.xml', 'w+');
	fwrite($f, $z);
	fclose($f);

	exec('wget http://127.0.0.1/autocampars.php?load=/tmp/spc_def.xml');
//	exec('/usr/html/autocampars.php --load /tmp/spc_def.xml');
//exec('/mnt/flash/html/autocampars.php --load /tmp/spc.xml');
}

// setup image and turn off all autocolors etc
apply_def_settings(1);
for($i = 0; $i < 6; $i++)
	exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');
sleep(3);
apply_def_settings(2);
for($i = 0; $i < 6; $i++)
	exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

//echo "done\n";
//echo "set test pattern...\n";

// set diagonal gradient test pattern on camera
_sensor_pattern_set();
//echo "done\n";

// load test image to /tmp/spc.jpeg
# skip next five images
# untill settings apply and wait for possible glitches
for($i = 0; $i < 10; $i++) {
	exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc_'.$i.'.jpeg');
//	exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null");
//	exec("wget http://127.0.0.1:8081/noexif/img -O /dev/null");
//	sleep(1);
}
# this one is an actual test image
exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
//exec("wget http://127.0.0.1:8081/noexif/img -O /tmp/spc.jpeg");

$md = "";
if(is_file("/tmp/spc.jpeg")) {
	$md = substr(exec("md5sum /tmp/spc.jpeg"), 0, $MD_LENGTH);
//	unlink("/tmp/spc.jpeg");
}

if($CMD == "check") {
//	$_md = $inits["rev_".$fpga_rev];
	if($md != $md5sum) {
//		$__result = "fail";
//		$__description = "MD5SUM is |$md| and must be |$_md|";
		res_update("FPGA_compress", "fail", "MD5SUM is |$md| and must be |$md5sum|");
		_sensor_pattern_unset();
//		$s_id = i2c_send(16, 0, 0x48A0, 0x00, 0);
		_finish();
	} else {
		res_update("FPGA_compress", "ok", "");
		$_pass = TRUE;
//		$__result="ok";
	}
} else {
	if($CMD == "add") {
		// check what the sensor/FPGA pair is a really new
		$rez = md5sum_find($md5sums, $sensor_id, $fpga_rev);
		if($rez->sensor && $rez->fpga) {
			res_update("FPGA_compress", "fail", "MD5SUM record already exist");
			_finish();
		}
		$new_md5 = $md5sums->addChild("md5sum", $md);
		$new_md5['sensor_id'] = $sensor_id;
		$new_md5['fpga_rev'] = $fpga_rev;
		$r = $md5sums->asXML();
$spc_xml_file = "/etc/test/spc.xml";
		$f = fopen($spc_xml_file, 'w');
		fwrite($f, $r);
		fclose($f);
		res_update("FPGA_compress", "ok", "MD5SUM record updated");
		$_pass = TRUE;
	} else {
		res_update("FPGA_compress", "unknown", "wrong command ${MODE}");
	}
}

// disable test pattern
//$s_id = i2c_send(16, 0, 0x48A0, 0x00, 0);
_sensor_pattern_unset();
_finish();

?>
