#!/usr/local/sbin/php -q
<?php
/*!
*! PHP script
*! FILE NAME  : test_10359.php
*! DESCRIPTION: The script is for tesing 10359 board with 5MPix sensors.
*!              1. Initializes 10359 board.
*!              2. Checks 10353-10359 I2C connection.
*!              3. Detects and programs sensors (up to 3).
*!              4. Adjusts clock phase (+/-40*2 steps) for each sensor until MD5 sum of a frame is correct. 
*! AUTHOR     : spectr, dzhimiev
*! Copyright (C) 2008 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log: test_inner_10359.php,v $
*!  Revision 1.17  2010/06/02 21:23:13  dzhimiev
*!  1. updated according to all recent changes
*!
*!  Revision 1.14  2009/11/20 16:45:02  dzhimiev
*!  1. commented out 353 fpga revision check
*!
*!  Revision 1.13  2009/10/07 14:40:21  dzhimiev
*!  1. fixed to work with prod353 again
*!
*!  Revision 1.12  2009/10/06 16:36:26  dzhimiev
*!  1. added work_range records to the result
*!  2. corrected the calculation of the shift to the middle of the work interval
*!  3. modified resulting logic of the shift_check function
*!
*!  Revision 1.11  2009/10/05 16:47:55  dzhimiev
*!  1. added frame waits between sensor's initialization procedures and channels switching
*!
*!  Revision 1.10  2009/09/29 22:15:55  elphel
*!  missing #!/usr/local/sbin/php -q at the  beginning of the file
*!
*!  Revision 1.9  2009/09/21 15:52:49  dzhimiev
*!  1. added an extra frame 'get' between the ones that are checked and compared
*!
*!  Revision 1.8  2009/09/04 06:42:49  dzhimiev
*!  1. a mistake in an index of the 'while' loop corrected
*!
*!  Revision 1.7  2009/08/28 10:19:52  dzhimiev
*!  1. new technique of finding the right phase shift
*!
*!  Revision 1.3  2009/03/03 18:56:59  dzhimiev
*!  1. added a 10-second delay before starting the test to avoid it during setting the sensor's frequency and the sensor's initialization
*!
*!  Revision 1.2  2009/02/26 14:04:45  dzhimiev
*!  1. pause after channel switch - more stable work
*!  2. test fails if one of the phases fails to be adjusted
*!
*!  Revision 1.1  2009/02/05 20:00:00  dzhimiev
*!  files for 10359 board test - ported from 7.x to 8.x software version
*!
*!
*/

require 'i2c.inc';

// result of the test will be return by test_10359.php on request
$res_fn = "/etc/test/board_10359.res";

// program name for launcher
$self_fn="test_inner_10359.php";

$bdetect_log="/var/log/bdetect";

//------------------------------------------------------------------------------
// general flow functions

//c12508bb1315bc90f94e5f49c9d2f405

function pr($text) {
	//echo "<pre>$text</pre>\n";
	flush();
}

function res_update($name, $result, $description) {
	global $res_rpt;

	if($result != "")
		$res_rpt.="<${name} result=\"${result}\">".$description."</${name}>\n";
	else
		$res_rpt.="<${name}>".$description."</${name}>\n";
}

# result - stored in a file
function res_update2($name, $result, $description) {
	global $res_fn;

	$file = fopen($res_fn, "a+");
	if($result != "")
		fwrite($file, "<${name} result=\"${result}\">".$description."</${name}>\n");
	else
		fwrite($file, "<${name}>".$description."</${name}>\n");
	fclose($file);

	pr("<${name} result=\"${result}\">".$description."</${name}>");
}

function _finish($res) {
	global $res_rpt;
	global $res_fn;

	res_update("finish", $res, "");

	$res_rpt.="</test>\n";

	$file = fopen($res_fn, "a+");
	fwrite($file, $res_rpt);
	fclose($file);
	flush();

	$res_xml=file_get_contents("/etc/test/board_10359.res");
	header("Content-Type: text/xml");
	header("Content-Length: ".strlen($res_xml)."\n");
	header("Pragma: no-cache\n");
	printf("%s", $res_xml);
	flush();
	exit(0);
}

$sensor_type = '5MPx';

function apply_def_settings($pass) {
	$DEF_QUALITY = 96;
	$DEF_WOI_WIDTH = 10000;
	$DEF_WOI_HEIGHT = 10000;
	// sample gamma == 1.0
	$DEF_GAMMA = 6554624;
	$DEF_COLOR = 2;	// color mode - "color, YCbCr 4:2:0, 3x3 pixels"

	// copy predefined settings with autocampars.php
	$conf_xml = simplexml_load_file("/etc/autocampars.xml");
	foreach($conf_xml->paramSets->children() as $paramSet) {
		$numSet = (integer)$paramSet->attributes()->number;
		if($numSet == 0) {
			if($pass == 1 || $pass == 2 || $pass == 0) {
				$paramSet->QUALITY = $DEF_QUALITY;
				$paramSet->WOI_WIDTH = $DEF_WOI_WIDTH;
				$paramSet->WOI_HEIGHT = $DEF_WOI_HEIGHT;
				$paramSet->GTAB_R = $DEF_GTAB_R;
				$paramSet->GTAB_G = $DEF_GTAB_G;
				$paramSet->GTAB_B = $DEF_GTAB_B;
				$paramSet->GTAB_GB = $DEF_GTAB_GB;
				$paramSet->SENSOR_RUN = 2;
				$paramSet->COMPRESSOR_RUN = 2;
				$paramSet->GAINR = 0x20000;
				$paramSet->GAING = 0x20000;
				$paramSet->GAINB = 0x20000;
				$paramSet->GAINGB = 0x20000;
				$paramSet->RSCALE = 0x10000;
				$paramSet->GSCALE = 0x10000;
				$paramSet->BSCALE = 0x10000;
				$paramSet->DGAINR = 32768;
				$paramSet->DGAING = 32768;
				$paramSet->DGAINGB = 32768;
				$paramSet->DGAINB = 32768;
				$paramSet->COLOR = 1;
				$paramSet->AUTOEXP_ON = 0;
				$paramSet->GAIN_MIN = 0x10000;
				$paramSet->GAIN_MAX = 0x10000;
				$paramSet->ANA_GAIN_ENABLE = 0;
				// daemons
				$paramSet->DAEMON_EN = 0;
				$paramSet->DAEMON_EN_AUTOEXPOSURE = 0;
				$paramSet->DAEMON_EN_STREAMER = 1;
				$paramSet->DAEMON_EN_CCAMFTP = 0;
				$paramSet->DAEMON_EN_CAMOGM = 0;
				$paramSet->DAEMON_EN_AUTOCAMPARS = 0;
			}
			if($pass == 0) { // for SDRAM test
		        	$paramSet->SENSOR_RUN = 0;
		        	$paramSet->COMPRESSOR_RUN = 0;
			}
	    }
	}
	$z = $conf_xml->asXml();
	$f = fopen('/tmp/cd359_def.xml', 'w+');
	fwrite($f, $z);
	fclose($f);
	exec('wget http://127.0.0.1/autocampars.php?load=/tmp/cd359_def.xml');
}

function apply_def_settings2($pass) {
	$DEF_QUALITY = 96;
	$DEF_WOI_WIDTH = 10000;
	$DEF_WOI_HEIGHT = 10000;
	// sample gamma == 1.0
	$DEF_GAMMA = 6554624;
	$DEF_COLOR = 2;	// color mode - "color, YCbCr 4:2:0, 3x3 pixels"


	if($pass == 1 || $pass == 2 || $pass == 0) {

		$parsForProgram=array(
			'QUALITY'         => $DEF_QUALITY,
			'WOI_WIDTH'       => $DEF_WOI_WIDTH,
			'WOI_HEIGHT'      => $DEF_WOI_HEIGHT,
			'GTAB_R'          => $DEF_GTAB_R,
			'GTAB_G'          => $DEF_GTAB_G,
			'GTAB_B'          => $DEF_GTAB_B,
			'GTAB_GB'         => $DEF_GTAB_GB,
			'SENSOR_RUN'      => 2,
			'COMPRESSOR_RUN'  => 2,
			'GAINR'           => 0x20000,
			'GAING'           => 0x20000,
			'GAINB'           => 0x20000,
			'GAINGB'          => 0x20000,
			'RSCALE'          => 0x10000,
			'GSCALE'          => 0x10000,
			'BSCALE'          => 0x10000,
			'DGAINR'          => 32768,
			'DGAING'          => 32768,
			'DGAINGB'         => 32768,
			'DGAINB'          => 32768,
			'COLOR'           => 1,
			'AUTOEXP_ON'      => 0,
			'GAIN_MIN'        => 0x10000,
			'GAIN_MAX'        => 0x10000,
			'ANA_GAIN_ENABLE' => 0x10000,
			//daemons
			'DAEMON_EN'              => 0,
			'DAEMON_EN_AUTOEXPOSURE' => 0,
			'DAEMON_EN_STREAMER'     => 1,
			'DAEMON_EN_CCAMFTP'      => 0,
			'DAEMON_EN_CAMOGM'       => 0,
			'DAEMON_EN_AUTOCAMPARS'  => 0
		);

		}
	if($pass == 0) { // for SDRAM test
		$parsForProgram=array(
			'SENSOR_RUN'      => 0,
			'COMPRESSOR_RUN'  => 0
		);
	}

	elphel_set_P_arr($parsForProgram);

}

function sensor_init($init_pars){
	for ($i=0;$i<256;$i++){
		i2c_send(16,0,0x4800+$i,$init_pars[$i],0);	
		$readout=i2c_receive(16,0,0x4800+$i,0);
	}
}

function send_i2c_4($a,$d) {
	$i2c  = fopen('/dev/xi2c16','w');
	fseek ($i2c, 2*$a) ;
	$res=fwrite($i2c, chr (floor($d/(256*256*256))).chr (($d - 256*256*256*floor($d/(256*256*256)))/(256*256)).chr (($d - 256*256*floor($d/(256*256)))/256).chr ($d - 256*floor($d/(256))) );
	fclose($i2c);
	return $res;
} // end of send_i2c()

function receive_i2c_4($width,$bus,$a,$raw=0) {
   $w=($width==16)?2:1;
   $i2c_fn='/dev/xi2c'.($raw?'raw':(($w==2)?'16':'8')).(($bus==0)?'':'_aux');
   $i2c  = fopen($i2c_fn, 'r');
	 fseek ($i2c, 4, SEEK_END);
   fseek ($i2c, $w*$a);
	 
   $data = fread($i2c, 2*$w);
   fclose($i2c);
   if (strlen($data)<2*$w) return -1;
   $v=unpack(($w==1)?'C*':'N*',$data);
	 //print_r($v);
	 //printf("0x%x\n",$v[1]);
	 return $v[1];
} // end of receive_i2c()

function _sensor_id() {
	global $sensor_type;

	$s_id = i2c_receive(16, 0, 0x4800, 0);
	$sensor_id = sprintf("0x%x",$s_id);
	if($sensor_id == '0xffffffff') {
		$sensor_type = '3MPx';
		$s_id = i2c_receive(16, 0, 0x5D00, 0);
		$sensor_id = sprintf("0x%x",$s_id);
	}
	return $sensor_id;
}

function _sensor_pattern_set() {
	global $sensor_type;

	if($sensor_type == '5MPx') {
//		i2c_send(16, 0, 0x48A0, 0x19, 0);
		// use other test pattern
		i2c_send(16, 0, 0x48A0, 0x41, 0);
		return;
	}
	if($sensor_type == '3MPx') {
		$addr = 0x5D00;
		i2c_send(16, 0, $addr + 0x32, 0xAAAA, 0);
		$ctrl = i2c_receive(16, 0, $addr + 0x07, 0);
		$ctrl |= 0x40;
		i2c_send(16, 0, $addr + 0x07, $ctrl, 0);
		return;
	}
}

function _sensor_pattern_unset() {
	global $sensor_type;

	if($sensor_type == '5MPx') {
		i2c_send(16, 0, 0x48A0, 0x00, 0);
		return;
	}
	if($sensor_type == '3MPx') {
		$addr = 0x5D00;
		$ctrl = i2c_receive(16, 0, $addr + 0x07, 0);
		$ctrl &= 0xFFBF;
		i2c_send(16, 0, $addr + 0x07, $ctrl, 0);
		return;
	}
}

function md5sum_find($xml_md5sum, $sensor_id, $fpga_rev) {
	$rez->sensor = false;
	$rez->fpga = false;
	$rez->md5sum = "";

	for($i = 0; $i < count($xml_md5sum->md5sum); $i++) {
		$md5 = $xml_md5sum->md5sum[$i];
		if($md5['sensor_id'] == $sensor_id)
			$rez->sensor = true;
		//if($md5['fpga_rev'] == $fpga_rev)
			$rez->fpga = true;
		//if($md5['sensor_id'] == $sensor_id && $md5['fpga_rev'] == $fpga_rev) {
		if($md5['sensor_id'] == $sensor_id) {
			$rez->md5sum = $md5;
			break;
		}
	}
	return $rez;
}

function phase_adjust($channel,$md5){
  $phase_shift=1;
  $md5_new = "";
  $md5_old = "";
  for($i=0;$i<241;$i++){
    if ($i==80) $phase_shift=2;
    exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');
    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift='.$phase_shift.'&n=2');
    //exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null");
    exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
    if(is_file("/tmp/spc.jpeg")) {
      $md5_new = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);
      unlink("/tmp/spc.jpeg");
    }      
    if (($md5_new==$md5)&&($md5_old==$md5)) return true;
    $md5_old=$md5_new;
  }
  return false;
}

function test_pattern_phase_adjusment() {
  $phase_shift=1;
  $md5_int="";
  $md5_new = "";
  $md5_old = "";
  $channel=-1;

  //go to 0
  exec("fpcf -w 8 f0");
/*
  $ch_phase_=receive_i2c_4(16,0,0x847,0);
  $ch_phase_value = $ch_phase_ & 0x1ff;

  if ($ch_phase_value>256) $ch_phase_value=$ch_phase_value-512;

  if ($ch_phase_value<0)
    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=1&n='.-$ch_phase_value);
  else
    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=2&n='.$ch_phase_value);
*/  

  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');
  
for($j=0;$j<4;$j++) {
  for($i=0;$i<301;$i++){
    if ($i==100) $phase_shift=2;
    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift='.$phase_shift.'&n=2');
    //exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null");

    $md5_old=0;
    $md5_int=0;
    $md5_new=0;

    exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
    if(is_file("/tmp/spc.jpeg")) {
      if (filesize("/tmp/spc.jpeg")<500000) $md5_old = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);
      unlink("/tmp/spc.jpeg");
    }      
    exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
    if(is_file("/tmp/spc.jpeg")) {
      if (filesize("/tmp/spc.jpeg")<500000) $md5_int = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);
      unlink("/tmp/spc.jpeg");
    }
    exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
    if(is_file("/tmp/spc.jpeg")) {
      if (filesize("/tmp/spc.jpeg")<500000) $md5_new = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);
      unlink("/tmp/spc.jpeg");
    }

    //if (($md5_new==$md5_int)&&($md5_old==$md5_int)&&($md5_int=='8d26686bda085d680f8527af430304eb' || $md5_int=='f6fb3e937f766cea88888a35009c856d' || $md5_int=='af6cdc53f51f6a44cd330b8121701d0f')) { 
    if (($md5_new==$md5_int)&&($md5_old==$md5_int)){
      test_pattern_fine_tuning();
      res_update("test_pattern_md5", "", $md5_int);
      return true;
    }


  }
  file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm=0&phase_shift=4&n=1');
}
  return false;
}

function test_pattern_fine_tuning(){
  $k_l=0;
  $k_r=0;
  global $dcm_range;

  $dcm_range[3]=0;

  $channel=-1;

  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

  while ((shift_check(-1,2,1,0))&&($k_l<50)) {
    $k_l=$k_l+1;
  }
  file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=1&n='.$k_l);

  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

  while ((shift_check(-1,1,1,0)&&($k_r<50))) {
    $k_r=$k_r+1;
  }

  $dcm_range[3]=$k_r+$k_l;

  file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=2&n='.floor(($k_r+$k_l)/2));

  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

  return shift_check(-1,1,0,0);
}

function coarse_adjustment($channel,$md5,$range,$step){
  $phase_shift=1;
  $md5_new = "";
  $md5_old = "";
  $k=0;

  //go to 0
  //file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=0xf&n=1');
  i2c_send(16,0,0x803+$channel,0xf,0);
/*
  $ch_phase_=receive_i2c_4(16,0,"0x86".$channel+1,0);

  $ch_phase_value = $ch_phase_ & 0x3ff;

  if ($ch_phase_value>256) $ch_phase_value=$ch_phase_value-512;

  if ($ch_phase_value<0) {
    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=1&n='.-$ch_phase_value);
  }
  else {
    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=2&n='.$ch_phase_value);
  }
*/

  // first check current position
  if (shift_check($channel,$phase_shift,0,$md5)) {
      return fine_tuning($channel,$md5);
  }

  //move to the left border
  file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=2&n='.floor($range/2));
  //apply_def_settings(2);
  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

  if (shift_check($channel,$phase_shift,$step,$md5)) {
      return fine_tuning($channel,$md5);
  }
  
  $k=$step-$range/2;
  //res_update('current_k_'.$channel, 'ok', $k);
 
  while ($k+2*$step<$range/2) {
    if (shift_check($channel,$phase_shift,2*$step,$md5)) {
      return fine_tuning($channel,$md5);
    }
    $k=$k+2*$step;
    if (($k+2*$step>$range/2)&&($step>5)){
      //back to 0
      if ($k>0) {
	file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=2&n='.$k);
      }
      else {
	file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=1&n='.-$k);
      }

      //apply_def_settings(2);
      exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

      //next iteration
      if (coarse_adjustment($channel,$md5,$range,$step/2)) return true;
    }
  }

  return false;
}

function fine_tuning($channel,$md5){
  $k_l=0;
  $k_r=0;
  global $dcm_range;

  $dcm_range[$channel]=0;

  while ((shift_check($channel,2,1,$md5))&&($k_l<50)) {
    $k_l=$k_l+1;
  }

  file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=1&n='.$k_l);

  //apply_def_settings(2);
  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

  while ((shift_check($channel,1,1,$md5)&&($k_r<50))) {
    $k_r=$k_r+1;
  }

  $dcm_range[$channel]=$k_r+$k_l;

  file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift=2&n='.floor(($k_r+$k_l)/2));

  //apply_def_settings(2);
  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

  if ($dcm_range[$channel]==0) return false;

  return shift_check($channel,1,0,$md5);
}

function shift_check($channel,$phase_shift,$number_of_steps,$md5){
  //for($i=0;$i<$number_of_steps;$i++){

    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($channel+1).'&phase_shift='.$phase_shift.'&n='.$number_of_steps);
    //apply_def_settings(2);
    exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');
    //res_update('current_phase_ch'.$channel, 'ok', receive_i2c_4(16,0,'0x86'.$channel+1,0));

    $sample1=1;
    $sample2=2;
    $sample3=3;

    exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
    if(is_file("/tmp/spc.jpeg")) { if (filesize("/tmp/spc.jpeg")<500000) $sample1 = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);unlink("/tmp/spc.jpeg");}

    exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
    if(is_file("/tmp/spc.jpeg")) { if (filesize("/tmp/spc.jpeg")<500000) $sample2 = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);unlink("/tmp/spc.jpeg");}

    exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
    if(is_file("/tmp/spc.jpeg")) { if (filesize("/tmp/spc.jpeg")<500000) $sample3 = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);unlink("/tmp/spc.jpeg");}

    if (($md5==0) && ($sample1==$sample2) && ($sample2==$sample3)) return true;
    else if (($sample1==$md5)||($sample2==$md5)||($sample3==$md5)) return true;
    else if (($sample1==$sample2) && ($sample2==$sample3))         return true;
    return false;
  //}
}

function ddr_page_write($shift) {
	//fill a buffer
	global $data_in;
	$j=0;

	for ($i=0;$i<128;$i++) {
	  $word=$i+$shift;
	  //send_i2c_4(0x0802,$i+$shift);
	  //i2c_send(16,0,0x870,$i+$shift,0);
	  //exec("fpcf -i2cw16 870 $word");

	  if ((($i-3)/8==floor(($i-3)/8))||(($i-4)/8==floor(($i-4)/8))||(($i-5)/8==floor(($i-5)/8))) $word=0xffff;
	  else                                                                                       $word=0x0000;

	  i2c_send(16,0,0x870,$word,0);

	  $data_in[$i]=$word;
	  //usleep(10);
	}

	//write page, data does not matter
	for ($i=0;$i<1;$i++) {
		i2c_send(16,0,0x863,0x0001,0);
	}
}

function ddr_page_read($color) {

	global $data_in;

	$j=0;
	//read page, data does not matter
	for ($i=0;$i<1;$i++) {
		i2c_send(16,0,0x864,0x0001,0);
	}
	$adr=0x875;
	for ($i=0;$i<128;$i++) {
	  $data[$i]=i2c_receive(16,0,0x870,0);
	}

	for ($i=0;$i<128;$i++) if ($data[$i]<>$data_in[$i]) return false;

	return true;
}

function memory_test() {

	//reset the current position
	$current_quarter = i2c_receive(16,0,0x860,0);
	$current_position_ = receive_i2c_4(16,0,0x860,0);
	$current_position  = $current_position_ & 0x3ff;

	for($i=0;$i<$current_quarter;$i++) i2c_send(16,0,0x802,0x8);

	if ($current_position>256) $current_position=$current_position-512;

	if ($current_position<0) {
		for($i=0;$i<(-$current_position);$i++) i2c_send(16,0,0x802,0x1);
	}
	else {
		for($i=0;$i<( $current_position);$i++) i2c_send(16,0,0x802,0x2);
	}

	 // start test
	for($i=0;$i<4;$i++) {
		for($j=0;$j<90;$j++) {
			if (memory_rw_check()) return fine_memory_test();
			if ($j<30) for ($k=0;$k<5;$k++) i2c_send(16,0,0x802,0x1); // shift
			else       for ($k=0;$k<5;$k++) i2c_send(16,0,0x802,0x2); // shift
		}
		// back to 0
		for ($k=0;$k<30;$k++) i2c_send(16,0,0x802,0x1);
		// quarter change
		i2c_send(16,0,0x802,0x4); // 90 degree shift
	}
	return true;
}

function fine_memory_test() {

	global $mem_dcm_range;
	$k_l=0;
	$k_r=0;

	// find the left border
	while ((memory_rw_check())&&($k_l<50)) {
		$k_l=$k_l+1;
		i2c_send(16,0,0x802,0x2);
	}

	// back to the center
	for ($i=0;$i<$k_l;$i++) {
		i2c_send(16,0,0x802,0x1);
	}

	// find the right border
	while ((memory_rw_check())&&($k_r<50)) {
		$k_r=$k_r+1;
		i2c_send(16,0,0x802,0x1);
	}

	// go to the center of the range
	for ($i=0;$i<floor(($k_r+$k_l)/2);$i++) {
		i2c_send(16,0,0x802,0x2);
	}

	$mem_dcm_range=$k_r+$k_l;

	if ($mem_dcm_range==0) return false;
	
	// final shift
	return memory_rw_check();
}

function memory_rw_check() {
    ddr_page_write(0); // ddr_page_write writes 1024 words to a buffer and then transfer it with 16 writes to SDRAM
    return ddr_page_read(0x00007F);
}

//------------------------------------------------------------------------------
// at the stage processing we must do this:
// if test passed:
// - update results record
// - increase stage number
// - reboot
// if test fail:
// - update results record
// - finish

$flag = FALSE;

set_time_limit(300);

$dcm_range=array('N/A','N/A','N/A','N/A');
$mem_dcm_range='N/A';

$save=false;
$dcm0=0;
$dcm1=0;
$dcm2=0;
$dcm3=0;

$dcm0_90=0;
$dcm1_90=0;
$dcm2_90=0;
$dcm3_90=0;

if(is_file($res_fn)) unlink($res_fn); // delete an old report file

$res_rpt="<?xml version='1.0' standalone='yes'?>\n<test name=\"test_10359\">\n";

/*	$file = fopen($res_fn, "a+");
	fwrite($file, "<?xml version='1.0' standalone='yes'?>\n<test_inner_10359>\n");
	fclose($file);
	flush();
*/

if(is_file($bdetect_log)) {
  if (file_get_contents("/var/log/bdetect")=="10359") $flag = true;
  res_update("board_id", "ok", file_get_contents("/var/log/bdetect"));
}

if(!$flag) {
  res_update("board_id", "fail", '');
  _finish("fail");
}

foreach($_GET as $key=>$value) {
	switch($key) {
		case "save"           : $save=true;               break;
		case "dcm0"           : $dcm0=$value+0;           break;
		case "dcm0_90"        : $dcm0_90=$value+0;        break;
		case "dcm1"           : $dcm1=$value+0;           break;
		case "dcm1_90"        : $dcm1_90=$value+0;        break;
		case "dcm2"           : $dcm2=$value+0;           break;
		case "dcm2_90"        : $dcm2_90=$value+0;        break;
		case "dcm3"           : $dcm3=$value+0;           break;
		case "dcm3_90"        : $dcm3_90=$value+0;        break;
	}
}

if ($save) {
  res_update("sclk0_dcm","", $dcm0);
  res_update("ch0_dcm","", $dcm1);
  res_update("ch1_dcm","", $dcm2);
  res_update("ch2_dcm","", $dcm3);

  res_update("sclk0_dcm_90", "", $dcm0_90);
  res_update("ch0_dcm_90","", $dcm1_90);
  res_update("ch1_dcm_90","", $dcm2_90);
  res_update("ch2_dcm_90","", $dcm3_90);
  _finish("ok");
}else{
  if(test_main() == false) {
    _finish("fail");
  }else{
    _finish("ok");
  }
}

function test_main() {
  $MD_LENGTH = 32;
  $spc_xml_file = "/etc/test/spc.xml";
  global $res_fn;
  $md5sums = "";
  global $dcm_range;
  global $mem_dcm_range;

  $control_md5=array('N/A','N/A','N/A');

  if(file_exists($spc_xml_file)) {
	  $md5sums = new SimpleXMLElement(file_get_contents($spc_xml_file));
  } else {
	  $md5sums = "<?xml version='1.0' standalone='yes'?><sensor_pattern_test>\n";
	  // sample of records: <md5sum sensor_id='0x1801' fpga_rev='0x353301e'>df5bf0c75fd10b1f9de3847076e32be2</md5sum>
	  $md5sums .= "</sensor_pattern_test>";
	  $md5sums = new SimpleXMLElement($md5sums);
  }

/*
  if(file_exists($res_fn)) {
	  $phases_values = new SimpleXMLElement(file_get_contents($res_fn));
  }
*/

// FPGA revision...-------------------------------------------------------------------------------------------------------------
  $data = elphel_fpga_read(0x13);
  $fpga_rev = sprintf("0x%x",$data);

// sensors present
  $sensors=array(false,false,false);
  $phases=array(false,false,false);

// 10353-10359 I2C connection check---------------------------------------------------------------------------------------------
  if (i2c_receive(16, 0, 0x800, 0)==0x359) {
    res_update("firmware", "ok",'');
  } else {
    res_update("firmware", "fail", "An old firmware revision or i2c doesn't work, got ".i2c_receive(16, 0, 0x800, 0));
    _finish("fail");
  }

// Detect sensors on channel 0, 1 and 2 -------------------------------------------------------------------------------------------
  for($i=0;$i<3;$i++){
    // Detect sensor on channel i

    if (i2c_receive(16, 0, 0x4800+($i+1)*0x200, 0)==0x1801){
      res_update("sensor_ch".$i.'_id', "ok", "0x1801");
      $sensors[$i]=true;
    } else
    res_update("sensor_ch".$i.'_id', "fail", "10338 is not connected to J".($i+2)." or I2C bus between 10359 and 10338 doesn't work. 	Got ".i2c_receive(16, 0, 0x4800, 0));
  }

// super mode on!
  exec("fpcf -w 4e 60"); 
//black calibration off
  exec("fpcf -i2cw16 4820 0000");
// 64f6e76342687047b2558ddd674c5927 is the correct sum with BLC off
//test mode on - vertical color bars - red-green-blue-gray-purple
  exec("fpcf -i2cw16 48a0 0041");

//initialize sensors
  exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');
  sleep(1); apply_def_settings2(0); exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null'); 
  sleep(1); apply_def_settings2(1); for($i = 0; $i < 6; $i++) exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');
  sleep(1); apply_def_settings2(2); for($i = 0; $i < 6; $i++) exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

// read sensor parameters
  for ($i=0;$i<256;$i++){
	  $init_pars3[$i]=i2c_receive(16,0,0x4800+$i,0);
  }

  $sensor_id = _sensor_id();

  //--------------------
  //get test pattern reference MD5 sum
  $md5sum = "";

  $rez = md5sum_find($md5sums, $sensor_id, $fpga_rev);
  if($rez->md5sum == "") {
    res_update('fpga_compress', 'unknown', 'control MD5SUM for sensor ID '.$sensor_id.' and FPGA revision '.$fpga_rev.'not found');
    _finish("fail");
  } else
    $md5sum = $rez->md5sum;

  //adjust phase for the system clock
  i2c_send(16, 0, 0x809, 0x20, 0);
  test_pattern_phase_adjusment();
  i2c_send(16, 0, 0x809, 0x00, 0);
  //return false;

  //////////////////////////////////////////////////////////////////////////////////////////
  // memory test ///////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////
  // initialization sequence
  i2c_send(16,0,0x850,0x0001,0); i2c_send(16,0,0x840,0x5555,0); // Disable DDR SDRAM read/write channels
  
  i2c_send(16,0,0x851,0x0001,0); i2c_send(16,0,0x841,0x7fff,0); // PRE : Addr[10] = 1, Bank = 11
  i2c_send(16,0,0x851,0x0000,0); i2c_send(16,0,0x841,0x2002,0); // Extended mode register - Enable DLL
  i2c_send(16,0,0x851,0x8000,0); i2c_send(16,0,0x841,0x0163,0); // Load Mode Register - Burst Length - 8, CAS latency - 2.5
  i2c_send(16,0,0x851,0x0000,0); i2c_send(16,0,0x841,0x8000,0); // Refresh
  i2c_send(16,0,0x851,0x0000,0); i2c_send(16,0,0x841,0x8000,0); // Refresh
  i2c_send(16,0,0x851,0x0001,0); i2c_send(16,0,0x841,0x7fff,0); // PRE : Addr[10] = 1, Bank = 11
  
  //i2c_send(16,0,0x850,0x0001,0); i2c_send(16,0,0x840,0x0000,0); // Init auto refresh
  //i2c_send(16,0,0x850,0x0000,0); i2c_send(16,0,0x840,0x5555,0); // Disable DDR SDRAM read/write channels
  i2c_send(16,0,0x850,0x0002,0); i2c_send(16,0,0x840,0x0a00,0); // Enable auto refresh
  
  $memory_test_result=memory_test();

  res_update("memory_test","", $memory_test_result);
  res_update("memory_phase_range","", $mem_dcm_range);
  
  //_finish("ok");

  //Adjust phases for detected channels-------------------------------------------------------------------------------------------
  while(list($key, $value) = each($sensors)) {
    if ($value) {
      switch($key) {
	case 0  : $chn=1; break;
	case 1  : $chn=2; break;
	case 2  : $chn=3; break;
	default : $chn=0;
      }   

      exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');
      //send_i2c_4(0x806,$chn);
      i2c_send(16, 0, 0x806, $chn, 0);

      _sensor_pattern_set();

      for ($cnt=0;$cnt<4;$cnt++) {
	if ($phases[$key]==false) {
	  if (coarse_adjustment($key,$md5sum,320,80)) {res_update('sensor_ch'.$key.'_phase', 'ok', 'success0'); $phases[$key]=true;}
	  else{
	    //file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($key+1).'&phase_shift=0xc&n=1'); /*dcm reset*/ 
	    file_get_contents('http://127.0.0.1/359/phases_adjust.php?dcm='.($key+1).'&phase_shift=0x4&n=1'); /*90 degree shift*/ 

	    //apply_def_settings(2);
	    for($i = 0; $i < 3; $i++) exec('wget http://127.0.0.1:8081/noexif/next/wait/img -O /dev/null');

	  }
	}
      }

      exec("wget http://127.0.0.1:8081/noexif/next/wait/img -O /tmp/spc.jpeg");
      if(is_file("/tmp/spc.jpeg")) {$control_md5[$key] = substr(exec("md5sum /tmp/spc.jpeg"), 0, 32);unlink("/tmp/spc.jpeg");}

      if ($phases[$key]==false) res_update('sensor_ch'.$key.'_phase', 'fail', 'Failed to adjust the correct phase');

      _sensor_pattern_unset();   
    }

    //_finish("woahahaha1");
  }

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

  //// back to broadcast channel
  //send_i2c_4(0x806,7);
  i2c_send(16, 0, 0x806, 0x39, 0);

  $ch0_phase_value=0;
  $ch1_phase_value=0;
  $ch2_phase_value=0;

  $sclk0_phase_=receive_i2c_4(16,0,0x847,0);

  $sclk0_phase_value   = $sclk0_phase_ & 0x3ff;
  $sclk0_phase_quarter = $sclk0_phase_ >> 16;

  $ch0_phase_=receive_i2c_4(16,0,0x861,0);

  $ch0_phase_value   = $ch0_phase_ & 0x1ff;
  $ch0_phase_quarter = $ch0_phase_ >> 16;

  if ($sensors[1]) $ch1_phase_=receive_i2c_4(16,0,0x862,0);

  $ch1_phase_value   = $ch1_phase_ & 0x1ff;
  $ch1_phase_quarter = $ch1_phase_ >> 16;

  if ($sensors[2]) $ch2_phase_=receive_i2c_4(16,0,0x863,0);

  $ch2_phase_value   = $ch2_phase_ & 0x1ff;
  $ch2_phase_quarter = $ch2_phase_ >> 16;


  if ($sclk0_phase_value>256) $sclk0_phase_value=$sclk0_phase_value-512;
  if ($ch0_phase_value>256) $ch0_phase_value=$ch0_phase_value-512;
  if ($ch1_phase_value>256) $ch1_phase_value=$ch1_phase_value-512;
  if ($ch2_phase_value>256) $ch2_phase_value=$ch2_phase_value-512;

  res_update("ch0_sum","", $control_md5[0]);
  res_update("ch1_sum","", $control_md5[1]);
  res_update("ch2_sum","", $control_md5[2]);

  res_update("sclk0_dcm","", $sclk0_phase_value);
  res_update("ch0_dcm","", $ch0_phase_value);
  res_update("ch1_dcm","", $ch1_phase_value);
  res_update("ch2_dcm","", $ch2_phase_value);

  res_update("sclk0_dcm_90", "", $sclk0_phase_quarter);
  res_update("ch0_dcm_90","", $ch0_phase_quarter);
  res_update("ch1_dcm_90","", $ch1_phase_quarter);
  res_update("ch2_dcm_90","", $ch2_phase_quarter);

  res_update("sclk0_dcm_work_range", "", $dcm_range[3]);
  res_update("ch0_dcm_work_range","", $dcm_range[0]);
  res_update("ch1_dcm_work_range","", $dcm_range[1]);
  res_update("ch2_dcm_work_range","", $dcm_range[2]);

  exec("fpcf -w 4e 40"); // super mode off!

  if ($phases[2] && $phases[1] && $phases[0])
    return true;
  else 
    return false;
}

/*
//SDRAM test -------------------------------------------------------------------------------------------------------------------

//$row_adr=0x3ff;
//ddr_init();

//$written_pages=array(array(),array(),array(),array(),array());

//for($i=$row_adr;$i<$row_adr+5;$i++) {
//  $written_pages[$row_adr-$i]=ddr_page_write($i);
//  res_update('Page_'.$i, 'ok', $written_pages[$row_adr-$i][0]);
//}
////for($i=$row_adr;$i<$row_adr+5;$i++) ddr_page_read($i);
*/
?>
